/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *   Copyright (C) 2003-2004 Alain Gibaud <alain.gibaud@free.fr>           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "AsmConfigWidget.h"

#include <qhbox.h>
#include <qvbox.h>
#include <qstringlist.h>
#include <qcombobox.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qpushbutton.h>
#include <qgroupbox.h>
#include <qlayout.h>
#include <klocale.h>
#include <kiconloader.h>
#include <qtooltip.h>
#include <qcheckbox.h>

#include "device_list.h"
#include "Config.h"
#include "memory.h"

DeviceChooserDialog::DeviceChooserDialog(const QString &device, QWidget *parent)
  : KDialogBase(parent, "device_chooser_dialog", true, i18n("Choose a device"), Ok|Cancel, Ok)
{
  _listView = new KListView(this);
  _listView->setAllColumnsShowFocus(true);
  connect(_listView, SIGNAL(doubleClicked(QListViewItem *, const QPoint &, int)),
          SLOT(listDoubleClicked(QListViewItem *)));
  setMainWidget(_listView);
  _listView->addColumn(i18n("Device"));
  for (uint i=0; i<DeviceUser::Nb_Types; i++)
    _listView->addColumn(i18n(DeviceUser::DATA[i].label));
  QStringList list = DeviceList::names();
  QStringList::const_iterator it = list.begin();
  KListViewItem *selected = 0;
  for(; it!=list.end(); ++it) {
    KListViewItem *item = new KListViewItem(_listView, *it);
    const DeviceList::Data *data = DeviceList::data(*it);
    Q_ASSERT(data);
    for (uint i=0; i<DeviceUser::Nb_Types; i++)
      item->setText(i+1, data->supported[i] ? "X" : "-");
    if ( device==*it ) selected = item;
  }
  if (selected) {
    _listView->setSelected(selected, true);
    _listView->ensureItemVisible(selected);
  }
}

QString DeviceChooserDialog::device() const
{
  QListViewItem *item = _listView->selectedItem();
  if ( item==0 ) item = _listView->firstChild();
  return item->text(0);
}

void DeviceChooserDialog::listDoubleClicked(QListViewItem *item)
{
  if ( item==0 ) return;
  accept();
}

DeviceChooser::DeviceChooser(bool withAuto, QWidget *parent)
  : QWidget(parent, "device_chooser")
{
  init(withAuto);
}

DeviceChooser::DeviceChooser(const QString &device, bool withAuto, QWidget *parent)
  : QWidget(parent, "device_chooser")
{
  init(withAuto);
  setDevice(device);
}

void DeviceChooser::init(bool withAuto)
{
  QHBoxLayout *top = new QHBoxLayout(this, 0, 10);
  _button = new QPushButton(this);
  connect(_button, SIGNAL(clicked()), SLOT(chooseDevice()));
  top->addWidget(_button);
  if (withAuto) {
    _auto = new QCheckBox(i18n("auto detect from source file."), this);
    connect(_auto, SIGNAL(toggled(bool)), SLOT(autoChanged(bool)));
    top->addWidget(_auto);
  } else _auto = 0;
}

void DeviceChooser::setDevice(const QString &device)
{
  QString name = (device=="*" && _auto==0 ? DeviceList::defaultDevice() : device);
  _device = (device=="*" ? DeviceList::defaultDevice() : device);
  _button->setText(label(name));
  _button->setEnabled( name!="*" );
  if (_auto) {
    _auto->blockSignals(true);
    _auto->setChecked( name=="*" );
    _auto->blockSignals(false);
  }
}

void DeviceChooser::chooseDevice()
{
  DeviceChooserDialog dialog(_device, this);
  if ( !dialog.exec() ) return;
  setDevice(dialog.device());
  emit changed();
}

void DeviceChooser::autoChanged(bool on)
{
  _button->setEnabled(!on);
  if (on) {
    _device = _button->text();
    _button->setText(label("*"));
  } else _button->setText(_device);
  emit changed();
}

QString DeviceChooser::device() const
{
  if ( _auto && _auto->isChecked() ) return "*";
  return _device;
}

QString DeviceChooser::label(const QString &device)
{
  if ( device=="*" ) return i18n("<auto>");
  return device;
}

//----------------------------------------------------------------------------
const char *AsmConfigWidget::RADIX[Nb_Radix] = { "dec", "hex" };
const char *AsmConfigWidget::WARNING_LEVELS[Nb_WarningLevels] = { "0", "1", "2" };

AsmConfigWidget::AsmConfigWidget()
{
    QVBoxLayout *top = new QVBoxLayout(this, 10, 10);

// default device
    QHBoxLayout *hbox = new QHBoxLayout(top);
    QLabel *label = new QLabel(i18n("Default device:"), this);
    hbox->addWidget(label);
    _device = new DeviceChooser(Config::defaultDevice(), true, this);
    hbox->addWidget(_device);
    hbox->addStretch(1);

// assembler config
    QString h, i, r, w, o;
    Config::getAsmConfig(h, i, r, w, o);

    QGroupBox *asmbox = new QGroupBox(2, Qt::Horizontal, i18n("Assembler"), this);
    top->addWidget(asmbox);

    // 2) hex file format
    QLabel *labpr = new QLabel(asmbox) ;
    labpr->setText(i18n("Hex file format")) ;
    ff = new QComboBox(asmbox) ;
    ff->insertItem("*") ;
    for (uint i=0; i<Memory::Nb_HexFormats; i++) {
      ff->insertItem(Memory::HEX_FORMATS[i]);
      if ( Memory::HEX_FORMATS[i]==h ) ff->setCurrentItem(i+1);
    }
    QToolTip::add(ff,i18n("Choose the default format to save hex files.\n"
                          "* means that format is specified by LIST directive" )) ;

    // 3) radix
    labpr = new QLabel(asmbox) ;
    labpr->setText(i18n("Radix")) ;
    ra = new QComboBox(asmbox) ;
    QToolTip::add(ra,i18n("Choose the default radix used by gpasm to parse integer constants\n"
                          "* means that radix is specified by LIST directive"));
    ra->insertItem("*");
    for (uint i=0; i<Nb_Radix; i++) {
      ra->insertItem(RADIX[i]);
      if ( RADIX[i]==h ) ra->setCurrentItem(i+1);
    }

    // 4) Warning level
    labpr = new QLabel(asmbox) ;
    labpr->setText(i18n("Warning level")) ;
    wa = new QComboBox(asmbox) ;
    wa->insertItem("*");
    for (uint i=0; i<Nb_WarningLevels; i++) {
      wa->insertItem(WARNING_LEVELS[i]);
      if ( WARNING_LEVELS[i]==h ) wa->setCurrentItem(i+1);
    }
    QToolTip::add(wa,i18n("For cleaner code, always prefer the highest warning level.\n"
                          "* means that warning level is specified by LIST directive")) ;


    // 5) Include Directories
    labpr = new QLabel(asmbox) ;
    labpr->setText(i18n("Include directories")) ;
    id = new QLineEdit(asmbox) ;
    id->setText(i) ;
    QToolTip::add(id,i18n("By using the pseudovariable $(SRCPATH), you can specify an include directory\n"
                          "relative to the source file directory:\n"
                          "ie: compiling /home/alain/bar/x.asm with the include directory set to $(SRCPATH)/foo\n"
                          "will set the gpasm include directory to /home/alain/bar/foo"	)) ;

    // 6) other option
    labpr = new QLabel(asmbox) ;
    labpr->setText(i18n("Other options")) ;
    oo = new QLineEdit(asmbox) ;
    oo->setText(o) ;
    QToolTip::add(oo,i18n("This item allows you to manually specify a gpasm \"exotic\" option.\n"
                          "The $(SRCPATH) pseudo variable if supported in this item as it is "
                          "in the \"Include Directories\" one.")) ;
     asmbox->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Fixed) ;

// diassembler config
    QString dother;
    bool ds;
    Config::getDAsmConfig(ds, dother);

    QGroupBox *dasmbox = new QGroupBox(2, Qt::Horizontal, i18n("Disassembler"),  this);
    top->addWidget(dasmbox);

    // B) short option
    (new QLabel(dasmbox))->setText(i18n("Plain source code")) ;
    dshort = new QCheckBox( dasmbox ) ;
    dshort->setChecked(ds) ;
    QToolTip::add(dshort,i18n("Output a source file reassemblable by gpasm"	)) ;

    // C) Custom options
    (new QLabel(dasmbox))->setText(i18n("Other options")) ;
    doo = new QLineEdit(dasmbox) ;
    doo->setText(dother) ;
    QToolTip::add(doo,i18n("This item allows you to manually specify a gpdasm \"exotic\" option.")) ;

    dasmbox->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Fixed) ;

    // END of gpdasm configuration

    top->addStretch(1);
}

void  AsmConfigWidget::doApply()
{
  Config::setAsmConfig(ff->currentText(), id->text(), ra->currentText(),
                       wa->currentText(), oo->text());
  Config::setDAsmConfig(dshort->isChecked(), doo->text());
  Config::writeDefaultDevice(_device->device());
}

QPixmap AsmConfigWidget::icon() const
{
  KIconLoader loader;
  return loader.loadIcon("piclab_config_asm", KIcon::Toolbar, KIcon::SizeMedium);
}
