/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *   Copyright (C) 2003-2004 Alain Gibaud <alain.gibaud@free.fr>           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "EditorManager.h"

#include <qiconset.h>
#include <qdragobject.h>

#include <kmessagebox.h>
#include <klocale.h>
#include <kiconloader.h>
#include <kpopupmenu.h>

#include "FNU.h"
#include "KBEditor.h"
#include "PrgWid.h"
#include "NewProject.h"
#include "PikMain.h"

EditorManager::EditorManager(PikMain *main, QWidget *parent)
  : QTabWidget(parent), _main(main), _current(0)
{
  tabBar()->installEventFilter(this);
  connect(this, SIGNAL(currentChanged(QWidget *)), SLOT(changeToEditor(QWidget *)));
}

bool EditorManager::openFile(const KURL &url)
{
  PURL purl(url);
  if ( purl.isEmpty() ) return false;
  Editor *e = findEditor(purl);
  if (e) { // document already loaded
    if ( KMessageBox::questionYesNo(this, i18n("File \"%1\" already loaded."), i18n("Reload?"), i18n("Cancel").arg(purl.filepath()))==KMessageBox::Yes )
      if ( !e->slotLoad() ) {
        closeEditor(e);
        return false;
      }
    return true;
  }
  if ( !simpleOpenFile(url) ) return false;
  KAction *a = _main->action("file_open_recent");
  static_cast<KRecentFilesAction *>(a)->addURL(url);
  return true;
}

bool EditorManager::simpleOpenFile(const PURL &url)
{
  if ( url.isEmpty() || findEditor(url) || !url.exists(true, this) ) return false;
  Editor *e = createEditor(url);
  if ( !e->slotLoad() ) {
    closeEditor(e);
    return false;
  }
  e->setFocus();
  return true;
}

bool EditorManager::openFile()
{
  Project *p = _main->projectManager()->project();
  PDirectory dir = (p ? p->directory() : PDirectory::home());
  PURL url = PURL::getOpen(dir, PURL::filter(PURL::Editable), this ,i18n("Open File"));
  return openFile(url.kurl());
}

void EditorManager::connectEditor(Editor *editor)
{
    disconnectEditor(currentEditor());
    if ( editor==0 ) return;
    for (uint i=0; ; i++) {
      const Editor::SlotData &d = Editor::SLOT_DATA[i];
      if ( d.action==0 ) break;
      if ( !(d.capability & editor->capabilities()) ) continue;
      connect(_main->action(d.action), SIGNAL(activated()), editor, d.slot);
    }
    connect(editor, SIGNAL(guiChanged()), SIGNAL(guiChanged()));
    connect(editor, SIGNAL(dropEventPass(QDropEvent *)), SLOT(slotDropEvent(QDropEvent *)));
    connect(editor, SIGNAL(statusChanged(const QString &)), SIGNAL(statusChanged(const QString &)));
}

void EditorManager::disconnectEditor(Editor *editor)
{
    if ( editor==0 ) return;
    for (uint i=0; ; i++) {
      const Editor::SlotData &d = Editor::SLOT_DATA[i];
      if ( d.action==0 ) break;
      if ( !(d.capability & editor->capabilities()) ) continue;
      disconnect(_main->action(d.action), SIGNAL(activated()), editor, d.slot);
    }
    disconnect(editor, SIGNAL(guiChanged()), this, SIGNAL(guiChanged()));
    disconnect(editor, SIGNAL(dropEventPass(QDropEvent *)), this, SLOT(slotDropEvent(QDropEvent *)));
    disconnect(editor, SIGNAL(statusChanged(const QString &)), this, SIGNAL(statusChanged(const QString &)));
}

void EditorManager::updateTitle(Editor *e)
{
  if ( e==0 ) return;
  QString name = (e->url().isEmpty() ? "<" + e->name() + ">" : e->url().filename());
  KIconLoader loader;
  QIconSet iset = loader.loadIconSet("filesave", KIcon::Small, 0), nulliset;
  changeTab(e, e->isModified() ? iset : nulliset, name);
}

EditorManager::~EditorManager()
{
  closeAllFiles();
}

bool EditorManager::closeAllFiles()
{
  while ( currentEditor() )
    if ( !closeCurrentEditor() ) return false;
  return true;
}

bool EditorManager::closeAllOthers()
{
  QValueList<Editor *> list = _editors;
  list.remove(currentEditor());
  QValueList<Editor *>::iterator it = list.begin();
  for (; it!=list.end(); ++it)
    if ( !closeEditor(*it) ) return false;
  return true;
}

bool EditorManager::closeEditor(const PURL &url)
{
  Editor *e = findEditor(url);
  if ( e==0 ) return true;
  return closeEditor(e);
}

bool EditorManager::closeEditor(Editor *e)
{
  if ( e==0 ) return true;
  if ( e->isModified() ) {
    QString name = (e->url().isEmpty() ? "<" + e->name() + ">" : "\"" + e->url().filepath() + "\"");
    int res = KMessageBox::questionYesNoCancel(this, i18n("File %1 not saved.").arg(name),
                                               0, i18n("Save"), i18n("Discard"), i18n("Cancel"));
    switch (res) {
      case KMessageBox::Yes: e->slotSave(); break;
      case KMessageBox::Cancel: return false;
    }
  }
  removePage(e);
  _editors.remove(e);
  Editor *g = static_cast<Editor *>(currentPage());
  changeToEditor(g);
  delete e;
  emit guiChanged();
  return true;
}

void EditorManager::changeToEditor(Editor *e)
{
  connectEditor(e);
  _current = e;
  if ( e==0 ) return;
  showPage(e);
  e->clearFocus();   // force a got focus signal
  e->setFocus();
  e->slotStatusChanged();
  emit guiChanged();
}

Editor *EditorManager::findEditor(const QString &name)
{
  QValueList<Editor *>::iterator it = _editors.begin();
  for (; it!=_editors.end(); ++it) if ( (*it)->name()==name ) return *it;
  return 0;
}

Editor *EditorManager::findEditor(const PURL &url)
{
  QValueList<Editor *>::iterator it = _editors.begin();
  for (; it!=_editors.end(); ++it) if ( (*it)->url()==url ) return *it;
  return 0;
}

Editor *EditorManager::createEditor(const PURL &url)
{
  Editor *e = 0;
  if ( url.fileType()==PURL::Hex )
    e = new PrgWid(url, _main->projectManager()->project(), this);
  else e = new KBEditor(url, this);
  addEditor(e);
  return e;
}

void EditorManager::addEditor(Editor *e)
{
  QValueList<Editor *>::iterator it = _editors.begin();
  for (; it!=_editors.end(); ++it) if ( *it==e ) return;
  _editors.append(e);
  addTab(e, QString::null);
  setTabEnabled(e, true);
  changeToEditor(e);
}

Editor *EditorManager::findOrCreateEditor(const PURL &url)
{
  Editor *e = findEditor(url);
  if ( e==0 ) e = createEditor(url);
  changeToEditor(e);
  return e;
}

void EditorManager::slotDropEvent(QDropEvent *event)
{
  QStringList urls;
  if ( !QUriDrag::decodeLocalFiles(event, urls)) return;
  QStringList::const_iterator it = urls.begin();
  for(; it!=urls.end(); ++it) {
    PURL url(*it);
    Editor *e = findOrCreateEditor(url);
    e->slotLoad();
  }
}

Editor *EditorManager::switchToEditor(const PURL &url)
{
  Editor *e = findEditor(url);
  if ( e==0 ) {
    if( !url.exists(true, this) ) return 0;
    e = createEditor(url);
    e->slotLoad();
  }
  changeToEditor(e);
  return e;
}

PURL EditorManager::currentFile() const
{
  Editor *e = currentEditor();
  return (e ? e->url() : PURL());
}

void EditorManager::saveAllFiles()
{
  QValueList<Editor *>::iterator it = _editors.begin();
  for (; it!=_editors.end(); ++it) {
    if ( !(*it)->isModified() ) continue;
    (*it)->slotSave();
    updateTitle(*it);
  }
  emit guiChanged();
}

PURL::List EditorManager::files() const
{
  PURL::List names;
  QValueList<Editor *>::const_iterator it = _editors.begin();
  for(; it!=_editors.end(); ++it) {
    if ( (*it)->url().isEmpty() ) continue;
    names.push_back((*it)->url());
  }
  return names;
}

void EditorManager::newFile()
{
  NewFileDialog dialog(_main->projectManager()->project(), this);
  if ( dialog.exec()!=QDialog::Accepted ) return;
  if ( dialog.url().exists(true, this) ) {
    KMessageBox::sorry(this, i18n("File \"%1\" already exists.").arg(dialog.url().filepath()));
    return;
  }
  if ( !dialog.url().create(this) ) {
    KMessageBox::sorry(this, i18n("Could not create file \"%1\".").arg(dialog.url().filepath()));
    return;
  }
  createEditor(dialog.url());
  if ( dialog.addToProject() ) _main->projectManager()->insertFile(dialog.url());
}

bool EditorManager::eventFilter(QObject *watched, QEvent *e)
{
  if ( watched!=tabBar() || e->type()!=QEvent::ContextMenu ) return false;
  QContextMenuEvent *ce = static_cast<QContextMenuEvent *>(e);
  QPoint p = tabBar()->mapFromGlobal(ce->globalPos());
  QTab *tab = tabBar()->selectTab(p);
  if ( tab==0 ) return false;
  KPopupMenu *popup = new KPopupMenu;
  for (int i=0; i<tabBar()->count(); i++)
    if ( tabBar()->tabAt(i)==tab ) {
      Editor *editor = static_cast<Editor *>(page(i));
      popup->insertTitle(editor->url().filename());
      popup->insertItem(i18n("Close"), 1);
      if ( nbEditors()>1 ) popup->insertItem(i18n("Close All Others"), 2);
      if ( editor->isModified() ) popup->insertItem(i18n("Save"), 3);
      if ( !editor->url().isEmpty() ) popup->insertItem(i18n("Reload"), 4);
      switch (popup->exec(ce->globalPos())) {
        case 1: closeEditor(editor); break;
        case 2: closeAllOthers(); break;
        case 3: editor->slotSave(); break;
        case 4: editor->slotLoad(); break;
      }
    }
  delete popup;
  return true;
}
