/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "FNU.h"

#include <qfileinfo.h>
#include <qdatetime.h>
#include <qdir.h>

#include <klocale.h>
#include <kio/netaccess.h>
#include <kfileitem.h>
#include <kfiledialog.h>
#include <kdirselectdialog.h>
#include <ktempfile.h>

#include "data/project.xpm"
#include "data/sourcefile.xpm"
#include "data/includefile.xpm"
#include "data/csourcefile.xpm"
#include "data/objectfile.xpm"

//----------------------------------------------------------------------------
PDirectory PDirectory::home()
{
  return KURL::fromPathOrURL(QDir::homeDirPath());
}

QString PDirectory::path() const
{
  return _url.path(1);
}

bool PDirectory::exists(QWidget *widget) const
{
  return KIO::NetAccess::exists(_url, true, widget);
}

bool PDirectory::create(QWidget *widget) const
{
  return KIO::NetAccess::mkdir(_url, widget);
}

PDirectory PDirectory::getExisting(const PDirectory &dir, QWidget *widget,
                                   const QString &caption)
{
  return KDirSelectDialog::selectDirectory(dir.path(), false, widget, caption);
}

//----------------------------------------------------------------------------
const PURL::Data PURL::DATA[Nb_FileTypes] = {
  { Source | Editable | Linkable,  I18N_NOOP("ASM file"),     { "asm", "src", "pic", 0 }, sourcefile_xpm  },
  { Source | Editable | Linkable,  I18N_NOOP("C file"),       { "c", 0 },                 csourcefile_xpm },
  { Linkable,                      I18N_NOOP("Object file"),  { "o", 0 },                 objectfile_xpm  },
  { Linkable,                      I18N_NOOP("Library file"), { "lib", "a", 0 },          0               },
  { Editable,                      I18N_NOOP("Hex file"),     { "hex", 0 },               0               },
  { NoProperty,                    I18N_NOOP("Project file"), { "pikprj", 0 },            project_xpm     },
  { Source | Editable,             I18N_NOOP("Include file"), { "inc", 0 },               includefile_xpm },
  { Editable,                      I18N_NOOP("List file"),    { "lst", 0 },               0               },
  { Editable,                      I18N_NOOP("Map file"),     { "map", 0 },               0               },
  { NoProperty,                    I18N_NOOP("COD file"),     { "cod", 0 },               0               },
  { NoProperty,                    I18N_NOOP("COFF file"),    { "coff", 0 },              0               },
  { NoProperty,                    I18N_NOOP("Unknown file"), { 0 },                      0               }
};

PURL::PURL(const PDirectory &dir, const QString &filename, FileType type)
{
  _url.setFileName(dir.path() + addExtension(filename, type));
}

PURL::PURL(const PDirectory &dir, const QString &filepath)
{
  _url.setFileName(dir.path() + filepath);
}

PURL::FileType PURL::fileType() const
{
  QFileInfo info(filename());
  for (uint k=0; k<Nb_FileTypes; k++)
    for (uint i=0; DATA[k].extensions[i]; i++)
      if ( info.extension(false).lower()==DATA[k].extensions[i] ) return FileType(k);
  return Unknown;
}

PURL PURL::toFileType(FileType type, bool replace, bool upper) const
{
  QFileInfo info(filename());
  QString fname = (replace ? info.baseName() : info.fileName());
  QString ext = DATA[type].extensions[0];
  if (upper) ext = ext.upper();
  return PURL(directory().path() + fname + '.' + ext);
}

QString PURL::basename() const
{
  QFileInfo info(_url.fileName());
  return info.baseName();
}

QString PURL::filename() const
{
  QFileInfo info(_url.fileName());
  return info.fileName();
}

QString PURL::filepath() const
{
//  qDebug("%s: %s + %s", _url.prettyURL().latin1(), directory().path().latin1(),
//         filename().latin1());
  return directory().path() + filename();
}

PDirectory PURL::directory() const
{
  return KURL::fromPathOrURL(_url.directory(false));
}

bool PURL::isInto(const PDirectory &dir) const
{
  return filepath().startsWith(dir.path());
};

QString PURL::relativeTo(const PDirectory &dir) const
{
  Q_ASSERT ( isInto(dir) );
  return filepath().right(filepath().length() - dir.path().length());
}

bool PURL::exists(bool source, QWidget *widget) const
{
  return KIO::NetAccess::exists(_url, source, widget);
}

bool PURL::copyTo(const PURL &destination, QWidget *widget, QString *error) const
{
  bool ok = KIO::NetAccess::copy(_url, destination._url, widget);
  if ( !ok && error ) *error = KIO::NetAccess::lastErrorString();
  return ok;
}

bool PURL::create(QWidget *widget) const
{
  if ( exists(true, widget) ) return true;
  KTempFile tmp;
  tmp.setAutoDelete(true);
  return KIO::NetAccess::upload(tmp.name(), _url, widget);
}

bool PURL::write(const QString &text, QWidget *widget, QString *error) const
{
  PFile file(*this);
  if ( file.openForWriting(widget) ) file.stream() << text;
  file.close();
  if (error) *error = file.errorString();
  return !file.hasError();
}

bool PURL::del(QWidget *widget, QString *error) const
{
  bool ok = KIO::NetAccess::del(_url, widget);
  if ( !ok && error ) *error = KIO::NetAccess::lastErrorString();
  return ok;
}

bool PURL::isDestinationUpToDate(FileType type, QWidget *widget) const
{
  if ( !data().properties & Source ) return true;
  KIO::UDSEntry suds, duds;
  if ( !KIO::NetAccess::stat(_url, suds, widget) ) return true; // source does not exists
  PURL url = toFileType(type);
  if ( !KIO::NetAccess::stat(url._url, duds, widget) ) return false; // destination does not exists
  KFileItem sitem(suds, _url), ditem(duds, url._url);
  return ( sitem.time(KIO::UDS_MODIFICATION_TIME)<ditem.time(KIO::UDS_MODIFICATION_TIME) );
}

QString PURL::addExtension(const QString &filename, FileType type)
{
  QFileInfo info(filename);
  if ( !info.extension().isEmpty() ) return filename;
  return filename + '.' + extension(type);
}

QString PURL::filter(FileType type)
{
  QString s1, s2;
  for (uint i=0; DATA[type].extensions[i]; i++) {
    if ( i!=0 ) {
      s1 += " ";
      s2 += ", ";
    }
    QString s = QString("*.") + DATA[type].extensions[i];
    s1 += s + " " + s.upper();
    s2 += s;

  }
  return s1 + "|" + i18n(DATA[type].label) + " (" + s2 + ")";
}

bool PURL::isDosFile(QWidget *widget) const
{
  PFile file(*this);
  if( !file.openForReading(widget) ) return false;
  int oldc = 0;
  for (;;) {
    int c = file.getch();
    if ( c==-1 )  break;
    if( c=='\n' && oldc=='\r' ) return true;
    oldc = c;
  }
  return false;
}

QString PURL::filter(int properties)
{
  QString filter;
  for (uint i=0; i<Nb_FileTypes; i++) {
    if ( DATA[i].properties & properties ) {
      if ( i!=0 ) filter += '\n';
      filter += PURL::filter(FileType(i));
    }
  }
  return filter;
}

PURL PURL::getOpen(const PDirectory &startDir, const QString &filter,
                   QWidget *widget, const QString &caption)
{
  return KFileDialog::getOpenURL(startDir.path(), filter, widget, caption);
}

PURL PURL::getSave(const PDirectory &startDir, const QString &filter,
                   QWidget *widget, const QString &caption)
{
  return KFileDialog::getSaveURL(startDir.path(), filter, widget, caption);
}

//----------------------------------------------------------------------------
PFile::PFile(const PURL &url)
  : _url(url), _stream(0), _tmp(0), _widget(0)
{}

void PFile::close()
{
  file().close();
  if ( !_file.name().isEmpty() ) {
    KIO::NetAccess::removeTempFile(_file.name());
    _file.setName(QString::null);
  }
  delete _stream;
  _stream = 0;
  if (_tmp) {
    if ( !KIO::NetAccess::upload(_tmp->name(), _url._url, _widget) )
      _error = KIO::NetAccess::lastErrorString();
    delete _tmp;
    _tmp = 0;
  }
}

bool PFile::openForReading(QWidget *widget)
{
  close();
  QString tmp;
  _widget = widget;
  if ( !KIO::NetAccess::download(_url._url, tmp, widget) ) {
    _error = KIO::NetAccess::lastErrorString();
    return false;
  }
  _file.setName(tmp);
  return _file.open(IO_ReadOnly);
}

bool PFile::openForWriting(QWidget *widget)
{
  close();
  _widget = widget;
  _tmp = new KTempFile;
  _tmp->setAutoDelete(true);
  return ( _tmp->status()==0 );
}

QTextStream &PFile::stream()
{
  if ( _stream==0 ) _stream = new QTextStream(&file());
  return *_stream;
}

bool PFile::hasError() const
{
  if ( !_error.isEmpty() ) return true;
  return ( file().status()!=IO_Ok );
}

QString PFile::errorString() const
{
  if ( _error.isEmpty() ) return file().errorString();
  return _error;
}

const QFile &PFile::file() const
{
  return _tmp ? *_tmp->file() : _file;
}

QFile &PFile::file()
{
  return _tmp ? *_tmp->file() : _file;
}
