/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#ifndef FNU_H
#define FNU_H

#include <qfile.h>
class QWidget;

#include <kurl.h>
class KTempFile;

//----------------------------------------------------------------------------
class PDirectory
{
public:
  PDirectory() {}
  PDirectory(const KURL &url) : _url(url) {}

  QString path() const; // with ending '/'
  bool isEmpty() const { return _url.isEmpty(); }
  bool exists(QWidget *widget) const;
  bool create(QWidget *widget) const;
  static PDirectory getExisting(const PDirectory &startWith, QWidget *widget,
                                const QString &caption);
  static PDirectory home();
  
private:
  KURL _url;
};

//----------------------------------------------------------------------------
class PURL
{
public:
  enum FileType { Asm = 0, CSource, Object, Library, Hex, Project, Include,
                  Lst, Map, Cod, Coff, Unknown, Nb_FileTypes };

  PURL() {}
  PURL(const QString &filePath) { _url.setFileName(filePath); }
  PURL(const KURL &url) : _url(url) {}
  // add correct extension if filename has no extension
  PURL(const PDirectory &path, const QString &filename, FileType type);
  PURL(const PDirectory &path, const QString &relativeFilepath);

  enum FileProperty { NoProperty = 0, Source = 1, Linkable = 2, Editable = 4};
  struct Data {
    int properties;
    const char *label;
    const char *extensions[10];
    const char **xpm_icon;
  };
  static const Data DATA[Nb_FileTypes];
  PURL toFileType(FileType type, bool replaceExtension = true,
                  bool upperCase = false) const;
  const Data &data() const { return DATA[fileType()]; }
  FileType fileType() const;
  bool isEmpty() const { return filename().isEmpty(); }
  QString basename() const; // filename without extension
  QString filename() const; // filename without path
  QString filepath() const; // filename with path
  PDirectory directory() const;
  KURL kurl() const { return _url; }
  bool isInto(const PDirectory &dir) const;
  QString relativeTo(const PDirectory &dir) const;
  bool isDestinationUpToDate(FileType type, QWidget *widget = 0) const;
  bool exists(bool source, QWidget *widget) const;
  bool create(QWidget *widget) const; // do not overwrite
  bool isDosFile(QWidget *widget) const;
  bool copyTo(const PURL &destination, QWidget *widget, QString *error = 0) const;
  bool write(const QString &text, QWidget *widget, QString *error = 0) const;
  bool del(QWidget *widget, QString *error = 0) const;

  static QString filter(FileType type);
  static QString filter(int properties);
  static QString extension(FileType type) { return DATA[type].extensions[0]; }
  static PURL getOpen(const PDirectory &startDir, const QString &filter,
                      QWidget *widget, const QString &caption);
  static PURL getSave(const PDirectory &startDir, const QString &filter,
                      QWidget *widget, const QString &caption);
  // add correct extension if filename has no extension
  static QString addExtension(const QString &filename, FileType type);

  typedef QValueList<PURL> List;
  bool operator ==(const PURL &url) const { return _url==url._url; }

private:
  KURL _url;

  friend class PFile;
};

//----------------------------------------------------------------------------
class PFile
{
public:
  PFile(const PURL &url);
  ~PFile() { close(); }

  const PURL &url() const { return _url; }
  bool openForReading(QWidget *widget);
  bool openForWriting(QWidget *widget);
  void close();
  bool hasError() const;
  QString errorString() const;

  int getch() { return _file.getch(); }
  Q_LONG readLine(char *p, Q_ULONG maxlen) { return _file.readLine(p, maxlen); }
  Q_LONG readLine(QString &s, Q_ULONG maxlen) { return _file.readLine(s, maxlen); }
  QTextStream &stream();

private:
  PURL  _url;
  QFile _file;
  QTextStream *_stream;
  QString    _error;
  KTempFile *_tmp;
  QWidget   *_widget;

  QFile &file();
  const QFile &file() const;
};

#endif
