/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *   Copyright (C) 2003-2004 Alain Gibaud <alain.gibaud@free.fr>           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "KBEditor.h"
#include "FNU.h"
#include <kfiledialog.h>
#include <klocale.h>
#include <kapplication.h>
#include <qhbox.h>
#include <qfont.h>
#include <qfontmetrics.h>
#include <qpoint.h>
#include <qevent.h>
#include <qapplication.h>
#include <qfile.h>
#include <qstring.h>
#include <qstringlist.h>
#include <qtextedit.h>
#include <qlayout.h>

#include <ktexteditor/printinterface.h>
#include <ktexteditor/document.h>

KBEditor::KBEditor(const PURL &url, QWidget *parent)
  : Editor(url, parent, "source_editor"), doc(0), view(0), old_read_write_state(true),
    selected_line(-1)
{
    // populateList();
    KLibFactory *factory ;
    factory = KLibLoader::self()->factory("libkatepart") ;

    KTextEditor::Document  *tmp ;
    tmp = (KTextEditor::Document *)
          factory->create(0L,"kate", "KTextEditor::Document") ;

    doc = (  Kate::Document * ) tmp ;

    // find HL scheme for ASM sources
    for( picHLNumber = 0 ; (picHLNumber < doc->hlModeCount())  && (doc->hlModeName( picHLNumber ) != "XPicAsm"); ++picHLNumber)  ;
    if(   picHLNumber == doc->hlModeCount() )
       for( picHLNumber = 0 ; (picHLNumber < doc->hlModeCount())  && (doc->hlModeName( picHLNumber ) != "PicAsm"); ++picHLNumber)  ;

    QVBoxLayout *top = new QVBoxLayout(this, 0, 10);
    split = new QSplitter(this) ; // ready for multiview
    split->setFrameStyle ( QFrame::TabWidgetPanel |  QFrame::Sunken);
    top->addWidget(split);

    connect(doc, SIGNAL(undoChanged()), SIGNAL(guiChanged()));
    connect(doc, SIGNAL(selectionChanged()), SIGNAL(guiChanged()));
    connect(doc, SIGNAL(textChanged()), SIGNAL(guiChanged()));
    connect(doc, SIGNAL(hlChanged()), SLOT(slotHlChanged()));
    setAcceptDrops(true);
    slotAddView();
}
//
// create a new view
//
void KBEditor::slotAddView()
{
    KTextEditor::View *v = doc->createView(split);
    Q_ASSERT(v);
    connect(v, SIGNAL(gotFocus(Kate::View *)), SLOT(slotGotFocus(Kate::View *)));
    connect(v, SIGNAL(cursorPositionChanged()), SLOT(slotStatusChanged()));
    connect(v, SIGNAL(dropEventPass(QDropEvent *)), SIGNAL(dropEventPass(QDropEvent *)));
    connect(v, SIGNAL(newStatus()), SLOT(slotStatusChanged()));
        //Change 2-30-04 Connecting to the cursorChange signal to allow the routine to run
        //on a loaded file with no text change
        //connect(v, SIGNAL(cursorPositionChanged()), this, SLOT(slotChangedText()) );

        // I don't know why this is needed for the new view to appears
        v->show();
        v->setFocus();

        // dispatch available space between views
        QValueList<int> list = split->sizes();
        QValueList<int>::Iterator it;
        int sum ;
        for (it = list.begin( ), sum = 0  ;  it != list.end()  ; ++it)   sum += *it;
        sum  /=   list.size() ;
        for (it = list.begin( )  ;  it != list.end()  ; ++it)   *it = sum ;
        split->setSizes(list) ;
        if( view == 0) view = (Kate::View * )v ; // if first view

        v->installEventFilter(this) ;

        emit guiChanged();
}
//
// Called when the focus is transfered to another view
//
void  KBEditor::slotGotFocus(Kate::View * v)
{
  view = v;
  setFocusProxy(view);
  slotStatusChanged();
  emit guiChanged();
}

void KBEditor::slotRemoveCurrentView()
{
  delete view;
  // transfer focus to the last view, which becomes  the proxy of KBEditor  AND current view
  doc->views().last()->setFocus();
  emit guiChanged();
}

uint KBEditor::value(ValueType type) const
{
  switch (type) {
  case None: return true;
  case Modified: return isModified();
  case AddViewCount: return doc->views().count() - 1;
  case UndoCount: return doc->undoCount();
  case RedoCount: return doc->redoCount();
  case SelectionCount: return (doc->hasSelection() ? 1 : 0);
  case LineNumberOn: return view->lineNumbersOn();
  case BorderOn: return view->iconBorder();
  case AutoWrapOn: return view->dynWordWrap();
  case DosEolOn: return ( view->getEol()==1 );
  }
  Q_ASSERT(false);
  return 0;
}

bool KBEditor::isModified() const
{
    return doc->isModified();
}

void KBEditor::slotUndo()
{
    doc->undo() ;
}

void KBEditor::slotRedo()
{
    doc->redo();
}

void KBEditor::slotCut()
{
  if (view) view->cut();
}

void KBEditor::slotCopy()
{
  if (view) view->copy();
}

void KBEditor::slotPaste()
{
  if (view) view->paste();
}

void KBEditor::slotSearch()
{
  if (view) view->find();
}

void KBEditor::slotReplace()
{
  if (view) view->replace();
}

void KBEditor::slotGotoLine()
{
  if (view) view->gotoLine();
}

bool KBEditor::save(const PURL &url)
{
  return doc->saveAs(url.kurl());
}

bool KBEditor::load(const PURL &url)
{
  if ( !doc->openURL(url.kurl()) ) return false;
  // set the DOS EOL status to the proper value
  view->setEol(url.isDosFile(0) ? 1 : 0);
  return true;
}

void KBEditor::slotToggleLineNumbersOn()
{
  if (view) view->toggleLineNumbersOn();
}

void  KBEditor::setReadWrite(bool rw)
{
    doc->setReadWrite(rw) ;
}
bool KBEditor::isReadWrite() const
{
    return doc->isReadWrite();
}

void  KBEditor::slotStatusChanged()
{
  uint l, c;
  view->cursorPosition(&l, &c) ;
  QString text = i18n("Line: %1  Col: %2").arg(l+1).arg(c+1);
  text += "  " + doc->hlModeName(doc->hlMode());
  if( !isReadWrite() ) text += "  " + i18n("Read only");
  //if( value(AutoWrapOn) ) text += "  " + i18n("wrap");
  //if( value(DosEolOn) ) text += "  " + i18n("Dos EOL");
  emit statusChanged(" " + text + " ");
}

void KBEditor::slotToggleBorderOn()
{
  view->toggleIconBorder();
}

void KBEditor::slotToggleDosEOLOn()
{
  // note : EOL codes
  // 0 = Unix, 1 = Dos, 2 = Mac
  view->setEol(value(DosEolOn) ? 0 : 1);
  slotStatusChanged();
  emit guiChanged(); // changing buffer status is equivalent to change text
}

void KBEditor::slotToggleWrapOn()
{
  view->setDynWordWrap( !value(AutoWrapOn) );
  slotStatusChanged();
}

// temporary mark a line as selected
// for the purpose of highlighting asm errors
// To avoid text corruption, the buffer is write-disabled
// when the line is selected.
// See   unSelectLine()  routine
void KBEditor::slotSelectLine( )
{
  // if already in readonly mode => the mouse has been clicked again
  // before unSelectLine has been called, so ignore the click	.
  if ( view && selected_line!=-1 ) {
    view->setCursorPosition(selected_line, 0);
    old_read_write_state =  isReadWrite()  ;   // save old R/W flag
    setReadWrite(false) ;           // disable W
    doc->setSelection(selected_line, 0, selected_line+1, 0) ;
  }
}

void KBEditor::slotUnSelectLine()
{
  if ( selected_line!=-1 ) {
    doc->clearSelection() ;
    selected_line = -1;
    setReadWrite(old_read_write_state) ;     // restore previous R/W state
    if (view) view->setFocus() ;
  }
}

void KBEditor::updateBookmarksMenu(KPopupMenu *pop)
{
    pop->clear ();
    QPtrList<KTextEditor::Mark> m = doc->marks ();

    for (int i=0; (uint) i < m.count(); i++)
    {
        if (m.at(i)->type& (1 << 0))
        {
            QString linetxt = doc->textLine(m.at(i)->line);
            linetxt.truncate(48);
            linetxt.append ("...");
            pop->insertItem ( QString("%1 - \"%2\"").arg(m.at(i)->line+1).arg(linetxt),
                              this, SLOT (slotGotoBookmark(int)), 0, i );
        }
    }
}

void KBEditor::slotGotoBookmark(int i)
{
  view->gotoMark(doc->marks().at(i)) ;
}

void KBEditor::slotComment()
{
  if (view) view->comment();
}

void KBEditor::slotUncomment()
{
  if (view) view->uncomment();
}

bool KBEditor::selectLine(int line)
{
    if( selected_line != -1) return false ;
    selected_line = line ;
    return true ;
}

#if 0
void KBEditor::slotChangedText()
{
    //This slot runs the instrucion set help bar,
    // finds the current command word and compares it to the set of instrucions
    //Found in the descript Qstringlist.


    QString currentword;
    QString testval;

    Kate::View *v =  currentView();
    currentword = v->currentTextLine();
    //prepare the string for compareing
    currentword = currentword.simplifyWhiteSpace();
    currentword = currentword.upper();

    for ( QStringList::Iterator it = descript.begin(); it != descript.end(); ++it )
    {
        testval = *it;
        if(testval.startsWith(currentword.left(5)))
        {
            //if (testval.left(5) == currentword.left(5) ) {
            lab_curword->setText(*it);
        }
        //	else {
        //		lab_curword->setText("Pic Instruction Help");
        //	}
    }


}
#endif
#if 0
void KBEditor::populateList()
{
    //Populate the qstringlist with the pic instruction set, and the text to go along with them

    descript	+= "CLRF  : Clear F	OPERANDS: f";
    descript	+= "ADDWF : Add W to F	OPERANDS: f,d";
    descript	+= "ANDWF : AND W with F OPERANDS: f,d";
    descript	+= "CLRW  : Clear W	OPERANDS: NONE";
    descript	+= "COMF  : Complement F OPERANDS: f,d";
    descript 	+= "DECF  : Decrement  F OPERANDS: f,d";
    descript 	+= "DECSSZ: Decrement F, Skip 0 OPERANDS: f,d";
    descript	+= "INCF  : Increment F  OPERANDS: f,d";
    descript	+= "INCFSZ: Increment F, Skip 0 OPERANDS: f,d";
    descript	+= "IORWF : Inclusive OR W with F OPERANDS: f,d";
    descript	+= "MOVF  : Move F,  OPERANDS: f,d";
    descript 	+= "MOVWF : Move W to F OPERANDS: f,d";
    descript	+= "NOP   : No Operation OPERANDS: NONE";
    descript	+= "RLF   : Rotate Left F through Carry OPERANDS: f,d";
    descript	+= "RRF   : Rotate Right F through Carry OPERANDS: f,d";
    descript	+= "SUBWF : Subtract W from F OPERANDS: f,d";
    descript	+= "SWAPF : Swap Nibbles in F OPERANDS: f,d";
    descript	+= "XORWF : Exclusive OR W with F OPERANDS: f,s";
    descript	+= "BCF   : Bit Clear F OPERANDS: f,b";
    descript	+= "BSF   : Bit Set F   OPERANDS: f,b";
    descript	+= "BTFSC : Bit Test F, Skip if Clear OPERANDS: f,b";
    descript	+= "BTFSS : Bit Test F, Skip if Set   OPERANDS: f,b";
    descript	+= "ADDLW : Add Literal and W OPERANDS: k";
    descript	+= "ANDLW : And Literal and W OPERANDS: k";
    descript	+= "CLRWDT: Clear Watchdog Timer OPERANDS: NONE";
    descript 	+= "CALL  : Call Subroutine OPERANDS: k";
    descript	+= "GOTO  : Goto address OPERANDS: k";
    descript	+= "IORLW : Inclusive OR literal with W OPERANDS: k";
    descript	+= "MOVLW : Move Literal with W  OPERANDS: k";
    descript	+= "RETFIE: Return From Interrupt OPERANDS: NONE";
    descript	+= "RETLW : Return with Literal in W OPERANDS: k";
    descript	+= "RETURN: Return from subroutine OPERANDS: NONE";
    descript	+= "SLEEP : Go into Standby Mode OPERANDS: NONE";
    descript	+= "SUBLW : Subtract W from Literal OPERANDS: k";
    descript	+= "XORLW : Exclusive OR Literal with W OPERANDS: k";

}
#endif

void KBEditor::slotFindNext()
{
  if (view) view->findAgain();
  else QApplication::beep();
}

void KBEditor::slotFindPrevious()
{
  if (view) view->findPrev();
  else QApplication::beep();
}

void KBEditor::slotPrintDialog()
{
  if(doc) KTextEditor::printInterface(doc)->printDialog();
}

void KBEditor::slotHlChanged()
{
   // cout << "current HL = " << doc->hlMode() << " (" << doc->hlModeName( doc->hlMode() ) << ")" << endl ;
   // QString modeName(doc->hlModeName( doc->hlMode() )) ;
   // modeName =   modeName.lower() ;

   if(  picHLNumber ==  doc->hlModeCount()) return ; // no HL for pic ASM

   // because of collisions between file extensions
   // used by PiKdev and other languages/softwares
   // (ex: other ASM and also PHP ..)
   // I have to manage HL myself
   if ( (url().fileType()==PURL::Include || url().fileType()==PURL::Asm) && doc->hlMode()!=picHLNumber )
     doc->setHlMode(picHLNumber);
}

void KBEditor::configureEditor()
{
  doc->configDialog() ;
}

KBEditor::~KBEditor()
{
  delete doc ;
}
