/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *   Copyright (C) 2003-2004 Alain Gibaud <alain.gibaud@free.fr>           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "MemoryDisplayer.h"

#include <qframe.h>
#include <qgroupbox.h>
#include <qlabel.h>
#include <qscrollbar.h>
#include <qgrid.h>
#include <qhbox.h>
#include <qtooltip.h>
#include <qregexp.h>
#include <qcolor.h>
#include <qlayout.h>

#include <klocale.h>

#include "programmer.h"
#include "memory.h"
#include "microchip/misc.h"

//-----------------------------------------------------------------------------
MemorySpaceEditor::MemorySpaceEditor(Memory &memory, MemorySpace space,
                              bool withLabel, QWidget *parent, const char *name)
  : QWidget(parent, name), _memory(memory), _space(space)
{
  _top = new QVBoxLayout(this, 0, 10);
  if (withLabel) {
    QLabel *label = new QLabel(memory.label(space), this);
    _top->addWidget(label);
  }
}

//-----------------------------------------------------------------------------
MemoryDisplayerBase::MemoryDisplayerBase(QWidget *parent, Memory &memory,
                        MemorySpace space, uint lin, uint col,
                        uint offset, int length)
  : MemorySpaceEditor(memory, space, false, parent, "memory_displayer_base"),
    _firstAddress(memory.hexFileStart(space) + offset),
    _nbLines(lin), _nbCols(col)
{
    decodeSPDescriptor(memory.specialLocations());

    QHBoxLayout *hbox = new QHBoxLayout(_top);
    QFont f("courier", font().pointSize());

    // addresses
    QGrid *adr = new QGrid(1, QGrid::Horizontal, this);
    hbox->addWidget(adr);
    adr->setSpacing(0);
    adr->setMargin(3);
    adr->setFrameStyle(QFrame::Panel | QFrame::Sunken);
    adr->setFont(f);
    for (uint k=0; k<_nbLines; ++k) {
      QLabel *label = new QLabel(toHex(0, memory.nbCharsAddress(space)), adr);
      adresses.append(label);
    }

    // memory
    QGrid *grid = new QGrid(_nbCols, QGrid::Horizontal, this);
    hbox->addWidget(grid);
    grid->setSpacing(0);
    grid->setMargin(3);
    grid->setFrameStyle(QFrame::Panel |  QFrame::Sunken) ;
    grid->setFont(f);
    for (uint i = 0; i<_nbCols*_nbLines; ++i) {
      HexValueEditor *h = new HexValueEditor(grid, 0, memory.mask(space), memory.nbCharsWord(space));
      editors.append(h);
      connect(h, SIGNAL(modified()), SIGNAL(modified()));
    }

    setBase(_firstAddress);

    //  scrollbar if there are more lines to display than visible
    uint len = (length==-1 ? memory.length(space) : length);
    if ( (len / _nbCols)>_nbLines ) {
        QScrollBar *scrollbar = new QScrollBar(QScrollBar::Vertical, this);
        hbox->addWidget(scrollbar);
        int min = memory.hexFileStart(space) / _nbCols;
        int max = ((memory.hexFileStart(space) + len + 1) / _nbCols) - _nbLines;
        if ( (memory.length(space) % _nbCols)!=0 ) ++max;
        if( max<0 ) max = 0;
        // max must be set before min
        scrollbar->setMaxValue(max) ;
        scrollbar->setMinValue(min) ;
        connect(scrollbar, SIGNAL(valueChanged(int)), SLOT(setFirstLine(int))) ;
    }
}

void MemoryDisplayerBase::setReadWrite(bool readWrite)
{
  for (HexValueEditor *h=editors.first(); h; h = editors.next())
    h->setReadWrite(readWrite);
}

void MemoryDisplayerBase::setBase(int addr)
{
    if ( !_memory.isAddressInside(_space, addr) ) return;

    // memory
    int i = addr;
    for (HexValueEditor *h=editors.first(); h; h = editors.next()) {
      h->setNewAdress(_memory.dataAtAddress(_space, i));
      if ( specialLocs.find(i)!=specialLocs.end() ) { // this is a special location
        h->setColor( QColor(specialLocs[i].first) );
        QToolTip::add(h, specialLocs[i].second);
      } else {
        h->setColor(Qt::white);
        QToolTip::remove(h);
      }
      ++i;
    }

    // adresses
    i = addr;
    for (QLabel *ad=adresses.first(); ad; ad = adresses.next()) {
      uint a = (_memory.nbCharsWord(_space)>4 ? 2 : 1) * i;
      ad->setText(toHex(a, _memory.nbCharsAddress(_space)));
      i += _nbCols;
    }

    _firstAddress = addr;
}

void MemoryDisplayerBase::decodeSPDescriptor(const char *descriptor)
{
  QRegExp r("[ ]*([0-9a-fA-F]+|\\[[0-9a-fA-F]+-[0-9a-fA-F]+\\])(?:=#([0-9a-fA-F]+))?(?:\\{([^\\}]+)\\})?") ;
  QRegExp r1("\\[([0-9a-fA-F]+)-([0-9a-fA-F]+)\\]") ;

    int low, hi, color = 0xFFFF6B /* yellow */, offset=0 ;

    while( r.search(descriptor+offset ) != -1 )
    {

        offset += r.matchedLength() ;

        if( r1.search(r.cap(1) ) == -1)
        {
            low = hi = r.cap(1).toInt(0,16) ;
        }
        else
        {
            low = r1.cap(1).toInt(0,16) ;
            hi = r1.cap(2).toInt(0,16) ;
        }

        if( r.cap(2) != "") color =  r.cap(2).toInt(0,16) ;
        // else keep default or previous color

        // now, populate map
        if( hi < low ) { int t = hi ; hi = low ; low = t ; }
        for(int i = low ; i <= hi ; ++i)
            specialLocs[i] = std::make_pair(color,r.cap(3)) ;
    }
}

//-----------------------------------------------------------------------------
MemoryDisplayer::MemoryDisplayer(QWidget *parent, Memory &memory,
                                 MemorySpace space, uint lin, uint col)
  : MemorySpaceEditor(memory, space, true, parent, "memory_editor")
{
  QHBoxLayout *hbox = new QHBoxLayout(_top);
  _base = new MemoryDisplayerBase(this, memory, space, lin, col);
  connect(_base, SIGNAL(modified()), SIGNAL(modified()));
  hbox->addWidget(_base);
  hbox->addStretch(1);
}
