/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *   Copyright (C) 2003-2004 Alain Gibaud <alain.gibaud@free.fr>           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "PikMain.h"

#include <qpixmap.h>
#include <qiconset.h>
#include <qlayout.h>
#include <qsplitter.h>
#include <qregexp.h>
#include <qstringlist.h>
#include <qtimer.h>
#include <qprogressbar.h>
#include <qeventloop.h>

#include <kstatusbar.h>
#include <klocale.h>
#include <kedittoolbar.h>
#include <kapplication.h>
#include <kconfig.h>
#include <kaction.h>
#include <kmessagebox.h>
#include <kcmdlineargs.h>
#include <kprocess.h>

#include "PrgWid.h"
#include "ProjectWid.h"
#include "ConfigCenter.h"
#include "FNU.h"
#include "Project.h"
#include "Config.h"
#include "PrgConfigWidget.h"
#include "editor.h"
#include "AsmConfigWidget.h"
#include "device_list.h"
#include "KBEditor.h"

//------------------------------------------------------------------------------
CompileLog::CompileLog(QWidget *parent, PikMain *main)
  : Log::Widget(parent, "compile_log"), _main(main)
{
  connect(this, SIGNAL(clicked(int, int)), SLOT(lineClicked(int)));
}

void CompileLog::clear()
{
  Log::Widget::clear();
  _map.clear();
}

void CompileLog::appendLine(const QString &line)
{
  QRegExp asmError("^([^:]*):([0-9]+):(.+)\\[([0-9]+)\\](.+)$");
  if ( asmError.search(line)!=-1 ) {
    PURL url(asmError.cap(1));
    append(Log::Error, url.filename() + ":" + asmError.cap(2)
                       + ": " + asmError.cap(3) + ":" + asmError.cap(5));
    _map[paragraphs()-1] = Data(url, asmError.cap(2).toUInt()-1);
    return;
  }

  QRegExp linkError("^error:.*$");
  if ( linkError.search(line)!=-1 ) {
    append(Log::Error, line);
    return;
  }

  QRegExp linkDebug("^debug:.*$");
  if ( linkDebug.search(line)!=-1 ) {
    append(Log::Debug, line);
    return;
  }

  QRegExp linkInfo("^message:.*$");
  if ( linkInfo.search(line)!=-1 ) {
    append(Log::Information, line);
    return;
  }

  append(Log::Data, line); // unrecognized ??
}

void CompileLog::lineClicked(int line)
{
  if ( !_map.contains(line) ) return;
  Editor *e =  _main->editorManager()->switchToEditor(_map[line].url);
  if ( e && e->selectLine(_map[line].line) ) {
    QTimer::singleShot(10, e, SLOT(slotSelectLine()));
    QTimer::singleShot(200, e, SLOT(slotUnSelectLine()));
  }
}

//------------------------------------------------------------------------------
const char *PikMain::COMMAND_LABELS[Nb_CommandTypes] = {
  I18N_NOOP("Building project..."), I18N_NOOP("Building project..."),
  I18N_NOOP("Assemble file..."), I18N_NOOP("Diassemble file..."),
};

PikMain::PikMain()
  : project_wid(0), _process(0)
{
   _commandData.type = Nb_CommandTypes;

// status bar
    _actionStatus = new QLabel(statusBar());
    statusBar()->addWidget(_actionStatus);
    _actionProgress = new QProgressBar(statusBar());
    statusBar()->addWidget(_actionProgress);
    _editorStatus = new QLabel(statusBar());
    statusBar()->addWidget(_editorStatus, 0, true);
    _deviceStatus = new QLabel(statusBar());
    statusBar()->addWidget(_deviceStatus, 0, true);

// interface
    QSplitter *splitv = new QSplitter(this);
    splitv->setOrientation(Horizontal);
    splitv->setOpaqueResize(true);
    setCentralWidget(splitv);

    project_wid = new ProjectWid(this, splitv);
    connect(project_wid, SIGNAL(guiChanged()), SLOT(updateGUI()));

    QSplitter *splith = new QSplitter(splitv);
    splith->setOrientation(Vertical) ;
    splith->setOpaqueResize(true);

    edimanager = new EditorManager(this, splith);
    connect(edimanager, SIGNAL(guiChanged()), SLOT(updateGUI()));
    connect(edimanager, SIGNAL(statusChanged(const QString &)),
            _editorStatus, SLOT(setText(const QString &)));

    logTab = new QTabWidget(splith);
    compileLog = new CompileLog(logTab, this);
    logTab->addTab(compileLog, i18n("Compile Log"));
    compileLog->setMinimumHeight(50);
    programLog = new Log::Widget(logTab);
    logTab->addTab(programLog, i18n("Program Log"));

// actions
    // file actions
    KStdAction::openNew(edimanager, SLOT(newFile()), actionCollection());
    KStdAction::open(edimanager, SLOT(openFile()), actionCollection());
    KRecentFilesAction *recent_file =
      KStdAction::openRecent(edimanager, SLOT(openFile(const KURL &)), actionCollection());
    recent_file->setMaxItems(20);    KStdAction::save(0, 0, actionCollection());
    KStdAction::saveAs(0, 0, actionCollection());
    (void)new KAction(i18n("Save All"), /*"file_save_all"*/0, 0,
                      edimanager, SLOT(saveAllFiles()),
                      actionCollection(), "file_save_all");
    KStdAction::print(0, 0, actionCollection());
    KStdAction::close(edimanager, SLOT(closeCurrentEditor()), actionCollection());
    (void)new KAction(i18n("C&lose All"), /*"file_close_all"*/0, 0,
                      edimanager, SLOT(closeAllFiles()),
                      actionCollection(), "file_close_all");
    (void)new KAction(i18n("Close All Others"), 0, 0,
                      edimanager, SLOT(closeAllOthers()),
                      actionCollection(), "file_close_all_others");
    KStdAction::quit(this, SLOT(close()), actionCollection());

    // edit actions
    KStdAction::undo(0, 0, actionCollection());
    KStdAction::redo(0, 0, actionCollection());
    KStdAction::cut(0, 0, actionCollection());
    KStdAction::copy(0, 0, actionCollection());
    KStdAction::paste(0, 0, actionCollection());
    KStdAction::find(0, 0, actionCollection());
    KStdAction::findNext(0, 0, actionCollection());
    KStdAction::findPrev(0, 0, actionCollection(), "edit_find_previous");
    KStdAction::replace(0, 0, actionCollection());
    KStdAction::gotoLine(0, 0, actionCollection(), "edit_goto_line");
    //ka->setIcon("goto"); // the standard gotoLine action has no icon ...
    (void)new KAction(i18n("Com&ment"), 0, Qt::CTRL+Qt::Key_NumberSign,
                 actionCollection(), "edit_comment");
    (void)new KAction(i18n("Uncomme&nt"), 0, Qt::CTRL+Qt::SHIFT+Qt::Key_NumberSign,
                 actionCollection(), "edit_uncomment");

    // view actions
    (void)new KAction(i18n("&Split View"), "view_left_right",
                 Qt::CTRL+Qt::SHIFT+Qt::Key_L, actionCollection(), "view_split_view");
    (void)new KAction(i18n("&Close Current View"), "view_remove",
                 Qt::CTRL+Qt::SHIFT+Qt::Key_R, actionCollection(), "view_remove_view");
    (void)new KToggleAction(i18n("Show &Line numbers"), 0,
                       0, actionCollection(), "view_line_numbers");
    (void)new KToggleAction(i18n("Show Icon &Border"), 0,
                       0, actionCollection(), "view_icon_borders");
    (void)new KToggleAction(i18n("Auto Wrap Lines"), 0,
                       0, actionCollection(), "view_auto_wrap");
    (void)new KToggleAction(i18n("Dos End Of Line"), 0,
                       0, actionCollection(), "view_dos_eol");

    // project actions
    (void)new KAction(i18n("New Project..."), "piclab_createproject",
                      0, project_wid, SLOT(slotNewProject()), actionCollection(), "project_new");
    (void)new KAction(i18n("Open Project..."), "piclab_openproject",
                      0, project_wid , SLOT(slotOpenProject()), actionCollection(), "project_open");
    KRecentFilesAction *recent_project =
      new KRecentFilesAction(i18n("Open Recent Project"), 0,
                             project_wid, SLOT(openProjectFile(const KURL &)),
                             actionCollection(), "project_open_recent");
    recent_project->setMaxItems(20);
    (void)new KAction(i18n("Project Options..."), "piclab_editproject",
                      0, project_wid , SLOT(slotEditProject()), actionCollection(), "project_options");
    (void)new KAction(i18n("Close Project"), "piclab_closeproject",
                      0, this, SLOT(closeProject()), actionCollection(), "project_close");
    (void)new KAction(i18n("Add File..."), "piclab_addfile",
                      0, project_wid, SLOT(slotInsertFile()), actionCollection(), "project_add_file");
    (void)new KAction(i18n("Add Current File"), "piclab_addcurrentfile",
                      0, project_wid, SLOT(slotInsertCurrentFile()), actionCollection(), "project_add_current_file");

    // bookmarks actions
    //KStdAction::addBookmark(0, 0, actionCollection());

    // build actions
    (void)new KAction(i18n("&Build Project"), "piclab_compile",
                 Qt::Key_F6, this, SLOT(buildProject()), actionCollection(), "build_build_project");
    (void)new KAction(i18n("&Assemble File"), "piclab_compile",
                 Qt::Key_F7, this, SLOT(assembleFile()), actionCollection(), "build_assemble_file");
    (void)new KAction(i18n("&Disassemble"), "piclab_decompile",
                 Qt::Key_F7+Qt::SHIFT, this, SLOT(disAssemblePrg()), actionCollection(), "build_disassemble");
    (void)new KAction(i18n("Pikloops"), 0,
                 0, this , SLOT(slotRunPikloops()), actionCollection(), "build_pikloops");
    (void)new KAction(i18n("&Clean Project"), 0,
                 0, this, SLOT(cleanProject()), actionCollection(), "build_clean_project");

    // prog actions
    (void)new KAction(i18n("&Program"), "piclab_burnchip",
                 Qt::Key_F8, this , SLOT(program()), actionCollection(), "prog_program");
    (void)new KAction(i18n("&Verify"), "piclab_verifychip",
                 Qt::Key_F9, this , SLOT(verify()), actionCollection(), "prog_verify");
    (void)new KAction(i18n("&Read"), "piclab_readchip",
                 0, this , SLOT(read()), actionCollection(), "prog_read");
    (void)new KAction(i18n("&Erase"), "piclab_erasechip",
                 0, this , SLOT(erase()), actionCollection(), "prog_erase");
    (void)new KAction(i18n("&Blank Check"), 0,
                 0, this , SLOT(blankCheck()), actionCollection(), "prog_blank_check");

    // settings actions
    (void)KStdAction::configureToolbars(this, SLOT(slotConfigureToolbar()), actionCollection());
    (void)new KAction(i18n("Configure Editor..."), Qt::CTRL+Qt::SHIFT+Qt::Key_E,
                      this, SLOT(confEditor()), actionCollection(), "options_configure_editor");
    (void)KStdAction::preferences(this, SLOT(slotConfigCenter()), actionCollection());

  KRecentFilesAction *a = static_cast< KRecentFilesAction *>(action("project_open_recent"));
  a->loadEntries(kapp->config(), "recent-projects");
  a = static_cast< KRecentFilesAction *>(action("file_open_recent"));
  a->loadEntries(kapp->config(), "recent-files");

    createGUI();
    setAutoSaveSettings();

// init
    QValueList<int> szv;
    szv+=80;
    szv+=350;
    splitv->setSizes(szv);
    QValueList<int> szh;
    szh+=300;
    szh+=80;
    splith->setSizes(szh);
    updateGUI();

    QTimer::singleShot(0, this, SLOT(initialLoading()));
}

void PikMain::initialLoading()
{
  // once everything is properly set up, open  editors or project
  // as specified by the command line
  KCmdLineArgs *args = KCmdLineArgs::parsedArgs();
  for (int i = 0; i<args->count(); i++) {
    PURL url = args->url(i);
    if ( url.fileType()==PURL::Project && project_wid->project()==0 )
      project_wid->openProjectFile(url);
    else if ( url.data().properties & PURL::Editable )
      edimanager->simpleOpenFile(url);
  }

  // now, try to reopen project/files
  // note: project or files specified with command line args
  // have precedence over previous session ones.
  PURL project;
  PURL::List files;
  Config::getOpened(project, files);
  if ( project_wid->project()==0 && !project.isEmpty() && project.exists(true, this) )
    project_wid->openProjectFile(project);
  PURL::List::const_iterator it = files.begin();
  for (; it!=files.end(); ++it) edimanager->simpleOpenFile(*it);

// check

// focus
  if ( edimanager->currentEditor() ) {
    if ( edimanager->currentEditor()->inherits("KBEditor") )
      static_cast<KBEditor *>(edimanager->currentEditor())->view->setFocus();
    //else edimanager->currentEditor()->setFocus(); #### FIXME
  }
}

PikMain::~PikMain()
{
  delete _process;
}

KAction *PikMain::action(const char* name)
{
  KAction *a = actionCollection()->action(name);
  if ( a==0 ) qWarning("unknown action %s", name);
  return a;
}

void PikMain::saveSettings()
{
    KRecentFilesAction *a = static_cast< KRecentFilesAction *>(action("project_open_recent"));
    a->saveEntries(kapp->config(), "recent-projects");
    a = static_cast< KRecentFilesAction *>(action("file_open_recent"));
    a->saveEntries(kapp->config(), "recent-files");

    PURL::List files = edimanager->files();
    PURL project;
    Project *p = project_wid->project();
    if (p) project = p->url();
    Config::setOpened(project, files);
}

void PikMain::assembleProjectFile()
{
  Q_ASSERT( !_commandData.todo.isEmpty() );
  PURL url = _commandData.todo[0];
  _commandData.todo.remove(url);

  QStringList args;
  args += "gpasm";
  args += "-c"; // relocatable code
  // includes
  Project *proj = project_wid->project();
  QString expanded_include = proj->includeDir() ;
  expanded_include.replace(QRegExp("\\$\\(PRJPATH\\)"), proj->directory().path());
  expanded_include.replace(QRegExp("//"), "/" );
  if( !expanded_include.isEmpty() ) {
    args += "-I";
    args += expanded_include;
  }
  // processor
  args += "-p";
  args += DeviceList::toGPUtilsName(proj->device());
  // radix
  QString x = proj->radix();
  if( x!="*" && !x.isEmpty() ) {
    args += "-r";
    args += x;
  }
  // warnlevel
  x = proj->warnLevel() ;
  if( x!="*" && !x.isEmpty() ) {
    args += "-w";
    args += x;
  }
  // misc option
  x = proj->otherOptions() ;
  if( !x.isEmpty() ) args += x;
  // source file
  args += url.relativeTo(proj->directory());
  _commandData.type = AssembleStep;
  _commandData.url = url;
  executeCommand(proj->directory(), args);
}

void PikMain::buildProject()
{
  if ( _process && _process->isRunning() ) _process->kill();
  compileLog->clear();
  logTab->showPage(compileLog);

  Project *proj = project_wid->project();
  PURL::List files = proj->absoluteFiles();
  PURL::List::const_iterator it = files.begin();
  _commandData.todo.clear();
  for (; it!=files.end(); ++it) {
    // save modified editors
    Editor *e = edimanager->findEditor(*it);
    if ( e && e->isModified() ) e->slotSave();
    // do not process non source or up to date source files
    if ( !((*it).data().properties & PURL::Source) ) continue;
    _commandData.todo += *it;
  }
  if ( _commandData.todo.count()==0 ) linkProject();
  else {
    if ( !hasGPUtils("gpasm") ) return;
    _commandData.link = true;
    assembleProjectFile();
  }
}

void PikMain::linkProject()
{
  if ( !hasGPUtils("gplink") ) return;

  Project *proj = project_wid->project();
  PURL::List files = proj->absoluteFiles();
  PURL::List::const_iterator it = files.begin();
  uint filecount = 0;
  for (; it!=files.end(); ++it)
    if ( (*it).data().properties & PURL::Linkable ) filecount++;
  if ( filecount==0 ) {
    KMessageBox::sorry(this, i18n("No file to link!"));
    return;
  }

  QStringList args;
  args += "gplink";
  PURL hex = proj->outputFile();
  args += "-o";
  args += hex.filename();
  // -a option
  QString x = proj->hexFormat();
  if ( !x.isEmpty() ) {
    args += "-a";
    args += x;
  }
  // -m option
  args += "-m";
  // -d option
  if ( proj->linkDebug() ) args += "-d";
  // -I option
  x = proj->libsDir();
  if ( !x.isEmpty() ) {
    args += "-I";
    args += x;
  }
  // misc options
  x = proj->otherLinkOptions();
  if( !x.isEmpty() ) args += x;
  // objects files corresponding to src files
  for (it=files.begin(); it!=files.end(); ++it)
    if ( (*it).data().properties & PURL::Source )
      args += (*it).toFileType(PURL::Object).filepath();
  // objects
  for (it=files.begin(); it!=files.end(); ++it)
    if ( (*it).fileType()==PURL::Object )
      args += (*it).filepath();
  // libs must be at the end of the command
  for (it=files.begin(); it!=files.end(); ++it)
    if ( (*it).fileType()==PURL::Library )
      args += (*it).filepath();
  _commandData.type = Link;
  _commandData.url = hex;
  executeCommand(proj->directory(), args);
}

void PikMain::assembleFile()
{
  if ( _process && _process->isRunning() ) _process->kill();
  compileLog->clear();
  logTab->showPage(compileLog);
  if ( !hasGPUtils("gpasm") ) return;

  Editor *e = edimanager->currentEditor();
  Q_ASSERT( e && e->url().fileType()==PURL::Asm );
  if ( e->isModified() ) e->slotSave(); // buffer is systematically saved

  const Project *p = project_wid->project();
  if ( p && p->absoluteFiles().contains(e->url()) ) {
    _commandData.todo.clear();
    _commandData.todo += e->url();
    _commandData.link = false;
    assembleProjectFile();
  } else assembleStandAloneFile(e->url());
}

void PikMain::assembleStandAloneFile(const PURL &url)
{
    QString includedirs, others, warnlevel, hexformat, radix;
    Config::getAsmConfig(hexformat, includedirs, radix, warnlevel, others);
    // expension of $(SRCPATH) in includedirs
    QString expanded_includedirs = includedirs ;
    expanded_includedirs.replace(QRegExp("\\$\\(SRCPATH\\)"), url.directory().path());
    // eventually remove //
    expanded_includedirs.replace(QRegExp("//"), "/" );
    // and in others sub-command
    QString expanded_others = others ;
    expanded_others.replace(QRegExp("\\$\\(SRCPATH\\)"), url.directory().path());
    // eventually remove //
    expanded_others.replace(QRegExp("//"), "/" );

    QStringList args;
    args += "gpasm";
    args += "-L"; // force list
    // output file
    PURL hex = url.toFileType(PURL::Hex);
    args += "-o";
    args += hex.filepath();
    // warn level
    if ( warnlevel!="*" ) {
      args += "-w";
      args += warnlevel;
    }
    if ( !expanded_includedirs.isEmpty() ) {
      args += "-I";
      args += expanded_includedirs;
    }
    QString proc = Config::defaultDevice();
    if( proc!="*" ) {
      args += "-p";
      args += DeviceList::toGPUtilsName(proc);
    }
    if( hexformat!="*" ) {
      args += "-a";
      args += hexformat;
    }
    if( radix!="*" ) {
      args += "-r";
      args += radix;
    }
    if( !expanded_others.isEmpty() ) args += expanded_others;
    args += url.filepath();
    _commandData.type = AssembleSingle;
    _commandData.url = hex;
    executeCommand(url.directory(), args);
}


void  PikMain::slotBookmarksManagement()
{
  Editor *e = edimanager->currentEditor();
  if (e) e->updateBookmarksMenu(((KActionMenu *)action("bookmarks"))->popupMenu()) ;
}

void PikMain::slotConfigureToolbar()
{
  saveMainWindowSettings(KGlobal::config(), "MainWindow");
  KEditToolbar dlg(actionCollection());
  connect(&dlg, SIGNAL(newToolbarConfig()), SLOT(slotApplyToolbarSettings()));
  if ( dlg.exec() ) createGUI();
}

void PikMain::slotApplyToolbarSettings()
{
  applyMainWindowSettings(KGlobal::config(), "MainWindow");
}

bool PikMain::executeCommand(const PDirectory &dir, const QStringList &arguments)
{
  _stdout = QString::null;
  _stderr = QString::null;
  QString command;
  QStringList::const_iterator it = arguments.begin();
  for (; it!=arguments.end(); ++it)
    command += " " + *it;
  compileLog->append(Log::Information, command);
  delete _process;
  _process = new KProcess;
  connect(_process, SIGNAL(processExited(KProcess *)), SLOT(commandDone()));
  connect(_process, SIGNAL(receivedStdout(KProcess *, char *, int)), SLOT(receivedStdout(KProcess *, char *, int)));
  connect(_process, SIGNAL(receivedStderr(KProcess *, char *, int)), SLOT(receivedStderr(KProcess *, char *, int)));
  _process->setWorkingDirectory(dir.path());
  *_process << arguments;
  if ( !_process->start(KProcess::NotifyOnExit, KProcess::AllOutput) ) {
    compileLog->append(Log::Error, i18n("Failed to execute command."));
    return false;
  }
  updateGUI();
  return true;
}

void PikMain::showExistingFile(const PURL &url, bool show, bool addToProject)
{
  qDebug("show existing %s", url.filepath().latin1());
  if ( url.exists(true, this) ) {
    if (show) {
      Editor *e = edimanager->findOrCreateEditor(url);
      e->slotLoad();
      e->setReadWrite(false);
    }
    if ( addToProject ) project_wid->addFile(url);
  } else {
    edimanager->closeEditor(url);
    if ( addToProject ) project_wid->removeFile(url);
  }
}

void PikMain::commandDone()
{
  if ( !_stderr.isEmpty() ) compileLog->appendLine(_stderr);
  int code = _process->exitStatus();
  delete _process;
  _process = 0;
  if ( code==0 ) {
    compileLog->append(Log::Data, i18n("*** Success ***"));
    switch (_commandData.type) {
      case AssembleStep:
        showExistingFile(_commandData.url.toFileType(PURL::Lst), false, true);
        if ( _commandData.todo.isEmpty() ) {
          if ( !_commandData.link ) break;
          linkProject();
        } else assembleProjectFile();
        return;
      case Link:
        showExistingFile(_commandData.url, true, true);
        showExistingFile(_commandData.url.toFileType(PURL::Map), false, true);
        showExistingFile(_commandData.url.toFileType(PURL::Lst), false, true);
        break;
      case AssembleSingle:
        showExistingFile(_commandData.url, true, false);
        showExistingFile(_commandData.url.toFileType(PURL::Lst), true, false);
        break;
      case Disassemble: {
        QString error;
        if ( _commandData.url.write(_stdout, this, &error) ) {
          KMessageBox::detailedSorry(this, i18n("Could not write output file \"%1\".").arg(_commandData.url.filepath()), error);
          break;
        }
        showExistingFile(_commandData.url, true, false);
        break;
      }
      default: Q_ASSERT(false); qDebug("type = %i", _commandData.type); break;
    }
  } else compileLog->append(Log::Error, i18n("*** Exited with status: %1 ***").arg(code));

  _commandData.type = Nb_CommandTypes;
  updateGUI();
}

void PikMain::receivedStdout(KProcess *p, char *buffer, int len)
{
  if ( _commandData.type==Disassemble ) _stdout += QString::fromLatin1(buffer, len);
  else receivedStderr(p, buffer, len);
}

void PikMain::receivedStderr(KProcess *, char *buffer, int len)
{
  for (int i=0; i<len; i++) {
    if ( buffer[i]=='\n' ) {
      compileLog->appendLine(_stderr);
      _stderr = QString::null;
    } else _stderr += QString::fromLatin1(buffer + i, 1);
  }
}

bool PikMain::isExecutablePresent(const QStringList &args) const
{
  KProcess *process = new KProcess;
  *process << args;
  bool found = process->start(KProcess::Block);
  delete process;
  return found;
}

bool PikMain::hasGPUtils(const QString &executable)
{
  QStringList args;
  args += executable;
  args += "-v";
  if ( !isExecutablePresent(args) ) {
    KMessageBox::sorry(this, i18n("<b>%1</b> not found.\n"
                                  "Please download \"gputils\" at http://gputils.sourceforge.net"));
    return false;
  }
  return true;
}

void PikMain::disAssemblePrg()
{
  if ( _process && _process->isRunning() ) _process->kill();
  if ( !hasGPUtils("gpdasm") ) return;

    PrgWid *e = static_cast<PrgWid *>(edimanager->currentEditor());
    Q_ASSERT( e && e->fileType()==PURL::Hex );
    if ( e->isModified() ) e->slotSave();
    compileLog->clear();

    QString other_opt;
    bool short_format;
    Config::getDAsmConfig(short_format, other_opt);

    QStringList args;
    args += "gpdasm";
    args += "-p";
    args += DeviceList::toGPUtilsName(e->device());
    if (short_format) args += "-s";
    if ( !other_opt.isEmpty() ) args += other_opt;
    args += e->url().filepath();
    _commandData.url = e->url().toFileType(PURL::Asm, false); // file name xxx.hex => xxx.hex.asm
    _commandData.type = Disassemble;
    executeCommand(e->url().directory(), args);
}

void PikMain::confEditor()
{
    Editor *e = edimanager->currentEditor();
    if (e) e->configureEditor();
}

void PikMain::slotRunPikloops()
{
  if ( !isExecutablePresent("pikloops") ) {
    KMessageBox::detailedSorry(this, i18n("Cannot run \"pikloops\""),
                               i18n("The Pikloops utility (http://pikloops.tripod.com) is not installed in your system:\n"
                                    " You can download pikloops from http://pikdev.free.fr"));
    return;
  }
  KProcess *process = new KProcess;
  *process << "pikloops";
  process->start(KProcess::Block);
}

bool PikMain::queryClose()
{
  saveSettings();
  return edimanager->closeAllFiles();
}

void PikMain::slotConfigCenter()
{
  DeviceUser::Type type = Config::programmerType();

  ConfigCenter dialog(this);
  dialog.exec();

  // if programmer type is different
  // reload hex editor
  if ( Config::programmerType()!=type ) {
    PURL::List list = edimanager->files();
    PURL::List::const_iterator it = list.begin();
    for(; it!=list.end(); ++it) {
      if ( (*it).fileType()!=PURL::Hex ) continue;
      static_cast<PrgWid *>(edimanager->findEditor(*it))->programmerChanged();
    }
  }
  updateGUI();
}

void PikMain::updateGUI()
{
  // update editor actions
  Editor *e = edimanager->currentEditor();
  edimanager->updateTitle(e);
  for (uint i=0; ; i++) {
    const Editor::SlotData &d = Editor::SLOT_DATA[i];
    if ( d.action==0 ) break;
    KAction *a = action(d.action);
    if ( e==0 ) a->setEnabled(false);
    else if ( !(d.capability & e->capabilities()) ) a->setEnabled(false);
    else {
      uint v = e->value(d.type);
      if ( a->inherits("KToggleAction") ) {
        a->setEnabled(true);
        static_cast<KToggleAction *>(a)->setChecked(v);
      } else a->setEnabled(v);
    }
  }
  action("file_save_all")->setEnabled(e);
  action("file_close")->setEnabled(e);
  action("file_close_all")->setEnabled(e);
  action("options_configure_editor")->setEnabled(e && e->canConfigureEditor());

  // update project actions
  Project *p = project_wid->project();
  if (p)
    static_cast< KRecentFilesAction *>(action("project_open_recent"))->addURL(p->url().kurl());
  action("project_close")->setEnabled(p);
  action("project_options")->setEnabled(p);
  action("project_add_file")->setEnabled(p);
  action("project_add_current_file")->setEnabled(p && e);

  // update build actions
  action("build_build_project")->setEnabled(p);
  action("build_assemble_file")->setEnabled(e && e->fileType()==PURL::Asm);
  action("build_disassemble")->setEnabled(e && e->fileType()==PURL::Hex);
  action("build_clean_project")->setEnabled(p);

  // update prog actions
  bool running = (_process && _process->isRunning());
  bool enabled = ( !running && (p || (e && e->fileType()==PURL::Hex)) );
  action("prog_program")->setEnabled(enabled);
  action("prog_verify")->setEnabled(enabled);

  // update status bar
  if ( _commandData.type==Nb_CommandTypes ) {
    _actionStatus->hide();
    _actionProgress->hide();
  } else {
    _actionStatus->show();
    _actionStatus->setText(i18n(COMMAND_LABELS[_commandData.type]));
    _actionProgress->show();
  }
  if ( p && e ) project_wid->select(e->url());
  QString device = Config::defaultDevice();
  DeviceUser::Type type = Config::programmerType();
  if ( e && e->fileType()==PURL::Hex ) {
    device = static_cast<PrgWid *>(e)->device();
    type = Programmer::current().type();
  } else if ( p && (e==0 || project_wid->contains(e->url())) ) device = p->device();
  _deviceStatus->setText(" " + i18n(DeviceUser::DATA[type].label) + " - "
                         + DeviceChooser::label(device) + " ");

  // caption
  QString caption;
  if (p) caption += p->name();
  if (e) {
    if (p) caption += " - ";
    caption += e->url().filepath();
  }
  setCaption(KApplication::kApplication()->makeStdCaption(caption));
}

void PikMain::closeProject()
{
  PURL::List files = edimanager->files();
  PURL::List::const_iterator it = files.begin();
  for (; it!=files.end(); ++it)
    if ( project_wid->contains(*it) ) edimanager->closeEditor(*it);
  project_wid->slotCloseProject();
  compileLog->clear();
  programLog->clear();
}

void PikMain::cleanProject()
{
  if ( _process && _process->isRunning() ) _process->kill();
  compileLog->clear();
  logTab->showPage(compileLog);

  Project *p = project_wid->project();
  PURL::List todo;
  PURL::List files = p->absoluteFiles();
  PURL::List::const_iterator it = files.begin();
  for (; it!=files.end(); ++it) {
    if ( (*it).fileType()!=PURL::Asm ) continue;
    todo += (*it).toFileType(PURL::Object);
    todo += (*it).toFileType(PURL::Lst);
  }
  todo += p->outputFile();
  todo += p->outputFile().toFileType(PURL::Lst);
  todo += p->outputFile().toFileType(PURL::Map);
  todo += p->outputFile().toFileType(PURL::Cod);
  todo += p->outputFile().toFileType(PURL::Coff);

  for (it=todo.begin(); it!=todo.end(); ++it) {
    edimanager->closeEditor(*it);
    project_wid->removeFile(*it);
    (*it).del(this);
  }
}

//----------------------------------------------------------------------------
void PikMain::doAction(Programmer::Action action)
{
  // prepare
  PrgWid *editor = 0;
  switch (action) {
    case Programmer::Program:
    case Programmer::Verify: editor = getHexEditor(); break;
    case Programmer::Read: editor = createHexEditor(i18n("read")); break;
    case Programmer::Erase:
    case Programmer::BlankCheck: editor = createHexEditor(i18n("blank")); break;
    case Programmer::Nb_Actions: Q_ASSERT(false); break;
  }
  if ( editor==0 ) return;

  // do action
  Programmer::current().setLogWidget(programLog);
  logTab->showPage(programLog);
  programLog->clear();
  _actionStatus->show();
  _actionStatus->setText(i18n(Programmer::ACTION_LABELS[action]));
  Programmer::current().disconnect(_actionProgress);
  connect(&Programmer::current(), SIGNAL(totalSteps(int)), SLOT(setTotalSteps(int)));
  connect(&Programmer::current(), SIGNAL(progress(int)), SLOT(setProgress(int)));
  bool success = false;
  success = editor->doAction(action);

  // finish
  switch (action) {
    case Programmer::Program:
    case Programmer::Verify: break;
    case Programmer::Read:
      if (success) {
        edimanager->addEditor(editor);
        edimanager->changeToEditor(editor);
        break;
      }
    case Programmer::Erase:
    case Programmer::BlankCheck: delete editor; break;
    case Programmer::Nb_Actions: Q_ASSERT(false); break;
  }
  _actionStatus->hide();
  _actionProgress->hide();
}

PrgWid *PikMain::getHexEditor()
{
  if ( edimanager->currentEditor()->url().fileType()==PURL::Hex )
    return static_cast<PrgWid *>(edimanager->currentEditor());
  Project * project =  project_wid->project();
  Q_ASSERT(project);
  edimanager->simpleOpenFile(project->outputFile());
  PrgWid *editor = static_cast<PrgWid *>(edimanager->findEditor(project->outputFile()));
  if ( editor==0 ) {
    KMessageBox::sorry(this, i18n("Cannot open project hex file \"%1\".").arg(project->outputFile().filepath()));
    return 0;
  }
  return editor;
}

void PikMain::program()
{
  doAction(Programmer::Program);
}

void PikMain::verify()
{
  doAction(Programmer::Verify);
}

PrgWid *PikMain::createHexEditor(const QString &basename)
{
  QString name = PURL::addExtension(basename, PURL::Hex);
  Project *p = project_wid->project();
  Editor *e = edimanager->currentEditor();
  QString device;
  if ( e && e->fileType()==PURL::Hex ) device = static_cast<PrgWid *>(e)->device();
  else if (p) device = p->device();
  else device = Config::defaultDevice();
  e = edimanager->findEditor(name);
  if (e) edimanager->closeEditor(e);
  return new PrgWid(name, device, this);
}

void PikMain::read()
{
  doAction(Programmer::Read);
}

void PikMain::erase()
{
  doAction(Programmer::Erase);
}

void PikMain::blankCheck()
{
  doAction(Programmer::BlankCheck);
}

void PikMain::setTotalSteps(int total)
{
  if ( total==0 ) return;
  _actionProgress->show();
  _actionProgress->setTotalSteps(total);
  _actionProgress->setProgress(0);
  QApplication::eventLoop()->processEvents(QEventLoop::ExcludeUserInput);
}

void PikMain::setProgress(int steps)
{
  _actionProgress->setProgress(steps);
  QApplication::eventLoop()->processEvents(QEventLoop::ExcludeUserInput);
}
