/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *   Copyright (C) 2003 Alain Gibaud <alain.gibaud@free.fr>                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "PrgWid.h"

#include <qpushbutton.h>
#include <qgroupbox.h>
#include <qhgroupbox.h>
#include <qregexp.h>
#include <qlayout.h>
#include <qtimer.h>
#include <qscrollview.h>

#include <kmessagebox.h>
#include <klocale.h>
#include <kfiledialog.h>
#include <ktempfile.h>

#include "MemoryDisplayer.h"
#include "BinWordsEditor.h"
#include "programmer.h"
#include "AsmConfigWidget.h"
#include "Project.h"
#include "Config.h"
#include "device_list.h"

PrgWid::PrgWid(const QString &name, const QString &device, QWidget *parent)
  : Editor(name, parent, "hex_editor"), _project(0)
{
  init();
  _deviceChooser->setDevice(device);
}

PrgWid::PrgWid(const PURL &url, Project *project, QWidget *parent)
  : Editor(url, parent, "hex_editor"), _project(project)
{
  init();
}

PrgWid::PrgWid(const PURL &url, const QString &device, QWidget *parent)
  : Editor(url, parent, "hex_editor"), _project(0)
{
  init();
  _deviceChooser->setDevice(device);
  // #### FIXME if device=="*": autodetect...
}

void PrgWid::init()
{
  _vbox = 0;
  QHBoxLayout *hbox = new QHBoxLayout(this, 0);
  QScrollView *sview = new QScrollView(this);
  sview->setResizePolicy(QScrollView::AutoOneFit);
  hbox->addWidget(sview);
  _widget = new QWidget(sview->viewport());
  sview->addChild(_widget);
  _top = new QVBoxLayout(_widget, 10, 10);

  // device
  hbox =  new QHBoxLayout(_top);
  _deviceLabel = new QLabel(_widget);
  hbox->addWidget(_deviceLabel);
  _deviceChooser = new DeviceChooser(false, _widget);
  connect(_deviceChooser, SIGNAL(changed()), SLOT(deviceChanged()));
  hbox->addWidget(_deviceChooser);
  hbox->addStretch(1);

  _editors.setAutoDelete(true);
}

bool PrgWid::inProject() const
{
  return ( _project && _project->outputFile()==url() );
}

bool PrgWid::updateDisplay(const PURL &lurl, const QString &defaultDevice, QString &error)
{
  Programmer &prog = Programmer::current();
  QString text = (inProject() ? i18n("Project Device:") : i18n("Device:"));
  QString name = (inProject() ? _project->device() : defaultDevice);
  if ( name=="*" ) {
    bool success;
    name = DeviceList::guessDeviceFromFile(url(), success, this);
    if (success) text = i18n("Device (guessed from file):");
    else {
      text = i18n("Device (default):");
      KMessageBox::sorry(this, i18n("Target device is not configured and cannot be guessed from source file.\n"
                                    "The source file either cannot be found or does not contain any processor directive."));
    }
  }
  _deviceChooser->setDevice(name);

  if ( DeviceList::isSupported(name, prog.type()) ) {
    _memory.setDevice(name);
    if ( !_memory.loadHexFile(lurl, error, this) ) return false;
    display();
  } else {
    text = i18n("Device (not supported):");
    _editors.clear();
    delete _vbox;
    _vbox = 0;
    _top->addStretch(1);
  }
  _deviceLabel->setText(text);

  emit bufferModified();
  return true;
}

void PrgWid::display()
{
  _editors.clear();
  delete _vbox;
  _vbox = new QVBoxLayout(_top);
  for (uint i=0; i<Nb_MemorySpaces; i++) {
    MemorySpaceEditor *e = _memory.createEditor(MemorySpace(i), _widget);
    if ( e==0 ) continue;
    _editors.append(e);
    e->show();
    connect(this, SIGNAL(bufferModified()), e, SLOT(updateDisplay()));
    connect(e, SIGNAL(modified()), SLOT(slotModified()));
    _vbox->addSpacing(10);
    _vbox->addWidget(e);
  }
  _top->addStretch(1);
  setReadWrite(!inProject());
}

PrgWid::~PrgWid()
{}

uint PrgWid::value(ValueType type) const
{
  switch (type) {
  case None: return true;
  case Modified: return isModified();
  case AddViewCount: return 0;
  case UndoCount: return 0;
  case RedoCount: return 0;
  case SelectionCount: return 0;
  case LineNumberOn: return 0;
  case BorderOn: return 0;
  case AutoWrapOn: return 0;
  case DosEolOn: return 0;
  }
  Q_ASSERT(false);
  return 0;
}

bool PrgWid::save(const PURL &url)
{
  QString error;
  if ( !_memory.saveHexFile(url, Memory::IHX32, error, this) ) {
    KMessageBox::sorry(this, error);
    return false;
  }
  return true;
}

bool PrgWid::load(const PURL &url)
{
  QString error;
  if ( !updateDisplay(url, Config::defaultDevice(), error) ) {
    KMessageBox::sorry(this, error);
    return false;
  }
  return true;
}

bool PrgWid::doAction(Programmer::Action action)
{
  if ( !verifyDeviceType() ) return false;
  if ( !Programmer::current().doAction(action, _memory) ) return false;
  if ( action==Programmer::Read ) {
    display();
    emit bufferModified();
    if ( !url().isEmpty() ) slotModified();
  }
  return true;
}

bool PrgWid::verifyDeviceType()
{
  Programmer &prog = Programmer::current();
  if ( !DeviceList::isSupported(device(), prog.type()) ) {
    KMessageBox::sorry(this, i18n("Pic device \"%1\" is not supported by %2.")
                             .arg(device()).arg(i18n(DeviceUser::DATA[prog.type()].label)));
    return false;
  }

  if ( !prog.connectDevice() ) return false;
  QString name;
  uint revision;
  if ( !prog.readDeviceId(name, revision) ) return false;
  QString message;
  if ( name.isEmpty() )
    message = i18n("Device cannot be detected (not autodetectable or missing).\n"
                   "Continue with the provided device name (%1)?").arg(device());
  else if ( name!="?" && name!=device() )
    message = i18n("Detected device (%1) does not match specified target device (%2).\n"
                   "Continue with the provided device name?").arg(name).arg(device());
  if ( !message.isEmpty() ) {
    int res = KMessageBox::warningYesNo(this, message, 0, i18n("Continue"), i18n("Abort"));
    if ( res==KMessageBox::Cancel ) return false;
    name = "?";
  }
  if ( name=="?" ) prog.log(Log::Information, i18n("Will use the provided device name: %1").arg(device()));
  if ( _memory.device()!=device() ) reload();
  return prog.setupTarget();
}

void PrgWid::slotModified()
{
  _modified = true;
  emit guiChanged();
}

void PrgWid::programmerChanged()
{
  reload();
}

void PrgWid::deviceChanged()
{
  if ( !inProject() ) _deviceLabel->setText(i18n("Device:"));
  reload();
}

void PrgWid::reload()
{
  QString error;
  if ( !isModified() && !url().isEmpty() ) {
    if ( !updateDisplay(url(), device(), error) )
      KMessageBox::sorry(this, error);
  } else {
    KTempFile file;
    file.setAutoDelete(true);
    if ( !_memory.saveHexFile(file.name(), Memory::IHX32, error, this) ) {
      KMessageBox::sorry(this, i18n("Internal error saving hex data to temporary file\n(%1).").arg(error));
      return;
    }
    if ( !updateDisplay(file.name(), device(), error) )
      KMessageBox::sorry(this, i18n("Internal error reading hex data from temporary file\n(%1).").arg(error));
  }
}

QString PrgWid::device() const
{
  return _deviceChooser->device();
}

void PrgWid::setReadWrite(bool readWrite)
{
  _readWrite = readWrite;
  _deviceChooser->setEnabled(readWrite);
  QPtrListIterator<MemorySpaceEditor> it(_editors);
  for (; *it; ++it)
    (*it)->setReadWrite(readWrite);
}
