/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *   Copyright (C) 2003 Alain Gibaud <alain.gibaud@free.fr>                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "Project.h"

#include <qstringlist.h>
#include <qfileinfo.h>
#include <qdatetime.h>

#include "device_list.h"

Project::Project(const PURL &url)
  : _url(url)
{
  config = new KConfig(url.filepath(), false, false);
}

Project::~Project()
{
  config->sync();
  delete config;
}

QStringList Project::relativeFiles() const
{
  config->setGroup("Files");
  return config->readListEntry("inputFiles");
}

PURL::List Project::absoluteFiles() const
{
  PURL::List abs;
  QStringList files = relativeFiles();
  QStringList::const_iterator it = files.begin();
  for (; it!=files.end(); ++it) abs += PURL(directory(), *it);
  return abs;
}

void Project::saveFiles(const QStringList &files)
{
  config->setGroup("Files");
  config->writeEntry("inputFiles", files);
  config->sync();
}

void Project::addFile(const PURL &url)
{
  QStringList list = relativeFiles();
  list += url.relativeTo(directory());
  saveFiles(list);
}

void Project::removeFile(const PURL &url)
{
  QStringList list = relativeFiles();
  QString file = url.relativeTo(directory());
  list.remove(file);
  saveFiles(list);
}

void Project::setDescription(const QString& desc)
{
  config->setGroup("General") ;
  config->writeEntry("description", desc) ;
  config->sync();
}

void Project::setVersion(const QString& ver)
{
  config->setGroup("General") ;
  config->writeEntry("version", ver) ;
  config->sync();
}

void Project::setOutputFile(const PURL &url)
{
  config->setGroup("General");
  config->writeEntry("output", url.filename());
  config->sync();
}
PURL Project::outputFile() const
{
  config->setGroup("General");
  QString file = config->readEntry("output", QString::null);
  if ( file==QString::null ) return PURL(directory(), name(), PURL::Hex);
  QFileInfo info(file);
  return PURL(directory(), info.fileName(), PURL::Hex);
}

void Project::restore()
{
  config->deleteGroup("General",true) ;
  config->deleteGroup("Files",true) ;
  config->reparseConfiguration() ;
}

QString Project::version() const
{
  config->setGroup("General") ;
  return config->readEntry("version","0.1") ;
}

QString Project::description() const
{
  config->setGroup("General") ;
  return config->readEntry("description","<none>") ;
}

QString Project::device() const
{
  config->setGroup("Assembler") ;
  QString s = DeviceList::checkName(config->readEntry("target-device"));
  if ( s=="*" ) return DeviceList::defaultDevice(); // compatibility
  return s;
}
void Project::setDevice(const QString& x)
{
  config->setGroup("Assembler") ;
  config->writeEntry("target-device",x) ;
  config->sync();
}

QString Project::radix() const
{
  config->setGroup("Assembler") ;
  return config->readEntry("radix","dec") ;
}
void Project::setRadix(const QString& x)
{
  config->setGroup("Assembler") ;
  config->writeEntry("radix",x) ;
  config->sync();
}

QString Project::warnLevel() const
{
  config->setGroup("Assembler") ;
  return config->readEntry("warn-level","2") ;
}
void Project::setWarnLevel(const QString& x)
{
  config->setGroup("Assembler") ;
  config->writeEntry("warn-level",x) ;
  config->sync();
}

QString Project::includeDir() const
{
  config->setGroup("Assembler") ;
  return config->readEntry("include-dir","") ;
}
void Project::setIncludeDir(const QString& x)
{
  config->setGroup("Assembler") ;
  config->writeEntry("include-dir",x) ;
  config->sync();
}

QString Project::otherOptions() const
{
  config->setGroup("Assembler") ;
  return config->readEntry("other-options","") ;
}
void Project::setOtherOptions(const QString& x)
{
  config->setGroup("Assembler") ;
  config->writeEntry("other-options",x) ;
  config->sync();
}

bool Project::linkDebug() const
{
  config->setGroup("Linker") ;
  return config->readBoolEntry("debug",false) ;
}
void Project::setLinkDebug(bool x)
{
  config->setGroup("Linker") ;
  config->writeEntry("debug",x) ;
  config->sync();
}

QString Project::hexFormat() const
{
  config->setGroup("Linker") ;
  return config->readEntry("hex-format","inhx32") ;
}
void Project::setHexFormat(const QString& x)
{
  config->setGroup("Linker") ;
  config->writeEntry("hex-format",x) ;
  config->sync();
}

QString Project::libsDir() const
{
  config->setGroup("Linker") ;
  return config->readEntry("objs-libs-dir","") ;
}
void Project::setLibsDir(const QString& x)
{
  config->setGroup("Linker") ;
  config->writeEntry("objs-libs-dir",x) ;
  config->sync();
}

QString Project::otherLinkOptions() const
{
  config->setGroup("Linker") ;
  return config->readEntry("other-options","") ;
}
void Project::setOtherLinkOptions(const QString& x)
{
  config->setGroup("Linker") ;
  config->writeEntry("other-options",x) ;
  config->sync();
}
