/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *   Copyright (C) 2003 Alain Gibaud <alain.gibaud@free.fr>                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "ProjectEditorWid.h"

#include <klocale.h>
#include <kiconloader.h>

#include <qlineedit.h>
#include <qlabel.h>
#include <qtextedit.h>
#include <qgroupbox.h>
#include <qtooltip.h>
#include <qcheckbox.h>
#include <qpushbutton.h>
#include <qlayout.h>
#include <qhbox.h>

#include "Project.h"
#include "AsmConfigWidget.h"
#include "FNU.h"
#include "device_list.h"

ProjectEditorWid::ProjectEditorWid(Project *proj, QWidget *parent)
  : KDialogBase(parent, "project_options", true, i18n("Project Options"), Ok|Cancel, Ok),
    project(proj)
{
  QWidget *cw = new QWidget(this);
  setMainWidget(cw);
  QGridLayout *top = new QGridLayout(cw, 0, 0, 10, 10);

  KIconLoader loader ;
  QIconSet iconset = loader.loadIcon("fileopen", KIcon::Toolbar) ;

  QStringList hfmt ;
  (((hfmt += "inhx8m") += "inhx8s") += "inhx16") += "inhx32" ;
  QStringList rad ;
  (rad += "dec") += "hex" ;
  QStringList warn ;
  ((warn +="0") += "1") += "2" ;

  // -------------------------- global section ----------------------
  QLabel *label = new QLabel(i18n("Description"), cw);
  top->addWidget(label, 0, 0);
  _description = new QTextEdit(cw);
  _description->setText(project->description());
  top->addWidget(_description, 0, 1);
  label = new QLabel(i18n("Version"), cw);
  top->addWidget(label, 1, 0);
  _version = new QLineEdit(cw);
  _version->setText(project->version());
  top->addWidget(_version, 1, 1);
  label = new QLabel(i18n("Device"), cw);
  top->addWidget(label, 2, 0);
  _device = new DeviceChooser(project->device(), false, cw);
  top->addWidget(_device, 2, 1);

   // -------------------------- asm params  section ----------------------
   QGroupBox *asmbox = new QGroupBox(2, Horizontal, i18n("Assembler"), cw);
   top->addMultiCellWidget(asmbox, 3,3, 0,1);

    //   radix
    new QLabel(i18n("Radix"),asmbox) ;
    asm_radix = new QComboBox(asmbox) ;
    QToolTip::add(asm_radix,i18n("Choose the default radix used by gpasm to parse integer constants\n"
                          "* means that radix is specified by LIST directive."));
    asm_radix->insertItem("*") ;
    asm_radix->insertStringList(rad);
    QString r = project->radix() ;
    int k = rad.findIndex(r);
    asm_radix->setCurrentItem(k+1) ;


    //  Warning level
    new QLabel(i18n("Warning level"),asmbox) ;
    asm_warning = new QComboBox(asmbox) ;
    asm_warning->insertItem("*") ;
    asm_warning->insertStringList(warn);
    QString w=project->warnLevel() ;
    k=warn.findIndex(w) ;
    asm_warning->setCurrentItem(k+1) ;
    QToolTip::add(asm_warning,i18n("For cleaner code, always prefer higher warning levels.\n"
                          "* means that warning level is specified by LIST directive.")) ;

    //  Include Directories
    new QLabel(i18n("Include directory"),asmbox) ;
    QHBox *h1 = new QHBox(asmbox) ;
    asm_include = new QLineEdit(h1) ;
    QString i=project->includeDir() ;
    asm_include->setText(i) ;
    QToolTip::add(asm_include,i18n("By using the pseudovariable $(PRJPATH), you can specify an include directory\n"
                          "relative to the project directory."	)) ;
    QPushButton *incdir_button = new  QPushButton(iconset,"", h1) ;
    connect(incdir_button,SIGNAL(clicked()), this, SLOT( slotSelectIncDir())) ;


    //  other option
    new QLabel(i18n("Other options"),asmbox) ;
    asm_other = new QLineEdit(asmbox) ;
    QString o=project->otherOptions() ;
    asm_other->setText(o) ;
    QToolTip::add(asm_other,i18n("This item allows you to manually specify a gpasm \"exotic\" option.\n"
                          "The $(PRJPATH) pseudo variable if supported in this item as it is "
                          "in the \"Include Directories\" one.")) ;


    // ------------------------------- linker section -----------------
    QGroupBox *linkbox = new QGroupBox(2, Horizontal, i18n("Linker"), cw);
    top->addMultiCellWidget(linkbox, 4,4, 0,1);
        // hex file format
    new QLabel(i18n("Hex file format"),linkbox) ;
    link_fformat = new QComboBox(linkbox) ;
    link_fformat->insertStringList(hfmt);
    QString h=project->hexFormat() ;
    if( (k=hfmt.findIndex(h)) == -1) k = 0 ; ;
    link_fformat->setCurrentItem(k) ;
    QToolTip::add(link_fformat,i18n("CAUTION: inhx8s format is currently not supported by the PiKdev programmer.\n"
                          "If you need this support, please contact the author." )) ;

    // debug option
    new QLabel(i18n("Debug messages"),linkbox) ;
    link_debug = new QCheckBox(linkbox) ;
    link_debug->setChecked(project->linkDebug()) ;

    //  object/libs Directories
    new QLabel(i18n("Libraries directory"),linkbox) ;
    QHBox *hb2 = new QHBox(linkbox) ;
    link_include = new QLineEdit(hb2) ;
    link_include->setText(project->libsDir()) ;
    QToolTip::add(link_include,i18n("By using the pseudovariable $(PRJPATH), you can specify a directory\n"
                          "relative to the project directory:" )) ;
    QPushButton *libdir_button = new  QPushButton(iconset,"", hb2) ;
    connect(libdir_button,SIGNAL(clicked()), this, SLOT( slotSelectLibDir())) ;

    // other options
    new QLabel(i18n("Other options"),linkbox) ;
    link_other = new QLineEdit(linkbox) ;
    link_other->setText(project->otherLinkOptions()) ;
    QToolTip::add(link_other,i18n("This item allows you to manually specify a gplink \"exotic\" option.\n"
                          "The $(PRJPATH) pseudo variable if supported in this item as it is "
                          "in the \"Libraries Directory\" one.")) ;
}

void ProjectEditorWid::slotOk()
{
	project->setDescription(_description->text()) ;
	project->setVersion(_version->text()) ;

	project->setDevice(_device->device()) ;
	project->setWarnLevel(asm_warning->currentText()) ;
	project->setOtherOptions(asm_other->text()) ;
	project->setIncludeDir(asm_include->text()) ;
	project->setRadix(asm_radix->currentText()) ;

	project->setHexFormat(link_fformat->currentText()) ;
	project->setLinkDebug(link_debug->isOn()) ;
	project->setLibsDir(link_include->text()) ;
	project->setOtherLinkOptions(link_other->text()) ;
        accept();
}

void ProjectEditorWid::slotSelectIncDir()
{
  PDirectory dir = PDirectory::getExisting(project->directory(), this, i18n("Select include directory"));
  if ( !dir.isEmpty() ) asm_include->setText(dir.path());
}

void ProjectEditorWid::slotSelectLibDir()
{
  PDirectory dir = PDirectory::getExisting(project->directory(), this, i18n("Select libraries directory"));
  if ( !dir.isEmpty() ) link_include->setText(dir.path());
}
