/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *   Copyright (C) 2003 Alain Gibaud <alain.gibaud@free.fr>                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "ProjectWid.h"

#include <qhbox.h>
#include <qpushbutton.h>
#include <qlabel.h>
#include <qapplication.h>
#include <qdragobject.h>

#include <klocale.h>
#include <kiconloader.h>
#include <kaction.h>
#include <kmessagebox.h>
#include <kpopupmenu.h>

#include "NewProject.h"
#include "ProjectEditorWid.h"
#include "FNU.h"
#include "PikMain.h"

//----------------------------------------------------------------------------
const char *ListViewItem::FILE_GROUP_LABELS[Nb_FileGroups] = {
  I18N_NOOP("Sources"), I18N_NOOP("Includes"), I18N_NOOP("Objects"),
  I18N_NOOP("Generated")
};

ListViewItem::ListViewItem(QListView *listview)
  : KListViewItem(listview), _group(Nb_FileGroups)
{
  init(PURL::Project);
  set(PURL());
}

ListViewItem::ListViewItem(ListViewItem *item, FileGroup group)
  : KListViewItem(item), _group(group)
{
  setText(0, i18n(FILE_GROUP_LABELS[group]));
}

ListViewItem::ListViewItem(ListViewItem *item, const PURL &url)
  : KListViewItem(item), _group(Nb_FileGroups)
{
  init(url.fileType());
  set(url);
}

ListViewItem::FileGroup ListViewItem::group(const PURL &url)
{
  switch (url.fileType()) {
    case PURL::Asm:
    case PURL::CSource: return Sources;
    case PURL::Include: return Includes;
    case PURL::Library:
    case PURL::Object: return Objects;
    case PURL::Hex:
    case PURL::Map:
    case PURL::Cod:
    case PURL::Coff:
    case PURL::Lst: return Generated;
    case PURL::Project:
    case PURL::Unknown:
    case PURL::Nb_FileTypes: break;
  }
  Q_ASSERT(false);
  return Nb_FileGroups;
}

void ListViewItem::init(PURL::FileType type)
{
  _type = type;
  setPixmap(0, QPixmap(PURL::DATA[_type].xpm_icon));
}

void ListViewItem::set(const PURL &url)
{
  _url = url;
  if ( _type==PURL::Project ) {
    if ( url.isEmpty() ) setText(0, i18n("<no project>"));
    else setText(0, url.basename());
  } else setText(0, url.filename());
}

//----------------------------------------------------------------------------
ProjectWid::ProjectWid(PikMain *mainwin, QWidget *parent)
  : QVBox(parent, "project_manager"), _project(0), _main(mainwin)
{
  _view = new KListView(this);
  _view->addColumn(i18n("Project Files"), 170);
  _view->setFullWidth(true);
  _view->setRootIsDecorated(false);

  _root  = new ListViewItem(_view);

  connect(_view, SIGNAL(rightButtonPressed(QListViewItem *, const QPoint &, int )),
          SLOT(slotProjectRButton(QListViewItem *, const QPoint &, int)));
  connect(_view, SIGNAL(clicked(QListViewItem *)),
          SLOT(slotProjectElementClicked(QListViewItem *)));
  setAcceptDrops(true);
}

void  ProjectWid::slotProjectRButton(QListViewItem *item, const QPoint &p, int)
{
  if ( item==0 ) return;
  ListViewItem *li = static_cast<ListViewItem *>(item);
  if ( li->group()!=ListViewItem::Nb_FileGroups ) return;

  KIconLoader loader ;
  QIconSet iconnewfile = loader.loadIconSet("filenew", KIcon::Small) ;
  QIconSet iconopenproject = loader.loadIconSet("piclab_openproject", KIcon::Small) ;
  QIconSet iconnewproject = loader.loadIconSet("piclab_createproject", KIcon::Small) ;
  QIconSet iconeditproject = loader.loadIconSet("piclab_editproject", KIcon::Small) ;
  QIconSet iconcloseproject = loader.loadIconSet("piclab_closeproject", KIcon::Small) ;
  QIconSet iconaddfile = loader.loadIconSet("piclab_addfile", KIcon::Small) ;
  QIconSet iconaddcurrentfile = loader.loadIconSet("piclab_addcurrentfile", KIcon::Small) ;

  KPopupMenu *pop = new KPopupMenu;
  if ( li==_root ) {
    pop->insertTitle(i18n("Project"));
    if ( !li->url().isEmpty() ) { // "open" project entry
      pop->insertItem(iconnewfile, i18n("New File..."), 5);
      pop->insertItem(iconaddfile, i18n("Add File..."), 1);
      if ( _main->editorManager()->currentEditor() )
        pop->insertItem(iconaddcurrentfile,  i18n("Add Current File"), 2);
      pop->insertItem(iconeditproject, i18n("Project Options..."), 3);
      pop->insertItem(iconcloseproject, i18n("Close Project"), 4);
      switch( pop->exec(p) ) {
        case 1: slotInsertFile(); break;
        case 2: slotInsertCurrentFile(); break;
        case 3: slotEditProject(); break;
        case 4: slotCloseProject(); break;
        case 5: _main->editorManager()->newFile(); break;
      }
    } else { // empty project entry
      pop->insertItem(iconnewproject, i18n("New Project..."), 1);
      pop->insertItem(iconopenproject, i18n("Open Project..."), 2);
      switch ( pop->exec(p)) {
        case 1: slotNewProject(); break;
        case 2: slotOpenProject(); break;
      }
    }
  } else {
    pop->insertTitle(li->text(0));
    pop->insertItem( i18n("Remove From Project"), 1);
    if ( pop->exec(p)==1 ) removeFile(li->url());
  }
  delete pop;
}

void ProjectWid::slotCloseProject()
{
  if ( _project==0 ) return;
  _root->set(PURL());
  ListViewItem *it = static_cast<ListViewItem *>(_root->firstChild());
  while (it) {
    if ( _main->editorManager()->switchToEditor(it->url()) )
      _main->editorManager()->closeCurrentEditor();
    ListViewItem *it1 = it ;
    it = static_cast<ListViewItem *>(it->nextSibling());
    delete it1 ;
  }
  delete _project;
  _project = 0;
  emit guiChanged();
}

void ProjectWid::slotInsertFile()
{
    if ( _project==0 ) return; // security
    QString filter = PURL::filter(PURL::Source | PURL::Linkable);
    PURL url = PURL::getOpen(_project->directory(), filter, this, i18n("Select file"));
    if ( url.isEmpty() ) return;
    insertFile(url);
}

void ProjectWid::insertFile(const PURL &url)
{
  PURL purl = url;
  if ( !url.isInto(_project->directory()) ) {
    int res = KMessageBox::warningYesNo(this, i18n("File \"%1\" is not inside the project directory.").arg(url.filepath()),
                                        i18n("Add File to Project"), i18n("Copy"), KStdGuiItem::cancel());
    if ( res==KMessageBox::Cancel ) return;
    PURL dest(_project->directory(), url.filename());
    if ( dest.exists(true, this) ) {
      KMessageBox::sorry(this, i18n("Destination file \"%1\" already exists.").arg(dest.filepath()));
      return;
    }
    if ( !url.copyTo(dest, this) ) {
      KMessageBox::sorry(this, i18n("Copying file \"%1\" to project directory failed.").arg(url.filepath()));
      return;
    }
    purl = dest;
  }
  if ( _project->absoluteFiles().contains(purl) ) {
    KMessageBox::sorry(this, i18n("File \"%1\" is already in the project.").arg(url.filepath()));
    return;
  }
  _project->addFile(purl);
  addFile(purl);
}

void ProjectWid::slotOpenProject()
{
  if (_project) slotCloseProject();
  QString filter = PURL::filter(PURL::Project);
  PURL url = PURL::getOpen(PDirectory::home(), filter, this, i18n("Select Project file"));
  openProjectFile(url.kurl());
}

ListViewItem *ProjectWid::findGroupItem(ListViewItem::FileGroup group)
{
  ListViewItem *item = static_cast<ListViewItem *>(_root->firstChild());
  while (item) {
    if ( item->group()==group ) return item;
    item = static_cast<ListViewItem *>(item->nextSibling());
  }
  return 0;
}

void ProjectWid::addFile(const PURL &url)
{
  qDebug("Project::addFile %s", url.filepath().latin1());
  if ( findItem(url) ) return;
  ListViewItem::FileGroup group = ListViewItem::group(url);
  ListViewItem *parent = findGroupItem(group);
  if ( parent==0 ) {
    parent = new ListViewItem(_root, group);
    _root->setOpen(true);
  }
  (void)new ListViewItem(parent, url);
  parent->setOpen(true);
  emit guiChanged();
}

void ProjectWid::removeFile(const PURL &url)
{
  _project->removeFile(url);
  ListViewItem *item = findItem(url);
  if ( item==0 ) return;
  ListViewItem *group = static_cast<ListViewItem *>(item->parent());
  delete item;
  if ( group->childCount()==0 ) delete group;
  emit guiChanged();
}

void ProjectWid::slotProjectElementClicked(QListViewItem *item )
{
  if ( item==0 ) return;
  ListViewItem *li = static_cast<ListViewItem *>(item);
  if ( li->group()!=ListViewItem::Nb_FileGroups ) return;
  if ( !(PURL::DATA[li->type()].properties & PURL::Editable) ) return;
  if ( _main->editorManager() && _main->editorManager()->switchToEditor(li->url())==0 )
    KMessageBox::sorry(this, i18n("Cannot find file \"%1\".").arg(li->url().filepath()));
}

void ProjectWid::slotInsertCurrentFile()
{
  PURL url = _main->editorManager()->currentFile();
  if ( url.data().properties & PURL::Source ) insertFile(url);
  else QApplication::beep(); // #### FIXME: message box
}

void ProjectWid::slotEditProject()
{
  ProjectEditorWid dialog(_project, this);
  dialog.exec();
  emit guiChanged();
}

void ProjectWid::slotNewProject()
{
  if ( _project==0 ) slotCloseProject();
  NewProjectDialog dialog(this);
  if ( dialog.exec()==QDialog::Rejected ) return;

  // #### should ask for overwrite ?
  if ( dialog.url().exists(false, this) ) {
    KMessageBox::sorry(this, i18n("The project file \"%1\" already exists" ).arg(dialog.url().filepath()));
    return;
  }
  // create dir if it does not exist
  PDirectory dir = dialog.url().directory();
  if ( !dir.exists(this) && !dir.create(this) ) {
    KMessageBox::sorry(this, i18n("Cannot create directory \"%1\"" ).arg(dir.path()));
    return;
  }

  slotCloseProject();
  _project = new Project(dialog.url());
  _project->setDescription(QString::null);
  _project->setVersion("0.1");
  _project->setOutputFile(dialog.url().toFileType(PURL::Hex));
  _project->setDevice(dialog.device());
  _root->set(dialog.url());
  emit guiChanged();
}

void ProjectWid::openProjectFile(const PURL &url)
{
  if ( url.isEmpty() ) return;
  if ( !url.exists(true, this) ) {
    KMessageBox::sorry(this, i18n("Cannot open project file \"%1\".").arg(url.filepath()));
    return;
  }
  if (_project) slotCloseProject();
  _project = new Project(url);
  _root->set(url);
  PURL::List files = _project->absoluteFiles();
  PURL::List::const_iterator it = files.begin();
  for(; it!=files.end(); ++it) addFile(*it);
  _root->setOpen(true);
  emit guiChanged();
}

void ProjectWid::dragEnterEvent(QDragEnterEvent *ev)
{
    // accept if dragged object can be decoded as an URI
    ev->accept( QUriDrag::canDecode(ev) ) ;
}

void ProjectWid::dropEvent(QDropEvent *ev)
{
  QStringList files;
  if ( !QUriDrag::decodeLocalFiles(ev, files) ) return;
  PURL url(*files.begin());
  // only one file name, and it is a project file and no project already open
  if ( _project==0 && files.count()==1 && url.fileType()==PURL::Project ) {
    openProjectFile(url);
    return;
  }
  if ( _project && files.count()==1 && url.fileType()==PURL::Project ) {
    openProjectFile(url);
    return;
  }
  if ( _project==0 ) return;
  // try to include files in the current project
  for (QStringList::Iterator it=files.begin(); it!=files.end(); ++it) {
    PURL url(*it);
    PURL::FileType type = url.fileType();
    if ( type==PURL::Asm || type==PURL::CSource || type==PURL::Library
         ||  type==PURL::Object || type==PURL::Include )
      insertFile(url); // multiple insertions are checked by inserAbsFile()
  }
}

ListViewItem *ProjectWid::findItem(const PURL &url) const
{
  QListViewItemIterator it(_view);
  while ( it.current() ) {
    ListViewItem *item = static_cast<ListViewItem *>(it.current());
    if ( item->url()==url ) return item;
    ++it;
  }
  return 0;
}

void ProjectWid::select(const PURL &url)
{
  ListViewItem *item = findItem(url);
  if (item) _view->setSelected(item, true);
  else _view->clearSelection();
}
