/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "device_list.h"

#include <qregexp.h>

#include <klocale.h>

#include "devices/devices.h"
#include "direct_prog/PicFactory.hh"
#include "FNU.h"

//----------------------------------------------------------------------------
const DeviceUser::Data DeviceUser::DATA[Nb_Types] = {
 { I18N_NOOP("gpasm"),              0          },
 { I18N_NOOP("gpdasm"),             0          },
 { I18N_NOOP("Direct Programmers"), Programmer },
 { I18N_NOOP("Picstart+"),          0          }, // #### Programmer not implemented
 { I18N_NOOP("Promate2"),           0          }, // #### Programmer not implemented
 { I18N_NOOP("ICD"),                0          }, // #### Programmer not implemented
 { I18N_NOOP("ICD2 Programmer"),    Programmer },
 { I18N_NOOP("ICD2 Debugger"),      Debugger   }
};
int DeviceUser::TYPE_MAPPING[Nb_Types] = {
 0, 0, 0, SUPPORT_PICSTART_PLUS, SUPPORT_PROMATE2, SUPPORT_ICD,
 SUPPORT_ICD2P, SUPPORT_ICD2D
};

DeviceUser::Type DeviceUser::defaultType(Property property)
{
  for (uint i=0; i<Nb_Types; i++)
    if ( DATA[i].properties & property ) return Type(i);
  return Nb_Types;
}

//----------------------------------------------------------------------------
QMap<QString, DeviceList::Data> *DeviceList::_map = 0;

QStringList DeviceList::names(DeviceUser::Type type)
{
  QStringList list;
  QMap<QString, Data>::const_iterator it = map().begin();
  for (; it!=map().end(); ++it)
    if ( type==DeviceUser::AllTypes || it.data().supported[type] )
      list += it.key();
  return list;
}

QString DeviceList::defaultDevice(DeviceUser::Type type)
{
  QStringList list = names(type);
  if ( list.contains("16F877") ) return "16F877";
  return list[0];
}

void DeviceList::addGPAsmDevices()
{
  FILE *f = popen("gpasm -l", "r");
  if ( f==0 ) return;
  char name[60];
  while( fscanf(f, "%59s", name)!=EOF ) {
    if ( name[0]=='p' ) { // pic
      QString s(name);
      s = s.mid(1).upper();
      (*_map)[s].supported[DeviceUser::GPAsm] = true;
    }
  }
  pclose(f);
}

void DeviceList::addGPDAsmDevices()
{
  FILE *f = popen("gpdasm -l", "r");
  if ( f==0 ) return;
  char name[60];
  while( fscanf(f, "%59s", name)!=EOF ) {
    if ( name[0]=='p' ) { // pic
      QString s(name);
      s = s.mid(1).upper();
      (*_map)[s].supported[DeviceUser::GPDAsm] = true;
    }
  }
  pclose(f);
}

void DeviceList::addDirectProgDevices()
{
  QStringList list = QStringList::split(' ', PicFactory::factory()->getPicNames());
  QStringList::const_iterator it = list.begin();
  for (; it!=list.end(); ++it) {
    QString s = (*it).mid(1).upper();
    (*_map)[s].supported[DeviceUser::DirectProgrammer] = true;
  }
}

void DeviceList::addMicrochipDevices()
{
  for (uint i=0; PIC_DEVICES[i]; i++) {
    QString name = QString(PIC_DEVICES[i]->name).upper();
    if ( name.startsWith("PIC") ) name = name.mid(3);
    if ( name.startsWith("DSPIC") ) name = name.mid(5);
    for (uint k=0; k<DeviceUser::Nb_Types; k++) {
      if ( !(PIC_DEVICES[i]->support & DeviceUser::TYPE_MAPPING[k]) ) continue;
      (*_map)[name].supported[DeviceUser::Type(k)] = true;
      (*_map)[name].device = PIC_DEVICES[i];
    }
  }
}

QMap<QString, DeviceList::Data> &DeviceList::map()
{
  if ( _map==0 ) {
    _map = new QMap<QString, Data>;
    addGPAsmDevices();
    addGPDAsmDevices();
    addDirectProgDevices();
    addMicrochipDevices();
  }
  return *_map;
}

const DeviceList::Data *DeviceList::data(const QString &name)
{
  if ( !map().contains(name) ) return 0;
  return &map()[name];
}

QString DeviceList::checkName(const QString &name)
{
  if ( name=="*" ) return name;
  QStringList list = names();
  if ( list.contains(name) ) return name;
  if ( name.startsWith("p") ) // compat mode for PiKdev
    return checkName(name.mid(1));
  return list[0]; // default...
}

QString DeviceList::guessDeviceFromFile(const PURL &url, bool &success, QWidget *widget)
{
  success = false;
  PURL src = url;
  if ( url.fileType()==PURL::Hex ) {
    src = url.toFileType(PURL::Asm);
    if ( !src.exists(true, widget) ) {
      src = url.toFileType(PURL::Asm, true, true);
      if ( !src.exists(true, widget) ) return defaultDevice();
    }
  }

  PFile file(src);
  if ( !file.openForReading(widget) ) return defaultDevice();

  QString device, line;
  //   QRegExp re1("^[ \\t]+(?:PROCESSOR|processor)[ \\t]+((?:p|sx|P|SX)[a-z0-9A-Z]+)" ) ;
  QRegExp re1("^[ \\t]+(?:PROCESSOR|processor)[ \\t]+([a-z0-9A-Z]+)" ) ;
  QRegExp re2("^[ \\t]+(?:LIST|list)[ \\t]+" ) ;
  while( file.readLine(line, 10000)!=-1 ) {
    // search PROCESSOR directive
    if ( re1.search(line, 0)!=-1 ) {
      device = re1.cap(1);
      success = true;
      break;
    }
    // search LIST p=... directive
    int k = re2.search(line,0);
    if ( k!=-1 ) {
      //QRegExp re3("(?:p|P)[ \\t]*=[ \\t]*((?:p|sx|P|SX)[a-z0-9A-Z]+)") ;
      QRegExp re3("(?:p|P)[ \\t]*=[ \\t]*([a-z0-9A-Z]+)") ;
      if ( re3.search(line, k+5)!=-1 ) {
        device = re3.cap(1);
        success = true;
        break;
      }
    }
  }
  device = device.upper() ;
  if( device[0]=='P') return device.mid(1);
  return device;
}

bool DeviceList::isSupported(const QString &name, DeviceUser::Type type)
{
  const Data *d = data(name);
  if ( d==0 ) return false;
  if ( type!=DeviceUser::AllTypes ) return d->supported[type];
  for (uint i=0; i<DeviceUser::Nb_Types; i++) if ( d->supported[i] ) return true;
  return false;
}

QString DeviceList::toGPUtilsName(const QString &name)
{
  return "p" + name.lower();
}
