/* --------------------------------------------------------------------- *
   Linux PICxxxxxx programming software

   Supports Russ Reiss's PICSPA84 PIC16C84 programmer board
   and David Tait's PD PIC16C84 programmer schematic and probably any
   custom parallel port programmer.

   The software has been inspirated  by Brian C. Lane's C code
   but has bee re-written from scratch by Alain Gibaud in C++

   --------------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <iostream>

#include "pic.hh"
#include "PicFactory.hh"

//
//  pic class implementation by A Gibaud,
//  This code has been written from Brian C. Lane work
//  but is 99% rewritten
//
pic::pic()
{
    _data = 0;
    debugmode = 0 ;
    hard_ = 0 ;
    memaddr1=memaddr2=
    idaddr1=idaddr2=
    eeaddr1=eeaddr2=
    cfaddr1=cfaddr2=
    backup_osccal1 = backup_osccal2 = -1 ;
    clk_delay = 0 ;
    programming_option = -1 ;
}
pic::~pic()
{
    // nothing to do
}
void pic::setPort(hardware *port)
{
    hard_ = port ;
}
void pic::debugon(int mask)
{
    debugmode |= mask ;
}

void pic::debugoff(int mask)
{
    debugmode &= ~mask ;
}

// ================================================= End of I/O routines
// -----------------------------------------------------------------------
// A standard word mask for 14 bits
// Must be reimplemented for different word length
// -----------------------------------------------------------------------
int pic::wmask()
{
    return 0x3FFF ;
}

// Programming routines =====================================================


void pic::send_cmd( unsigned int d )
{
    send_bits(d, 6) ;
}

void pic::send_bits( unsigned int d , int nbb)
{
    int x;

    //++Mirko!!
    //Keep the write state...
    hard_->rwwrite();
    //--Mirko!!

    for( x = nbb ;  x  ; --x )
    {
        hard_->clkhi();				/* Clock high */

        if ( d & 0x01 )
            hard_->outhi();				/* Output a 1 */
        else
            hard_->outlo();				/* Output a 0 */

        waitus(1+clk_delay) ;
        hard_->clklo();				/* Clock low */
        waitus(1+clk_delay) ;
        d >>= 1;
    }
    hard_->outhi();

    //++Mirko!!
    //Normal (safe) state...
    hard_->rwread();
    //--Mirko!!

}




void pic::send_data( unsigned int d )
{
    int x;

    //++Mirko!!
    //Keep the write state...
    hard_->rwwrite();
    //--Mirko!!

    d &= wmask() ;		// mask unused msb
    d <<= 1;                // insert start bit

    for( x = 0; x < 16; x++ )
    {
        hard_->clkhi();
        if ( d & 0x01 )
            hard_->outhi();                		/* Data bit is a one */
        else
            hard_->outlo();                		/* Data bit is a zero */
        waitus(1+clk_delay) ;  // needed for slow programmers/fast PCs
        hard_->clklo();
        waitus(1+clk_delay) ;
        d >>= 1;                     /* Move the data over 1 bit */
    }
    hard_->outhi(); 			/* Added for ITU-1 support */

    //++Mirko!!
    //Normal (safe) state...
    hard_->rwread();
    //--Mirko!!
}



/* -----------------------------------------------------------------------
   Read 14 bits of data from the PIC
   clock idles low, change data. 1 start bit, 1 stop bit, data valid on
   falling edge.
   ----------------------------------------------------------------------- */
/** Get a word of data from device
May be reimplemented
(the default implementation is correct for (most ?) pic12 and pic16 devices)
*/
unsigned int pic::get_word()
{
    unsigned int x, ind;

    //++Mirko!!
    //Normal (safe) state...
    hard_->rwread();
    //--Mirko!!

    ind = 0;

    hard_->outhi();
    for( x = 16; x ; --x )
    {
        hard_->clkhi();
        waitus(1+clk_delay) ;
        if ( hard_->inbit() )
            ind |= 0x8000;				/* Bit is 1	*/
        else
            ind &= 0x7FFF;				/* Bit is 0	*/
        ind >>= 1 ;
        hard_->clklo();
        waitus(1+clk_delay) ;
    }
    return  ind & wmask() ;
}

/** Read a byte of data from the device
May be reimplemented
The default implementation just call get_word() and is correct
for pic12 and pic16 devices
  */
unsigned int pic::get_byte()
{
    return get_word() & 0xFF ;
}



/* ------------------------------------------------------------------------
   Delay for a number of uS -- Works better than usleep for me!
   ++Gib: maybe a better implementation might use nanosleep()
   (but nanosleep() is linux-specific and the present implementation
   seems to work fine ... )
   ------------------------------------------------------------------------ */
void pic::waitus( unsigned int us )
{
    struct timeval tv, wt;

    gettimeofday( &tv, NULL );

    wt.tv_usec = ( tv.tv_usec + us ) % 1000000;
    wt.tv_sec = tv.tv_sec + ( ( tv.tv_usec + us ) / 1000000);

    /* loop until time is up */
    for (;;)
    {
        gettimeofday( &tv, NULL );
        if ( tv.tv_sec > wt.tv_sec )
            return;
        if ( tv.tv_sec == wt.tv_sec )
            if ( tv.tv_usec > wt.tv_usec )
                return;
    }
}

// -----------------------------------------------------------------
// General command sequence interpreter
// The commands to be executed are located in "cmd"
// the "word" parameter is unusued by many commands, but may be
// used as input (by S and s and V and v) or as output (by R and r)
// Returns:
//	number of  errors during verification process
// ------------------------------------------------------------------
/*
 Here are the coding conventions for sending commands
 to the pic

----------- electric signals level commands
UPPER case means signal High - lower case means Low
  Example: "cPCcp" means "clock-low, , power ON, clock high,
  clock low, power OFF".

c/C = clock
d/D = data
p/P = power (vdd)
b/B = burn (vpp)

----------- higher level commands
wnnnnn = wait nnnnn microseconds (nnnnn is decimal)
S  = send data word
R = read data word
s = send byte  (not impemented)
r = read byte
F = fake word read (read from pic, but don't send back via *word,
    keep the value for subsequent Verify)
f = dummy byte read (read from pic, but don't send back via *word,
    keep the value for subsequent verify)
knn = send 6 bits command nn to PIC (nn is decimal)
v = verify byte data (compare previously readed byte  against *word & 0xFF)
V = verify word data (compare previously readed word  against *word)
; = NOP (can be used as separator - usefull in debug mode)
? = return previously readed word
			WARNING ! No blanks allowed

------------ COMMANDS SPECIFIC TO 18Fxxx SUPPORT :
Knn = send 4 bits command nn to PIC (nn is decimal)
xhh = send a 8 bits constant (hh must be a 2 digits hex constant)
Xhhhh =  send a 16 bits constant (hhhh must be a 4 digits hex constant)
, = a shorthand for w1

*/

int pic::pulseEngine(const char *cmd, unsigned int *word)
{
    static unsigned int rword = INVALID ;
    unsigned int  n ;
    int errors , cmdcod  ;
    char buf[100] ;


    for( errors = 0 ;  *cmd ; ++cmd )
    {
        cmdcod = *cmd ;
        switch( cmdcod  )
        {
        default:
            printf("\npic::pulseEngine : invalid command [ %c ]\n", cmdcod);
            exit(0) ;
            break ;
        case ';' : // NOP
            if(debugmode & 1) cout << endl ;
            break ;
        case 'c' :
            if(debugmode & 1)
                cout << "CLOCK L" << endl ;
            else
                hard_->clklo() ;
            break ;
        case 'C' :
            if(debugmode & 1)
                cout << "CLOCK H" << endl ;
            else
                hard_->clkhi() ;
            break ;

        case 'd' :
            if(debugmode & 1)
                cout << "DATA L" << endl ;
            else
                hard_->outlo() ;
            break ;
        case 'D' :
            if(debugmode & 1)
                cout << "DATA H" << endl ;
            else
                hard_->outhi() ;
            break ;

        case 'p' :
            if(debugmode & 1)
                cout << "VDD Off" << endl ;
            else
            {
                hard_->vddoff() ;
                waitus(10000) ;
            }
            break ;
        case 'P' :
            if(debugmode & 1)
                cout << "VDD On" << endl ;
            else
            {
                hard_->vddon() ;
                // wait 10ms for ALL devices
                // because some programmers (or target board)
                // have capacitors to charge
                waitus(10000) ;
            }
            break ;

        case 'b' :
            if(debugmode & 1)
                cout << "VPP Off" << endl ;
            else
                hard_->vppoff() ;
            break ;
        case 'B' :
            if(debugmode & 1)
                cout << "VPP On" << endl ;
            else
            {
                hard_->vppon() ;
                waitus(100000) ;
            }
            break ;

        case 'w' :
            for(n=0 , ++cmd ; *cmd && isdigit((int)*cmd) ; ++cmd )
                n = (n * 10) + *cmd - '0' ;
            --cmd ;
            if(debugmode & 1)
                cout << "WAIT " << n << " micro-sec." << endl ;
            else
                hard_->waitus(n) ;
            break ;
        case 'S' :
            if(debugmode & 1)
            {
                sprintf(buf, "SEND 0x%04X" , *word ) ;
                cout << buf << endl ;
            }
            else
                send_data(*word);
            break ;
        case 'R' :
            if(debugmode & 1)
                cout << "Read Word" << endl ;
            else
                *word = get_word()  ;
            break ;
        case 'r' :
            if(debugmode & 1)
                cout << "Read Byte" << endl ;
            else
                *word = get_byte()  ;
            break ;

        case 'F' :
            if(debugmode & 1)
                cout << "fake read Word" << endl ;
            else
                rword = get_word()  ;
            break ;
        case 'f' :
            if(debugmode & 1)
                cout << "fake read Byte" << endl ;
            else
                rword = get_byte() ;
            break ;

        case 'k' :
            for(n=0 , ++cmd ; *cmd && isdigit((int)*cmd) ; ++cmd )
                n = (n * 10) + *cmd - '0' ;
            --cmd ;
            if(debugmode & 1)
                cout << "SEND cmd " << n << endl ;
            else
                send_cmd(n ) ;
            break ;
        case 'K' :
            assert("pulseEngine : K command removed, please use k instead" == 0) ;
            break ;
        case 'V':
            if(debugmode & 1)
                cout << "COMPARE Word " << endl ;
            else
            {
                if(rword != (*word & wmask())) ++errors ;
            }
            break ;

        case 'v':
            if(debugmode & 1)
                cout << "COMPARE Byte " << endl ;
            else
            {
                if(rword != (*word & 0xFF)) ++errors ;
            }
            break ;
        case 'x':
            sscanf(++cmd,"%02X",&n) ;
            ++cmd ;

            if(debugmode & 1)
            {
                sprintf( buf,  "SEND 1 byte constant : 0x%02X",n) ;
                cout << buf << endl ;
            }
            else
            {
                send_bits(n,8) ;
            }
            break ;
        case 'X':
            sscanf(++cmd,"%04X",&n) ;
            cmd+=3 ;

            if(debugmode & 1)
            {
                sprintf( buf,  "SEND 2 bytes constant : 0x%04X",n) ;
                cout << buf << endl ;
            }
            else
            {
                send_bits(n,16) ;
            }
            break ;
        case ',':

            if(debugmode & 1)
            {
                cout << "WAIT 1 micro-sec" << endl ;
            }
            else
            {
                waitus(1)  ;
            }
            break ;
        case '?':

            if(debugmode & 1)
            {
                cout << "Fetch FAKE cache : "<<  hex << rword << dec << endl ;
            }
            else
            {
                *word = rword & wmask()  ;
            }
            break ;
        }
    }
    return errors ;
}

void pic::set_memory(unsigned int adr, unsigned int val)
{
    mem()[adr] = val ;
}


/* -------------------------------------------------------------------------
   displays contents of the mem buffer
   invalid (ie: 0xFFFF words) are not displayed
   and acts as blocks separators

   Returns:
   the number of detected blocks
   ------------------------------------------------------------------------ */
/*
int pic::display_mem(ostream &out)
{
    int start = 0 , len , nblocks ;

    FragBuffer::iterator block = mem.begin() ;
    string adrtype = byteAdresses() ? " (bytes adresses) " : " (words adresses) "  ;

    for( start = nblocks = 0 ; fetch_next_block(block,&len) ; ++nblocks  )
    {
        // block found, display it
        out << "======= Block #" << hex << nblocks << adrtype << endl ;
        display_block(out, len, block );
    }
    return nblocks ;
}
*/
/*
	Display a block of memory buffer on "out" stream
  iterator block is updated by the function

 */
// number of values per line
#define ASCIIBLKSIZE 8

void pic::display_block(ostream &out,
                        int reclen, 	/* length (in words) */
                        FragBuffer::iterator& block) 	/* iterator to block  */

{
    char buf[20] ;

    while(reclen > ASCIIBLKSIZE)
    {
        display_block(out, ASCIIBLKSIZE, block ) ;
        reclen -= ASCIIBLKSIZE ;
    }

    if(reclen <= 0) return ; /* Oops, block has just a ASCIIBLKSIZE * n size */


    // adress of first word of the line
    // form is not ANSI conformant, so the following line is not portable
    // out.form("%04X :", loc );
    // uppercase is not always implemented, so the following line is not portable
    // out << hex << setw(4) << setfill('0') << uppercase << loc << " :" ;
    // so I use an old-fashoned sprintf

    unsigned int adr =     (*block).first  ;
    if( byteAdresses()) adr *= 2 ;
    sprintf(buf, "%08X :", adr  ); out << buf ;
    for( ; reclen > 0; ++block , --reclen)
    {
        sprintf(buf," %04X" , (*block).second ); out << buf ;
    }
    out << endl ;
}
/* -----------------------------------------------------------
	Return an english message corresponding to the error code "code"
*/

const char *pic::error_msg(int code)
{
    static char m[500] ;

    *m = 0 ;
    if( code == 0)
    {
        strcpy(m,"No error");
    }
    else
    {
        if( code & 1)
        {
            if( *m ) strcat(m, "+") ;
            strcat(m,"prog. memory error") ;
        }

        if( code & 2)
        {
            if( *m ) strcat(m, "+") ;
            strcat(m,"EEPROM data error") ;
        }

        if( code & 4)
        {
            if( *m ) strcat(m, "+") ;
            strcat(m,"user ID error") ;
        }

        if( code & 8)
        {
            if( *m ) strcat(m, "+") ;
            strcat(m,"config fuses error") ;
        }

        if( code & 16)
        {
            if( *m ) strcat(m, "+") ;
            strcat(m,"non applicable op. error") ;
        }

        if( code & 32)
        {
            if( *m ) strcat(m, "+") ;
            strcat(m,"information") ;
        }

        if( code & 64)
        {
            if( *m ) strcat(m, "+") ;
            strcat(m,"calibration") ;
        }
    }

    return m ;
}


void pic::getBuffers(int& mem1, int& mem2,
                     int& ee1,  int& ee2,
                     int& id1,  int& id2,
                     int& cf1, int& cf2,
                     int& oscbackup1, int& oscbackup2 )
{
    mem1 = memaddr1 ; mem2 = memaddr2 ;
    ee1 = eeaddr1 ; ee2 = eeaddr2 ;
    id1 = idaddr1 ; id2 = idaddr2 ;
    cf1 = cfaddr1 ; cf2 = cfaddr2 ;
    oscbackup1 =  backup_osccal1 ;
    oscbackup2 =  backup_osccal2 ;
}

//
// Non virtual interfaces to program/verify routines
// Ouptut stream is forced to cout
//
int pic::program_pic()
{
    return program_pic(cout) ;
}
int pic::verify_pic()
{
    return verify_pic(cout) ;
}
//
// must be used by program/verify algorithms to output diags
//
void pic::warning(ostream& out, const char *what,
                  int where,  int chipvalue,  int memvalue )
{
    out << "Warning: [" << what << "] (addr = 0x" << hex << where ;
    if( chipvalue != -1) out << ", device = 0x" << chipvalue  ;
    if( memvalue != -1) out << ", buffer = 0x" << memvalue  ;
    out << ")" << endl << dec << flush;

}


/** Deduces the device family from the pic names

\return 12, 16 , 17 or 18 or 0 if picname is not recognized
 */
int  pic::picFamily()
{
    const char *n = names() ;
    while(*n == ' ') ++n ;
    int f = 0 ;
    sscanf(n,"p%d",&f) ;
    return f ;
}

/** No descriptions */
bool pic::byteAdresses() const
{
    return false ;
}
/** No descriptions */
std::string  pic::deviceID2Name(unsigned int id)
{
    PicFactory *fac = PicFactory::factory() ;
    // AFAIK, all PICs have a 5 bits revision number
    return fac->findDeviceName(id & ~0x1F )  ;
}
unsigned int pic::deviceID2Revision(unsigned int id)
{
    return id & 0x1F ;
}
/** Device autodetection
\return The raw type ID of the device (including revision number) */
unsigned int pic::get_deviceID()
{
    return 0 ; // no autodetection by default
}
/** Device autodetection support
\return
A blanks separated list of pic ID. this list MUST correspond item by item
to the names list returned by names().
Each ID of the list is the device ID word in which the revision number is masked.
It must be specified in hex.
If none of the devices returned by names() is autodetectable, IDs() can
return an empty string. If some devices are not identifiable,their  ID must
be 0.

The default implementation is to return an empty string, so old non
autodetectable devices do not need this method to be implemented. */
const char * pic::IDs()
{
    return "" ;
}


  /** Special locations descriptor
    This method returns a list of special memory locations (in device's memory space)
    The main purpose of this method is to allow the programmer widget to highlight
    some special words.
    Be careful : as special word needs color changes and tooptips allocations/deallocation,
    using too many special words can slow down the programming widget.

    General syntax :
    adresses_range=color{help_text} ...

    adresses_range:
    [low_addr:hi_addr] | unique_addr
    ( unique_addr is a shorthand for [unique_addr:unique_addr], all addresses
    are hexadecimal values )

    color:
    #RRGGBB
    (A HTML-like RGB value)

    help_text:
    Arbitrary long text, without } character. \n are allowed.

    =color field can be ommited. In this case, the previous color is used.
    If no color is provided, a default color is used.

    {help_text} can be ommited. In this case, the location does not have any
    attached text.

    The method can return several adress specifications, separated by blanks

    example:
    return "[0-3]=#FF0000{first four words} 3FF=#AAAAAA{OSCCAL word}" ;

    \return A blank separated list of locations descriptor.
    Must return an empty string (ie:"") if not used.
    Never return a NULL pointer !
   */


const char* pic::SPLocs()
{
  return "" ;
}
/** No descriptions */
const char * pic::progOptions()
{
  return "" ;
}
/** No descriptions */
int pic::setProgOptionNumber(int opt)
{
  int k = programming_option ;
  programming_option = opt ;
  return k ;
}
