#include <stdio.h>
#include <ctype.h>
#include <getopt.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>


#include "pic12C508.hh"


// ========================== 12C508	specific code ======================

pic12C508::pic12C508()
{
    memaddr1 = 0x0 ; memaddr2 = 0x1FF ;
    idaddr1 = 0x200 ; idaddr2 = 0x203 ;
    cfaddr1 = cfaddr2 = 0xFFF ;
    backup_osccal1 = 0x100 ; // test only !!!!
    backup_osccal2 = 0x101 ; // test only !!!!
}



// 12Cxx chips have 12 bits words
int pic12C508::wmask()
{
    return 0xFFF ;
}

/* -----------------------------------------------------------------------
   Program our RAM buffer into the PIC

   Returns:
   0 == OK
   1 == programm memory error
   2 == EEPROM data error (never used for 12Cxxx)
   4 == ID error
   8 == config fuses error

   the error codes are OR-ed
   ----------------------------------------------------------------------- */
int pic12C508::program_pic(ostream& out)
{
    int stat , fail  ;
    int x;

    stat = 0 ;

    // programming init sequence
    pulseEngine("cdP,Bw100", 0) ;

    // After entering programming mode, PC points to 0xFFF (Config fuses)
    if( mem()[cfaddr1] != INVALID)
    {

        unsigned int safe_cf = mem()[cfaddr1] ;
        // comment out the following line to authorize CF == 0
        //safe_cf |= 0x8 ;

        for(fail = 1 ; fail < 100 ; ++fail)
        { // conf fuses require a total of 10ms Pulses
            //  == 100 x 100 us
            if( pulseEngine("k2,S,k8w100k14w5000k4,F,V",&safe_cf ) == 0 ) break  ;
        }
        if(fail >= 100)
        {
            stat |= 8 ;
            warning(out,"Config fuses",cfaddr1) ;
        }
        else // 11 x fail additionnal passes
        {
            for(fail *= 11 ; fail > 0 ; --fail)
                pulseEngine("k2w1Sw1k8w100k14w5000", &safe_cf) ;
        }
    }

    //
    // increment adress  (code == 6) instruction rolls PC from 0xFFF to 0x000
    //
    // Program the PIC's main program code memory  + ID location
    // (this is possible because they are contiguous)
    for ( x = memaddr1; x <= idaddr2 ; ++x )
    {
        pulseEngine("k6,",0) ;

        // memaddr2 contains clock calibration value and should be
        // reprogramed with its original value
        if(mem()[x] != INVALID && pulseEngine("k4,F,V" , mem()+x) != 0)
        {
            for(fail = 1 ; fail <= 8 ; ++fail)
                if(pulseEngine("k2,S,k8w200k14w5000k4,F,V", mem()+x) == 0)
                    break  ;
            if(fail > 8 )
            {
                if( x >=idaddr1 )
                {
                    stat |= 4 ;
                    warning(out,"ID data", x) ;
                }
                else
                {
                    stat |= 1 ;
                    warning(out,"Prog mem()ory", x) ;
                }
            }
            else // 11 x fail additionnal passes
            {
                for(fail *= 11 ; fail > 0 ; --fail)
                    pulseEngine("k2w1Sw1k8w100k14w5000", mem()+x) ;
            }

        }
    }

    /* shutdown sequence VPP off, VDD off, etc.. */
    pulseEngine("w1bw1pw1" , 0);

    return stat ;
}

/* -----------------------------------------------------------------------
   Read the PIC's memory and compare it with RAM buffer.
   RAM buffer words equals to INVALID (0xFFFF) are considered as
   dummy and thus are NOT significant for the tests.

   Returns:
   0 == OK
   1 == programm memory error
   2 == EEPROM data error
   4 == ID error
   8 == config fuses error

   the error codes are OR-ed
   ----------------------------------------------------------------------- */
int pic12C508::verify_pic(ostream& out)
{
    int status;
    int i ;

    status = 0 ;
    // save ram buffer
    FragBuffer ram ;
    mem().swap(ram) ;  // exchange buffers, leave adresses in mem inchanged

    read_pic() ;

    // compare  program memory
    for(i = memaddr1 ; i <= memaddr2 ; ++i)
        if(mem()[i] != ram[i] && ram[i] != INVALID )
        {
            status |= 1 ;
            warning(out,"Prog mem()ory",i,mem()[i],ram[i]);
        }

    // compare ID
    for(i = idaddr1 ; i <= idaddr2 ; ++i)
        if(mem()[i] != ram[i] && ram[i] != INVALID)
        {
            status |= 4 ;
            warning(out,"ID data",i,mem()[i],ram[i]);
        }

    // compare config fuses
    if(mem()[cfaddr1] != ram[cfaddr1] && ram[cfaddr1] != INVALID)
    {
        status |= 8 ;
        warning(out,"Config fuses",cfaddr1,mem()[cfaddr1],ram[cfaddr1]);
    }

    // restore ram buffer
    mem().swap(ram) ;

    return status;
}
// -------------------------------------------------------------------------
//	Erase the pic
//	Always return 16 (non applicable)
// 	Because 12C508 is only UV erasable
// -------------------------------------------------------------------------

int pic12C508::erase_pic()
{
    return 16  ;
}


// -----------------------------------------------------------------------
//   Read the PIC's memory into our RAM buffer
// -----------------------------------------------------------------------

void pic12C508::read_pic()
{
    int x;
    unsigned int word ;

    /* general read initialization */
    pulseEngine("cdPw1Bw1", &word) ;

    // read Config Fuses
    pulseEngine("k4w1Rw1", mem()+cfaddr1) ;


    /* Read the PIC's  program memory  + ID */
    for ( x = memaddr1 ; x <= idaddr2 ; x++ )
    {

        pulseEngine("k6w1k4w1Rw1", mem()+x ) ;
    }

    // power off sequence
    pulseEngine("pw1bw1", 0 )  ;
}

const char * pic12C508::names()
{
    return "p12c508 p12c508a p12ce508 p12ce518" ;
}

const char * pic12C508::config_bits()
{
    // note: documentation specify bits 11-5 as "write as 0"
    // but they  actually "read as 1" - I force them to 1
    // to avoid spurious error messages during verifications
    return "1(11-5) MCLRE[4] CP[3] WDTE[2] FOSC[1-0]" ;
}
const char * pic12C508::SPLocs()
{
  return "1FF{OSCCAL word}" ;
}

/** No descriptions */
const char * pic12C508::progOptions()
{
  return "{First option}{second option}{third one}" ;
}
