#include <stdio.h>
#include <ctype.h>
#include <getopt.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>


#include "pic12C671.hh"


// ========================== 12C671	specific code ======================

pic12C671::pic12C671()
{
    memaddr1 = 0x0 ; memaddr2 = 0x3FF ;
    idaddr1 = 0x2000 ; idaddr2 = 0x2003 ;
    cfaddr1 = cfaddr2 = 0x2007 ;
}


// 12C6xx chips use 14 bits words
int pic12C671::wmask()
{
    return 0x3FFF ;
}

/* -----------------------------------------------------------------------
   Program our RAM buffer into the PIC

   Parametrized version

   Returns:
   0 == OK
   1 == programm memory error
   2 == EEPROM data error (never used for 12Cxxx)
   4 == ID error
   8 == config fuses error

   the error codes are OR-ed
   ----------------------------------------------------------------------- */
int pic12C671::program_pic(ostream &out)
{
    int stat , fail  ;
    int x;


    stat = 0 ;

    // programming init sequence
    pulseEngine("cdP,Bw10", 0) ;


    // PC points to Addr 0
    // program memory
    for ( x = memaddr1; x <= memaddr2 ; ++x )
    {
        if(mem()[x] != INVALID && pulseEngine("k4,F,V,", mem()+x) != 0)
        {
            for(fail = 1 ; fail <= 25 ; ++fail)
                if(pulseEngine("k2,S,k8w100k14,w5000k4,F,V,", mem()+x) == 0) break  ;
            if(fail > 25 )
            {
                warning(out,"Prog memory", x) ;
                stat |= 1 ;
            }
            else // 3 x fail additionnal passes
            {
                for(fail *= 3 ; fail > 0 ; --fail)
                    pulseEngine("k2w1Sw1k8w100k14w5000", mem()+x) ;
            }

        }
        pulseEngine("k6,",0) ;  // next location
    }

    // enter configuration mode, PC points to 0x2000
    unsigned int dummy = wmask() ;
    pulseEngine("k0,S," , &dummy);

    // program ID locations
    for( x = idaddr1 ; x <= idaddr2 ; ++x)
    {
        if( mem()[x] != INVALID)
        {

            for(fail = 1 ; fail <= 25 ; ++fail)
            {
                if( pulseEngine("k2w1Sw1k8w100k14w5000k4w1Fw1V",mem()+x) == 0 ) break  ;
            }
            if(fail >= 100)
            {
                stat |= 4 ;
                warning(out,"ID data",x) ;
            }
            else // 3 x fail additionnal passes
            {
                for(fail *= 3 ; fail > 0 ; --fail)
                    pulseEngine("k2w1Sw1k8w100k14w5000", mem()+x) ;
            }
        }
        pulseEngine("k6w1",0) ;  // next location
    }

    // go to Config fuse location
    for(  ; x < cfaddr1 ; ++x ) pulseEngine("k6w1",0) ;  // next location

    // must apply 100 programming cycle

    // I remove memory protection because it is irreversible
    // and can make a JW chip unusable (protection bit cannot be UV erased)
    unsigned int safe_cf = mem()[cfaddr1] ;
    //safe_cf |= 0x3F60 ;

    for(fail = 1 ; fail <= 100 ; ++fail)
    {
        pulseEngine("k2w1Sw1k8w100k14w3000",&safe_cf)  ;
    }
    // verify
    if( pulseEngine("k4w1Fw1V",&safe_cf) != 0)
    {
        warning(out,"Config fuses",cfaddr1) ;
        stat |= 8 ;
    }

    /* shutdown sequence VPP off, VDD off, etc.. */
    pulseEngine("cdw1bw1pw1" , 0);

    return stat ;
}



// -----------------------------------------------------------------------
//   Read the PIC's memory into our RAM buffer
// -----------------------------------------------------------------------

void pic12C671::read_pic()
{
    int x;
    unsigned int word ;

    /* general read initialization */
    pulseEngine("cdPw1Bw1", &word) ;

    /* Read the PIC's  program memory  */
    for ( x = memaddr1 ; x <= memaddr2 ; x++ )
    {
        pulseEngine("w1k4w1Rw1k6", mem()+x ) ;
    }

    // enter configuration mode, PC points to 0x2000
    unsigned int dummy = wmask() ;
    pulseEngine("k0w1Sw1" , &dummy);

    for( x = idaddr1; x <= idaddr2 ; x++ )
    {
        pulseEngine("k4w1Rw1k6w1", mem()+x) ;
    }

    // go to Config fuse location
    for(  ; x < cfaddr1 ; ++x ) pulseEngine("k6w1",0) ;

    // read Config Fuses
    pulseEngine("k4w1Rw1", mem()+cfaddr1) ;

    // power off sequence
    pulseEngine("pw1bw1", 0 )  ;
}

const char * pic12C671::names()
{
    return "p12c671 p12ce673" ;
}

const char * pic12C671::config_bits()
{
    return "CP[13-8] MCLRE[7] CP[6-5] /PWRTE[4]"
           " WDTE[3] FOSC[2-0]" ;
}

