/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "log.h"

#include <qpopupmenu.h>
#include <qeventloop.h>
#include <qapplication.h>

#include <kmessagebox.h>

#include "Config.h"

using namespace Log;

//-----------------------------------------------------------------------------
const Widget::LogData Widget::DATA[Nb_LogTypes] = {
  { 0, "black",     DontShow, 0                     }, // command
  { 0, "red",       Show,     0                     }, // error
  { 0, "orange",    Show,     0                     }, // warning
  { 0, "black",     Show,     0                     }, // data
  { 0, "darkGreen", DontShow, "Show debug messages" }, // debug
  { 0, "blue",      Show,     0                     }  // info
};

Widget::Widget(QWidget *parent, const char *name)
  : QTextEdit(parent, name)
{
  for (uint i=0; i<Nb_LogTypes; i++)
    _modes[i] = DATA[i].mode;
  _modes[Debug] = (Config::showDebugLog() ? Show : DontShow);
  setTextFormat(LogText);
  setMinimumWidth(300);
}

void Widget::append(Type type, const QString &text)
{
  switch (_modes[type]) {
  case Show: {
    if ( text.isEmpty() ) break; // nothing to display...
    QTextEdit::append(QString("<font color=%1>%2</font>")
                      .arg(DATA[type].color).arg(text));
    updateContents(); // #### fix bug in Qt (display is messed up)
    ensureVisible(0, contentsHeight());
    // immediately visible...
    QApplication::eventLoop()->processEvents(QEventLoop::ExcludeUserInput);
    break;
  }
  case DontShow: break;
  }
}

void Widget::appendToLastLine(const QString &text)
{
  uint p = paragraphs() - 1;
  insertAt(text, p, paragraphLength(p));
  updateContents(); // #### fix bug in Qt (display is messed up)
  ensureVisible(0, contentsHeight());
  // immediately visible...
  QApplication::eventLoop()->processEvents(QEventLoop::ExcludeUserInput);
}

QPopupMenu *Widget::createPopupMenu(const QPoint &pos)
{
  _popup = QTextEdit::createPopupMenu(pos);
  _popup->insertItem("Clear", this, SLOT(clear()));
  _popup->insertSeparator();
  for (uint i=0; i<Nb_LogTypes; i++) {
    if ( DATA[i].visibleConfig==0 ) continue;
    _id[i] = _popup->insertItem(DATA[i].visibleConfig);
    _popup->setItemChecked(_id[i], _modes[i]!=DontShow);
  }
  connect(_popup, SIGNAL(activated(int)), SLOT(toggleVisible(int)));
  return _popup;
}

void Widget::toggleVisible(int id)
{
  for (uint i=0; i<Nb_LogTypes; i++) {
    if ( _id[i]==id ) {
      _modes[i] = (_modes[i]==DontShow ? Show : DontShow);
      if ( i==Debug ) Config::writeShowDebugLog(_modes[i]==Show);
      break;
    }
  }
}

//-----------------------------------------------------------------------------
void Client::log(Type type, const QString &message)
{
  if ( type==Error ) _error = message;
  if (_log) _log->append(type, message);
  if (_parent) _parent->log(type, message);
}

void Client::appendToLastLine(const QString &text)
{
  if (_log) _log->appendToLastLine(text);
  if (_parent) _parent->appendToLastLine(text);
}

void Client::setError(const QString &message, ShowMode showBoxMode)
{
  if ( showBoxMode==Show ) KMessageBox::error(0, message);
  log(Error, message);
}
