/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *   Copyright (C) 2002 Alain Gibaud <alain.gibaud@free.fr>                *
 *   Copyright (C) 1994-1998 Brian C. Lane                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "memory.h"

#include <qregexp.h>

#include <klocale.h>
#include <kmessagebox.h>

#include "direct_prog/FragBuffer.hh"
#include "direct_prog/pic.hh"
#include "direct_prog/direct_programmer.h"
#include "microchip/icd2.h"

#include "MemoryDisplayer.h"
#include "BinWordsEditor.h"
#include "FNU.h"
#include "microchip/misc.h"

const Memory::ArchitectureData Memory::ARCHITECTURE_DATA[Nb_Architectures] = {
// 16C5X: 9, 10, 11 or 12-bit program counter; 12-bit instructions; 8-bit data
  { "16C5X", 12, 12,  8 },
// 16X: 13-bit program counter; 14-bit instructions; 8-bit data
  { "16X",   13, 14,  8 }, // eeprom: max = 256 words
// 17X-18X: 21-bit program counter; 16-bit instructions; 8-bit data
  { "17X",   21, 16,  8 },
  { "18X",   21, 16,  8 },
// dsPIC: 23-bit program counter; 24-bit instructions; 16-bit data
  { "30X",   23, 24, 16 }, // eeprom: max = 2 kwords = 4 kbytes
};

const struct Memory::SpaceData Memory::SPACE_DATA[Nb_MemorySpaces] = {
  { "program",        I18N_NOOP("Program memory"),      true,  16 }, // program
  { "calibration",    I18N_NOOP("Calibration Bits"),    false, 0  }, // calibration
  { "user_ids",       I18N_NOOP("User ID Locations"),   true,  1  }, // user id
  { "device_ids",     I18N_NOOP("Device IDs"),          false, 0  }, // device ids
  { "config",         I18N_NOOP("Configuration Bits"),  true,  0  }, // config
  { "eeprom",         I18N_NOOP("Data EEPROM"),         true,  8  }, // eeprom
  { "debug_vector",   I18N_NOOP("Debug Vector"),        false, 0  }, // debug vector
  { "hardware_stack", I18N_NOOP("Hardware Stack"),      false, 0  }  // hardware stack
};

Memory::Memory()
{
  // security...
  for (uint i=0; i<Nb_MemorySpaces; i++)
    _spec[i].present = 0;
}

void Memory::setDevice(const QString &name)
{
  // #### FIXME: this shouldn't require a programmer...
  qDebug("Memory::setDevice %s", name.latin1());
  _device = name;
  Programmer &prog = Programmer::current();
  prog.setDevice(name);
  switch (prog.type()) {
    case DeviceUser::DirectProgrammer: {
      const pic *apic = static_cast<DirectProgrammer &>(prog).getPic();
      setDevice(*apic);
      break;
    }
    case DeviceUser::ICD2Programmer: {
      const pic_device &dpic = static_cast<Icd2Programmer &>(prog).dpic();
      setDevice(dpic);
      break;
    }
    default: Q_ASSERT(false); break;
  }
  clearData();
}

// XX[a-b] means XX0 ... XXa-b+1
// XX(a-b) means that all the bits from b to a are identical
//         and are (identically) named XX
// XX[a] and XX(a) have the same meaning
// optionnal commment in {}
Memory::Configs Memory::processPikdevConfig(pic *p, uint startAddress, uint nbCharsAddress)
{
  Memory::Configs configs;
  if ( p==0 ) return configs;
  const QString &s = p->config_bits();
  QRegExp r("\\s*([A-Z/0-9-]+)([\\[(])(\\d+)(-?)(\\d*)([])])?\\s*(\\{([^}]*)\\})?\\s*(;)?\\s*");
  int offset = 0;
  uint a = 0;
  Memory::Config v;
  for (;;) {
    int i = r.search(s, offset);
    if ( i!=offset ) {
      qWarning("%s: cannot match config string %s", p->names(), s.mid(offset).latin1());
      return configs;
    }
    uint start = r.cap(3).toUInt();
    uint end = start;
    bool readOnly = ( r.cap(1)=="0" || r.cap(1)=="1" || r.cap(1)=="X" );
    if ( !r.cap(4).isEmpty() ) {
      if ( r.cap(5).isEmpty() || (r.cap(2)=="(" && r.cap(6)=="]")
           || (r.cap(2)=="[" && r.cap(6)==")" ) ) {
        qWarning("%s: wrong format for bit range in %s", p->names(), r.cap(0).latin1());
        return configs;
      }
      end = r.cap(5).toUInt();
      if ( start>end ) {
        uint tmp = start;
        start = end;
        end = tmp;
      }
    }
    uint mask = 0;
    for (uint l=start; l<=end; l++) mask |= 1 << l;
    uint nb = (r.cap(2)=="(" ? 2 : 1 << (end-start) + 1);
    for (uint l=0; l<nb; l++) {
      Memory::ConfigBit b;
      b.readOnly = readOnly;
      b.comment = r.cap(8);
      b.mask_name = r.cap(1);
      b.value_name = (readOnly ? r.cap(1) : QString::number(l));
      b.mask = mask;
      b.value = (r.cap(2)=="(" ? (l==0 ? 0 : mask) : l << start);
      //qDebug("%s: %s %i %i", b.mask_name.latin1(), b.value_name.latin1(), b.mask, b.value);
      v.append(b);
    }
    offset += r.matchedLength();
    bool atEnd = ( offset==int(s.length()) );
    if ( r.cap(9)==";" || atEnd ) {
      uint address = startAddress + (p->byteAdresses() ? 2 : 1)*a;
      v.name = toHexLabel(address, nbCharsAddress);
      qDebug("add config %s %i %i", v.name.latin1(), address, v.count());
      configs[address] = v;
      v.clear();
      a++;
    }
    if (atEnd) break;
  }
  return configs;
}

void Memory::setDevice(const pic &capic)
{
  // #### this is a temporary hack to support pikdev devices... rfpic are
  // not included anyway.
  pic &apic = const_cast<pic &>(capic);

  switch (apic.picFamily()) {
    case 12: _arch = ARCH_16C5X; break;
    case 16: _arch = ARCH_16X; break;
    case 17: _arch = ARCH_17X; break;
    case 18: _arch = ARCH_18X; break;
    default: Q_ASSERT(false); _arch = ARCH_16X;
  }

  uint add_mult = (_arch==ARCH_18X ? 2 : 1);
  int add[10];
  const MemorySpace spaces[5] = { MemoryCode, MemoryEeprom, MemoryUserId, MemoryConfig, MemoryCal };
  apic.getBuffers(add[0],add[1], add[2],add[3], add[4],add[5], add[6],add[7], add[8],add[9]);
  for (uint i=0; i<5; i++) {
    _spec[spaces[i]].present = ( add[2*i]!=-1 );
    switch (spaces[i]) {
      case MemoryCode:
      case MemoryUserId:
      case MemoryCal:
      case MemoryConfig:
        _spec[spaces[i]].start = add_mult * add[2*i];
        _spec[spaces[i]].end = add_mult * add[2*i+1];
        _spec[spaces[i]].width = 2;
        _spec[spaces[i]].mask = apic.wmask();
        break;
      case MemoryEeprom:
        _eepromHexFileOffset = add_mult * add[2*i];
        _spec[MemoryEeprom].start = 0;
        _spec[MemoryEeprom].end = add_mult * add[2*i+1] - _eepromHexFileOffset;
        _spec[MemoryEeprom].width = 1;
        _spec[MemoryEeprom].mask = 0xFF;
        break;
      case MemoryDbgvec:
      case MemoryHwstack:
      case MemoryDeviceId:
      case Nb_MemorySpaces: Q_ASSERT(false); break;
    }
    // inc and init: not filled
  }
  _spec[MemoryDbgvec].present = 0;
  _spec[MemoryHwstack].present = 0;
  _spec[MemoryDeviceId].present = 0;
  _specialLocations = apic.SPLocs();

  _configs = processPikdevConfig(&apic, _spec[MemoryConfig].start, nbCharsAddress(MemoryConfig));
}

void Memory::setDevice(const pic_device &device)
{
  _arch = Architecture(device.arch);
  _spec[MemoryCode] = device.memory_code;
  _spec[MemoryCal] = device.memory_cal;
  _spec[MemoryUserId] = device.memory_id;
  _spec[MemoryConfig] = device.memory_config;
  _spec[MemoryEeprom] = device.memory_eeprom;
  _spec[MemoryDbgvec] = device.memory_dbgvec;
  _spec[MemoryHwstack] = device.memory_hwstack;
  _spec[MemoryDeviceId].present = 0;
  _eepromHexFileOffset = device.eeprom_hexfile_offset;

  // #### TODO specialLocations

  _configs.clear();
  for (uint i=0; device.config[i].name; i++) {
    Config v;
    //qDebug("memory config %i %s", i, device.config[i].name);
    Q_ASSERT( !_configs.contains(device.config[i].address) );
    for (uint k=0; device.config[i].bits[k].mask_name; k++) {
      //qDebug("  add value %i %s", k, device.config[i].bits[k].mask_name);
      ConfigBit b;
      b.mask = device.config[i].bits[k].mask;
      b.value = device.config[i].bits[k].val;
      b.mask_name = device.config[i].bits[k].mask_name;
      b.value_name = device.config[i].bits[k].val_name;
      v.append(b);
    }
    v.name = device.config[i].name;
    _configs[device.config[i].address] = v;
  }
}

int Memory::hexFileStart(MemorySpace space) const
{
  if ( space==MemoryEeprom ) return _spec[space].start + _eepromHexFileOffset;
  return _spec[space].start;
}

int Memory::hexFileEnd(MemorySpace space) const
{
  if ( space==MemoryEeprom ) return _spec[space].end + _eepromHexFileOffset;
  return _spec[space].end;
}

uint Memory::length(MemorySpace space) const
{
  if ( !_spec[space].present ) return 0;
  return _spec[space].end - _spec[space].start + 1;
}

QString Memory::label(MemorySpace space) const
{
//    if ( _packedEeprom ) return i18n("Data memory (packed 8-bit words)");
  return i18n("%1 (%2-bit words)").arg(i18n(SPACE_DATA[space].label))
         .arg(nbBitsWord(space));
}

uint Memory::nbBitsWord(MemorySpace space) const
{
  uint n = 0;
  for (int i = 1; i; i<<=1)
    if ( i & _spec[space].mask ) n++;
  return n;
}

uint Memory::nbBitsAddress(MemorySpace) const
{
  return ARCHITECTURE_DATA[_arch].nbBitsPC;
}

uint Memory::nbCharsWord(MemorySpace space) const
{
  //qDebug("%s: nbBitsWord=%i nbCharsWord=%i", SPACE_DATA[space].name, nbBitsWord(space), nbBitsWord(space)/4 + (nbBitsWord(space)%4 ? 1 : 0));
  return nbBitsWord(space)/4 + (nbBitsWord(space)%4 ? 1 : 0);
}

uint Memory::nbCharsAddress(MemorySpace space) const
{
  return nbBitsAddress(space)/4 + (nbBitsAddress(space)%4 ? 1 : 0);
}

MemorySpaceEditor *Memory::createEditor(MemorySpace space, QWidget *parent)
{
  if ( !_spec[space].present ) return 0;
  uint nbLines = SPACE_DATA[space].nbDisplayLines;
  if ( nbLines==0 ) nbLines = (length(space)/8>1 ? 2 : 1);
  switch (space) {
    case MemoryCode:
    case MemoryCal:
    case MemoryUserId:
    case MemoryEeprom:
      return new MemoryDisplayer(parent, *this, space, nbLines, 8);
    case MemoryConfig:
      return new BinWordsEditor(*this, parent);
    case MemoryDeviceId:
    case MemoryDbgvec:
    case MemoryHwstack: break;
    case Nb_MemorySpaces: Q_ASSERT(false); break;
  }
  return 0;
}

void Memory::clearData()
{
  _data = FragBuffer();
  for (uint i=0; i<Nb_MemorySpaces; i++) {
    MemorySpace space = MemorySpace(i);
    if ( !_spec[space].present ) continue;
    qDebug("clear space: %i", space);
    _data.setBlock(hexFileStart(space), hexFileEnd(space));
  }
}

bool Memory::isAddressInside(MemorySpace space, int address) const
{
  return ( address>=hexFileStart(space) && address<=hexFileEnd(space) );
}

uint *Memory::dataAtOffset(MemorySpace space, uint offset)
{
  return dataAtAddress(space, hexFileStart(space) + offset);
}

uint *Memory::dataAtAddress(MemorySpace space, int address)
{
  if ( !isAddressInside(space, address) ) return 0;
  return _data + address;
}

const uint *Memory::dataAtOffset(MemorySpace space, uint offset) const
{
  return dataAtAddress(space, hexFileStart(space) + offset);
}

const uint *Memory::dataAtAddress(MemorySpace space, int address) const
{
  if ( !isAddressInside(space, address) ) return 0;
  return const_cast<FragBuffer &>(_data) + address;
}

//-----------------------------------------------------------------------------
const char *Memory::HEX_FORMATS[Nb_HexFormats] = {
  "inhx8m", "inhx8s", "inhx16", "inhx32"
};

bool Memory::saveHexFile(const PURL &url, HexFormat format,
                         QString &error, QWidget *widget)
{
  if ( format==IHX8S ) {
    KMessageBox::sorry(widget, i18n("IHX8S hex format is currently not supported..."));
    return false;
  }

  PFile file(url);
  if ( !file.openForWriting(widget) ) {
    error = i18n("Cannot write file \"%1\"").arg(url.filepath());
    return false;
  }

  unsigned int oldseg = 0xFFFFFFFF;
  FragBuffer::iterator block =  _data.begin();
  int len;
  while ( fetch_next_block(block, &len) ) {
    // block found, write it
    uint seg = (*block).first >> 15 ; // 2 * seg address
    if ( format == IHX32 && seg != oldseg ) {
      char buf[50];
      unsigned int check = 0x02 + 0x04 + (seg >> 8) + (seg & 0xFF);
      sprintf(buf,":02000004%04X%02X\n", seg, -check & 0xFF );
      file.stream() << buf;
      oldseg = seg ;
    }
    write_hex_block(file.stream(), len,block, format);
  }
  file.stream() << ":00000001FF\n" ;
  return true;
}

/* Write one line of Intel-hex file
 * Original code source from Timo Rossi,
 * modified by Alain Gibaud to support large blocks write
 */
void Memory::write_hex_block(QTextStream &stream,
                          int reclen, 	/* length (in words) */
                          FragBuffer::iterator& data , /* pointer to 1st data word (incremented by function)*/
                          HexFormat format)
{
    int check = 0;
    char buf[20] ;
    unsigned int loc ;

    /* A recursive algorithm is nicer but I don't want to stress the stack */

    while(reclen > HEXBLKSIZE)
    {
        write_hex_block(stream, HEXBLKSIZE,  data,format) ;
        reclen -= HEXBLKSIZE ;
    }

    if(reclen <= 0) return ; /* Oops, block has just a HEXBLKSIZE * n size */

    // write out line start ...
    if(format == IHX8M || format == IHX32)
    {
        loc = (*data).first * 2 ;
        sprintf(buf,  ":%02X%04X00", 2*reclen, loc & 0xFFFF );
        stream << buf ;
        check += ((loc) & 0xff) + (((loc) >> 8) & 0xff) + 2*reclen;
    }
    else if( format == IHX16 )
    {
        loc =  (*data).first   ;
        sprintf(buf, ":%02X%04X00", reclen, loc & 0xFFFF );
        stream << buf ;
        check += (loc & 0xff) + ((loc >> 8) & 0xff) + reclen;
    }
    /* else Hmmm .. something wrong */

    for( ; reclen > 0; ++data, --reclen)
    {
        unsigned int word =  (*data).second  ;
        if(format == IHX8M || format == IHX32)
        {
            sprintf( buf, "%02X%02X", (int)(word & 0xff), (int)((word >> 8) & 0xff));
            stream << buf ;
        }
        else if( format == IHX16)
        {
            sprintf(buf, "%02X%02X", (int)((word >> 8) & 0xff), (int)(word & 0xff));
            stream << buf ;
        }
        /* else Hmmm .. */

        check += (word & 0xff) + ((word >> 8) & 0xff);
    }

    /* write checksum, assumes 2-complement */
    sprintf(buf, "%02X\n", (-check) & 0xff);
    stream << buf ;

}

/* -------------------------------------------------------------------------
    This routine detects the next block to output
    A block is a set of consecutive addresse words not
    containing 0xFFFFFFFF
    @return true if a block has been detected
    'it' is updated to point to the first address of the block
    '*len' contains the size of the block
*/
bool Memory::fetch_next_block(FragBuffer::iterator& it,  int *len)
{
    unsigned int startadr , curadr ;
    // for( i = *start ; (i < MAXPICSIZE) && (Mem[i] == INVALID) ; ++i) ;
    // skip non-used words

    // Search block start
    while( (it != _data.end()) && (*it).second == 0xFFFFFFFFU ) ++it ;

    // if(i >= MAXPICSIZE ) return false ;
    if( it == _data.end() ) return false ;

    //for( *start = i ; (i < MAXPICSIZE) && (Mem[i] != INVALID) ; ++i) ;
    //*len = i - *start ;
    // search block end - a block may not cross a segment boundary
    FragBuffer::iterator itt(it) ;
    for( curadr = startadr = (*itt).first  , ++itt ; itt != _data.end() ; ++itt)
    {
        if( (*itt).first != curadr +1 ) break ;    // non contiguous addresses
        if((*itt).second == 0xFFFFFFFFU ) break ;   // unused word found
        if( (((*itt).first) & 0xFFFF0000U) != (curadr & 0xFFFF0000U)) break ; // cross segment boundary

        curadr =   (*itt).first  ;
    }
    *len = curadr - startadr + 1 ;

    return *len != 0  ;
}

bool Memory::loadHexFile(const PURL &url, QString &error, QWidget *widget)
{
  int res = read_hex(url, widget);
  switch (res) {
  case NoError: return true;
  case CRCError: error = i18n("CRC mismatch in file \"%1\"").arg(url.filepath()); return false;
  case UnrecognizedError: error = i18n("Unrecognized file \"%1\"").arg(url.filepath()); return false;
  case OpenError: error = i18n("Cannot open file \"%1\"").arg(url.filepath()); return false;
  }
  Q_ASSERT(false);
  return false;
}

/* -------------------------------------------------------------------------
   Read a Intel HEX file of either INHX16 or INHX8M format type, detecting
   the format automagicly by the wordcount and length of the line
   Tested  in 8 and 16 bits modes
   ------------------------------------------------------------------------ */
Memory::ReadHexError Memory::read_hex(const PURL &url, QWidget *widget)
{
    clearData(); // fill all usefull words with INVALID (ie 0xFFFFFFFF)

    PFile file(url);
    if ( !file.openForReading(widget) ) return OpenError;

    char	buf[1024];
    char	*p;
    unsigned int	count, addr, cksum, type, x, data, line, kend, bytecount,
    datah,datal ;
    unsigned int addrH = 0 ;  // upper 16 bits of 32 bits address  (inhx32 format)
    bool is_inhx32 = false ;

    line = 0;

    // all the configuration words must be programmed
    // the following turn any INVALID config word to a valid one
    // initialized with All 1
    for(uint k=0; k<length(MemoryConfig); ++k)
      *dataAtOffset(MemoryConfig, k) &= mask(MemoryConfig);

    /* Read each line of the file  (getline() remove trailing '\n')*/
    while ( file.readLine(buf, 1023) ) {
        line++;

        p = buf;
        /* If it isn't a valid intel hex line, then move to the next line */
        if( *p != ':' ) continue;

        /* because we need the exact line length for format detection,
           strip out CR/LF (and Any non-hex digit !)
        */
        for( kend = strlen(buf)-1 ; !isxdigit(buf[kend]) ; --kend) ;
        if(buf[kend] == ':') continue ; // ':' is alone
        buf[kend+1] = 0 ;
        bytecount = (kend-9) / 2 ; // number of data bytes of this record


        /* Get the byte count, the address and the type for this line.
           Because the checksum is computed from theses elements,
           i assume they are mandatory, thus I can read them together */

        if( sscanf( ++p, "%02X%04X%02X", &count , &addr, &type) != 3) return UnrecognizedError;
        /* Point to the first data byte */
        p += 8 ;

        /* Start calculating the checksum */
        cksum = count +
                (addr >> 8) +
                (addr & 0xFF) +
                type ;

        /* End of file field ?, exit now */
        if( type == 0x01 )
        {
            if( sscanf(p, "%02X", &data) != 1 )
                return UnrecognizedError;  // bad EOF record
            else if(((cksum+data) & 0xFF) != 0 )
                return CRCError;  // bad CRC
            else
                return NoError;   // OK, EOF
        }
        // example : linear extended record for adr 0x21xxxx
        // (blanks inserted for clarify)
        // :02 0000 04 0021 D9
        else if (type == 0x04)
        {
            if( sscanf(p, "%04X%02X", &addrH , &data) != 2 )
                return UnrecognizedError;  // bad address record

            cksum += (addrH & 0xFF) ;
            cksum += (addrH >> 8)   ;

            is_inhx32 = true ;

            if((cksum + data) & 0xFF != 0)
                return CRCError;  // bad CRC
            else
                continue ;   // goto next record
        }


        /* Figure out if its INHX16 or INHX8M
        	if count is a 16 bits words count => INHX16
         	if count is a byte count => INHX8M or INHX32
         */
        if( bytecount == count)
        {
            /* Processing a INHX8M line */
            for( x = 0; x < (count/2) ; x++ )
            {

                if( sscanf( p, "%02X%02X", &datal, &datah ) != 2) return UnrecognizedError;
                p += 4 ;

                _data[ (((addrH << 16) | addr)/2) + x ] = datal | (datah << 8);

                cksum += datah ;
                cksum += datal ;
            }
        }
        else if( bytecount == count * 2 )
        {
            if( is_inhx32) return UnrecognizedError; // inconsistant format
            /* Processing a INHX16 line */
            for( x = 0; x < count; x++ )
            {
                if( sscanf( p, "%02X%02X", &datah, &datal ) != 2) return UnrecognizedError;
                p += 4 ;

                _data[addr+x] = datal | (datah << 8);
                cksum += datah ;
                cksum += datal ;
            }
        } else return UnrecognizedError; // Brrrr !! Strange format.

        /* Process the checksum */
        if( sscanf( p, "%02X", &data ) != 1) return UnrecognizedError; // fmt error

        if( ((data + cksum) & 0xFF) != 0 ) return CRCError;
    }

    return NoError;
}
