/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *   Copyright (C) 2002-2003 Stephen Landamore <stephen@landamore.com>     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "icd2.h"

#include <klocale.h>

#include "misc.h"
#include "devices.h"
#include "device_list.h"
#include "icd2_prog_config.h"
#include "memory.h"

//-----------------------------------------------------------------------------
const char *Icd2::TARGET_MODE_LABELS[Nb_TargetModes] = {
  I18N_NOOP("Stopped"), I18N_NOOP("Running"), I18N_NOOP("Debug")
};
const uchar Icd2::TARGET_MODE_VALUES[Nb_TargetModes] = {
  0, // stopped
  1, // running
  2, // debug
};
const char *Icd2::MCLR_MODE_LABELS[Nb_MclrModes] = {
  I18N_NOOP("Reset Held"), I18N_NOOP("Reset Released")
};
const uchar Icd2::MCLR_MODE_VALUES[Nb_MclrModes] = {
  0, // held
  1, // released
};
const Icd2::VoltageTypeData Icd2::VOLTAGE_TYPE_DATA[Nb_VoltageTypes] = {
 { "37", 0.07988 }, // icd Vpp
 { "34", 0.0306  }, // target Vdd
 { "35", 0.07988 }, // target Vpp
};
const uchar Icd2::TEST_RESULT_VALUES[Nb_TestResults] = {
  0x00, // pass
  0x01, // low
  0x80  // high
};

Icd2::Icd2()
  : _dpic(0)
{
  _port.setParent(this);
}

void Icd2::serialInit(int speed, termios &info) const
{
  // always runs at 57600, 8n1
  cfsetispeed(&info, speed);
  cfsetospeed(&info, speed);
  // 8 data bits
  info.c_cflag &= ~CSIZE;
  info.c_cflag |= CS8;
  info.c_iflag &= ~ISTRIP;
  // no parity
  info.c_cflag &= ~(PARENB|PARODD);
  info.c_iflag &= ~INPCK;
  // 1 stop bit
  info.c_cflag &= ~CSTOPB;
  // raw access to serial port (uncooked)
  info.c_iflag &= ~(IGNCR|INLCR|ICRNL|IXOFF|IXON);
  info.c_oflag &= ~OPOST;
  info.c_cflag &= ~HUPCL;
  info.c_cflag |= (CLOCAL|CREAD);
  info.c_lflag &= ~(ICANON|ECHO|ECHONL|ISIG|IEXTEN|TOSTOP);
  // misc stuff
  info.c_cc[VMIN]  = 0; // wait for 1 char or timeout
  info.c_cc[VTIME] = 0; // wait this long, then fail
}

bool Icd2::connectDevice(const QString &port, char mode)
{
  resetError();

  _port.init(*this, port, 19200);
  _port.setFlowControl(true); // enable hardware flow control
  _port.setSpeed(19200);
  if ( hasError() ) return false;
  _port.setDtr(0); // reset
  msleep(10);
  _port.setDtr(1); // remove reset
  msleep(1);
  _port.send("Z");
  for (int i=0; i<4; i++) _port.getc(2000);
  if ( hasError() ) return false;
  _port.setDtr(0); // reset
  msleep(10);
  _port.setDtr(1); // remove reset
  _port.send(mode);
  char c = _port.getc(2000);
  if ( hasError() ) return false;
  return ( toupper(c)==mode );
}

void Icd2::setDevice(const QString &name)
{
  const DeviceList::Data *data = DeviceList::data(name);
  Q_ASSERT(data);
  _dpic = data->device;
  Q_ASSERT(_dpic);
}

bool Icd2::setTarget()
{
  // check that firmware is compatible
  FirmwareVersion fv = getFirmwareVersion();
  if ( fv.family()!=_dpic->icd2.efid ) {
    setError(i18n("Incorrect firmware loaded (non-matching family)."));
    return false;
  }
  // #### FIXME: check that revision is recent enough (6.43 ?)
  //if ( revision!=current_firmware(dpic.icd2.efid) ) {
    //printf( "incorrect firmware loaded, use lplab-maintain to load\n" );
    //printf( "firmware icd%08x.hex (available from mplab)\n",
    //        current_firmware(pic->icd2.efid) );
  //}
  // set target family
  return ( command("2A" + toHex(_dpic->icd2.famid, 2))
           && command("06" + toHex(_dpic->memory_code.end, 6)) );
}

bool Icd2::sendCommand(const QCString &s)
{
  QCString cs = s.upper();
  //format: <LL.....CC>
  QCString tx = "<";
  tx += toHex(cs.length() + 6, 2);
  tx += cs;
  uchar chk = tx[1] + tx[2];
  for (uint i=0; i<s.length(); i++)
    chk += cs[i];
  tx += toHex(chk, 2);
  tx += '>';
  qDebug("send_command: '%s'", tx.data());
  return _port.send(tx);
}

bool Icd2::receiveResponse(int timeout)
{
  // <LL...CC>
  _rx = _port.receive(3, timeout); // receive the first 2 bytes (i.e. the length)
  if ( _rx.length()!=3 ) return false;
  uint len = fromHex(_rx.mid(1, 2));
  _rx += _port.receive(len-3, timeout);
  if ( _rx.length()!=len ) return false;
  if ( _rx[0]!='[' || _rx[len-1]!=']' ) {
    setError(i18n("receive_response: Malformed string"));
    return false;
  }
  // verify the checksum
  uchar chk = 0;
  for (uint i=1; i<len-3; i++)
    chk += _rx[i];
  if ( chk!=fromHex(_rx.mid(len-3, 2)) ) {
    setError(i18n("receive_response: Bad checksum"));
    return false;
  }
  return true;
}

bool Icd2::command(const QCString &command)
{
  if ( !sendCommand(command) ) return false;
  if ( !receiveResponse(2000) ) return false; // timeout = 2 ms per default
  if ( command.mid(0, 2)!=_rx.mid(3, 2) ) {
    setError(i18n("command: Wrong return value (\"%1\"; was expecting \"%2\")")
             .arg(_rx.mid(3, 2)).arg(command.mid(0, 2)));
    return false;
  }
  return true;
}

bool Icd2::setFastSpeed()
{
    if ( !command("4D") ) return false; // go faster
    // no need to delay here - no more in-flight messages...
    if ( !_port.setSpeed(57600) ) {
      setError(i18n("set_fast_speed: Cannot set speed at 57600 baud"));
      return false;
    }
    msleep(100); // ...but we do need to delay here
    return true;
}

int Icd2::FirmwareVersion::value() const
{
  return bin2bcd(v[0]) << 24 | bin2bcd(v[1]) << 16 |
         bin2bcd(v[2]) << 8 | bin2bcd(v[3]);
}

Icd2::FirmwareVersion Icd2::getFirmwareVersion()
{
  FirmwareVersion fv;
  if ( !command("01") ) return fv;
  fv.v[1] = fromHex(_rx.mid(5, 2));
  fv.v[2] = fromHex(_rx.mid(7, 2));
  fv.v[3] = fromHex(_rx.mid(9, 2));
  if ( !command("07") ) return fv;
  fv.v[0] = fromHex(_rx.mid(5, 2));
  return fv;
}

bool Icd2::setTargetVddSource(VddSourceType type)
{
  return command("05" + toHex(type, 2));
}

double Icd2::getVoltage(VoltageType type)
{
  if ( !command(VOLTAGE_TYPE_DATA[type].command) ) return 0.0;
  return VOLTAGE_TYPE_DATA[type].factor * fromHex(_rx.mid(5, 2));
}

Icd2::TargetMode Icd2::getTargetMode()
{
  if ( !command("2C") ) return Nb_TargetModes;
  uchar mode = fromHex(_rx.mid(5, 2));
  for (uint i=0; i<Nb_TargetModes; i++)
    if ( mode==TARGET_MODE_VALUES[i] ) return TargetMode(i);
  Q_ASSERT(false);
  return Nb_TargetModes;
}

bool Icd2::setTargetMclr(MclrMode mode)
{
  return command("33" + toHex(MCLR_MODE_VALUES[mode], 2));
}

Icd2::TestData Icd2::selfTest()
{
  TestData td;
  if ( !command("02") ) return td;
  td.pass = true;
  for (uint k=0; k<TestData::Nb_VoltageTypes; k++) {
    uchar res = fromHex(_rx.mid(5 + 2*k, 2));
    uint i = 0;
    for (; i<Nb_TestResults; i++)
      if ( res==TEST_RESULT_VALUES[i] ) {
        td.voltages[k] = TestResult(i);
        break;
      }
    if ( i==Nb_TestResults ) td.voltages[k] = TestFail;
    if ( td.voltages[k]!=TestPass ) td.pass = false;
  }
  return td;
}

WordArray Icd2::readBlock(uint nbBytes, uint len)
{
    Q_ASSERT( nbBytes==1 || nbBytes==2 );
    char c = _port.getc(2000);
    if ( c!='{' ) {
      setError(i18n("read_block: Invalid begin character"));
      return WordArray();
    }
    WordArray data(len);
    uchar chk = 0;
    for (uint i=0; i<len; i++) {
      QCString s = _port.receive(2*nbBytes, 2000);
      if ( s.length()!=2*nbBytes ) {
        setError(i18n("read_block: Unspecified error reading data"));
        return WordArray();
      }
      data[i] = fromHex(s.mid(0, 2));
      if ( i<=1 ) qDebug("%c%c", s[0], s[1]);
      chk += s[0] + s[1];
      if ( nbBytes==2 ) {
        data[i] += fromHex(s.mid(2, 2)) * 256;
        if ( i<=1 ) qDebug("%c%c", s[2], s[3]);
        chk += s[2] + s[3];
      }
      if ( i!=0 && i%256==0 ) emit addProgress(256);
    }
    emit addProgress(len % 256);
    QCString s = _port.receive(2, 2000);
    if ( chk!=fromHex(s) ) {
      setError(i18n("read_block: Bad checksum: %1 (%2 expected)").arg(s).arg(toHex(chk, 2)));
      return WordArray();
    }
    if ( _port.getc(2000)!='}' ) {
      setError(i18n("read_block: Invalid end character"));
      return WordArray();
    }
    return data;
}

const Icd2::MemorySpaceData Icd2::MEMORY_SPACE_DATA[Nb_MemorySpaces] = {
//  read, write
  { "47", "43" }, // code
  { 0,    0    }, // cal ####
  { "4A", "46" }, // user id
  { 0,    0    }, // device id ####
  { "49", "45" }, // config
  { "48", "44" }, // eeprom
  { "40", "41" }, // debug vector
  { 0   , 0    }  // hw stack ####
};
WordArray Icd2::readMemory(MemorySpace space, const Memory &memory)
{
  const char *r = MEMORY_SPACE_DATA[space].read;
  if ( r==0 || !memory.isPresent(space) ) return WordArray();
  qDebug("readMemory nbBytes=%i length=%i", memory.width(space), memory.length(space));
  QCString com = r + toHex(memory.start(space), 8) + toHex(memory.length(space), 8);
  if ( !command(com) ) return WordArray();
  WordArray data = readBlock(memory.width(space), memory.length(space));
  if ( data.isEmpty() ) return data;
  (void)receiveResponse(-1);
  return data;
}

bool Icd2::writeBlock(uint nbBytes, const WordArray &data, uint offset, uint size)
{
  Q_ASSERT( nbBytes==1 || nbBytes==2 );
  Q_ASSERT( offset+size<=data.size() );
  uchar chk = 0;
  if ( ! _port.send("{") ) return false;
  for (uint i=0; i<size; i++) {
    QCString cs = toHex(data[offset+i] & 0xFF, 2);
    chk += cs[0] + cs[1];
    if ( nbBytes==2 ) {
      cs += toHex(data[offset+i] / 0x100, 2);
      chk += cs[2] + cs[3];
    }
    if ( i<10 ) qDebug("send; %i %s (%s)", i, cs.data(), toHex(data[offset+i] & 0xFFFF, 4).data());
    if ( !_port.send(cs) ) return false;
    if ( i!=0 && i%256==0 ) {
      emit addProgress(256);
      qDebug("sent %i bytes/words", i);
    }
  }
  emit addProgress(size % 256);
  qDebug("done sending %i bytes/words", size);
  return _port.send(toHex(chk, 2) + "}");
}

bool Icd2::writeMemory(MemorySpace space, const Memory &memory, const WordArray &data)
{
  const char *w = MEMORY_SPACE_DATA[space].write;
  if ( w==0 ) {
    setError(i18n("write_memory: writing not implemented"));
    return false;
  }
  if ( !memory.isPresent(space) ) {
    setError(i18n("Cannot write absent memory space"));
    return false;
  }
  Q_ASSERT( data.size()<=memory.length(space) );
  uint offset = 0;
  do {
    uint size = QMIN(data.size()-offset, 2048);
    QCString com = w + toHex(memory.start(space)+offset, 8) + toHex(size, 8);
    if ( !command(com) ) return false;
    if ( !writeBlock(memory.width(space), data, offset, size) ) return false;
    if ( !receiveResponse(-1) ) return false;
    offset += size;
  } while ( offset!=data.size() );
  return true;
}

bool Icd2::eraseAll()
{
  if ( !command("3300") ) return false; // mclr = 0 (put device in reset)
  if ( !command("29") ) return false;  // erase
  // #### need to erase config ??
  return true;
}

//-----------------------------------------------------------------------------
Icd2Programmer::Icd2Programmer()
{
  _icd2.setParent(this);
  connect(&_icd2, SIGNAL(addProgress(uint)), SLOT(slotAddProgress(uint)));
}

QString Icd2Programmer::port() const
{
  return Icd2Config::readPort();
}

bool Icd2Programmer::internalConnectDevice()
{
  resetError();
  log(Log::Debug, "connect device");
  if ( !_icd2.connectDevice(port(), 'U') ) return false;
  // ##### fast speed lock ICD2 on programming (drop bits ??)
  //log(Log::Debug, "set fast speed");
  //if ( !_icd2.setFastSpeed() ) return false;
  return !hasError();
}

bool Icd2Programmer::internalReadDeviceId(QString &name, uint &)
{
  // #### TODO
  // read_single_word(0x4a,0x2006) ??
  name = "?";
  log(Log::Warning, i18n("Read device id not implemented!"));
  return true;
}

bool Icd2Programmer::internalRead(Memory &memory)
{
  _steps = 0;
  for (uint i=0; i<Nb_MemorySpaces; i++) {
    MemorySpace space = MemorySpace(i);
    qDebug("read %i", i);
    WordArray data = _icd2.readMemory(space, memory);
    if ( hasError() ) return false;
    for (uint i=0; i<data.size(); i++) *memory.dataAtOffset(space, i) = data[i];
  }
  return true;
}

bool Icd2Programmer::internalErase()
{
  return _icd2.eraseAll();
}

bool Icd2Programmer::internalVerify(const Memory &memory)
{
  _steps = 0;
  for (uint i=0; i<Nb_MemorySpaces; i++) {
    MemorySpace space = MemorySpace(i);
    WordArray data = _icd2.readMemory(space, memory);
    if ( hasError() ) return false;
    for (uint i=0; i<data.size(); i++) {
      uint v = *memory.dataAtOffset(space, i) & memory.mask(space);
      uint d = data[i] & memory.mask(space);
      //qDebug("%i %i: %i %i", space, i, v, d);
      if ( v!=d ) {
        setError(i18n("Device memory doesn't match hex file (in %1 at address 0x%2: reading 0x%3 and expecting 0x%4).")
                 .arg(i18n(Memory::SPACE_DATA[space].label)).arg(toHex(memory.start(space)+i, 8))
                 .arg(toHex(d, 4)).arg(toHex(v, 4)));
        return false;
      }
    }
  }
  return true;
}

bool Icd2Programmer::internalProgram(const Memory &memory)
{
  _icd2.eraseAll();
  _steps = 0;
  for (uint i=0; i<Nb_MemorySpaces; i++) {
    MemorySpace space = MemorySpace(i);
    if ( !Memory::SPACE_DATA[i].writable || !memory.isPresent(space) ) continue;
    qDebug("program %i", i);
    //uint len = QMIN(10, memory.length(space)); // #### REMOVE ME
    WordArray data(memory.length(space));
    for (uint i=0; i<data.size(); i++) data[i] = *memory.dataAtOffset(space, i);
    if ( !_icd2.writeMemory(space, memory, data) ) return false;
  }
  return true;
}

void Icd2Programmer::internalSetDevice(const QString &name)
{
  _icd2.setDevice(name);
}

bool Icd2Programmer::internalSetupTarget()
{
  log(Log::Debug, "set target");
  if ( !_icd2.setTarget() ) return false;
  log(Log::Debug, "set source");
  if ( !_icd2.setTargetVddSource(Icd2Config::readVddSourceType()) ) return false;
  double v = _icd2.getVoltage(Icd2::IcdVpp);
  log(Log::Debug, QString("ICD2 Vpp = %1 V").arg(v));
  v = _icd2.getVoltage(Icd2::TargetVdd);
  log(Log::Debug, QString("Target Vdd = %1 V").arg(v));
  v = _icd2.getVoltage(Icd2::TargetVpp);
  log(Log::Debug, QString("Target Vpp = %1 V").arg(v));
  Icd2::TargetMode mode = _icd2.getTargetMode();
  log(Log::Debug, QString("Target mode: %1").arg(Icd2::TARGET_MODE_LABELS[mode]));
  log(Log::Debug, "hold reset");
  if ( !_icd2.setTargetMclr(Icd2::ResetHeld) ) return false;
  log(Log::Debug, "self-test");
  Icd2::TestData td = _icd2.selfTest();
  if ( !td.pass ) {
    setError(i18n("Self-test failed"));
    return false;
  }
  return !hasError();
}

uint Icd2Programmer::internalTotalSteps(Action action, const Memory &memory) const
{
  if ( action==Erase ) return 1;
  uint length = 0;
  for (uint i=0; i<Nb_MemorySpaces; i++) {
    MemorySpace space = MemorySpace(i);
    qDebug("space %i: %i %i", space, memory.isPresent(space), memory.length(space));
    if ( !memory.isPresent(space) ) continue;
    if ( action==Program && !Memory::SPACE_DATA[i].writable ) continue;
    length += memory.length(space);
  }
  return length;
}

void Icd2Programmer::slotAddProgress(uint steps)
{
  _steps += steps;
  emit progress(_steps);
}
