/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *   Copyright (C) 2002-2003 Stephen Landamore <stephen@landamore.com>     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#ifndef ICD2_H
#define ICD2_H

#include "programmer.h"
#include "serial.h"
#include "devices/devices.h"

typedef QMemArray<ushort> WordArray;

class Icd2 : public QObject, public Log::Client, public Device
{
  Q_OBJECT
public:
  Icd2();

// initialization
  void setDevice(const QString &name);
  bool connectDevice(const QString &port, char mode);
  bool setTarget();
  bool setFastSpeed();

// check
  class FirmwareVersion {
  public:
    FirmwareVersion() { v[0] = 0; }
    char v[4];
    char family() const { return v[0]; }
    int value() const;
  };
  FirmwareVersion getFirmwareVersion();

  enum VddSourceType { TargetVddSource = 0, Icd2VddSource = 1 };
  bool setTargetVddSource(VddSourceType type);
  enum VoltageType { IcdVpp = 0, TargetVpp, TargetVdd, Nb_VoltageTypes };
  double getVoltage(VoltageType type);

  enum TargetMode { TargetStopped = 0, TargetRunning, TargetDebug, Nb_TargetModes};
  static const char *TARGET_MODE_LABELS[Nb_TargetModes];
  TargetMode getTargetMode();

  enum MclrMode { ResetHeld = 0, ResetReleased, Nb_MclrModes};
  static const char *MCLR_MODE_LABELS[Nb_MclrModes];
  bool setTargetMclr(MclrMode mode);

  enum TestResult { TestPass = 0, TestLow, TestHigh, Nb_TestResults,
                    TestFail = Nb_TestResults };
  class TestData {
  public:
    TestData() : pass(false) {}
    bool pass;
    enum VoltageType { TargetVdd = 0, ModuleVpp, MclrGround, MclrVdd, MclrVpp,
                       Nb_VoltageTypes };
    TestResult voltages[Nb_VoltageTypes];
  };
  TestData selfTest();

// program
  const pic_device &dpic() const { return *_dpic; }
  WordArray readMemory(MemorySpace space, const Memory &memory);
  bool writeMemory(MemorySpace space, const Memory &memory, const WordArray &data);
  bool eraseAll();

signals:
  void addProgress(uint steps);

private:
  SerialPort _port;
  const pic_device *_dpic;
  QCString   _rx;

  struct VoltageTypeData {
    const char *command;
    double factor;
  };
  static const VoltageTypeData VOLTAGE_TYPE_DATA[Nb_VoltageTypes];
  static const uchar TARGET_MODE_VALUES[Nb_TargetModes];
  static const uchar MCLR_MODE_VALUES[Nb_MclrModes];
  static const uchar TEST_RESULT_VALUES[Nb_TestResults];
  struct MemorySpaceData {
    const char *read, *write;
  };
  static const MemorySpaceData MEMORY_SPACE_DATA[Nb_MemorySpaces];

  virtual bool needDrain() const { return true; }
  virtual void serialInit(int speed, termios &info) const;

  bool sendCommand(const QCString &s);
  bool receiveResponse(int timeout);
  bool command(const QCString &command);
  QCString receivedData() const { return _rx.mid(5, _rx.length()-8); }

  WordArray readBlock(uint nbBytes, uint length);
  bool writeBlock(uint nbBytes, const WordArray &data, uint offset, uint size);
};

class Icd2Programmer : public Programmer
{
  Q_OBJECT
public:
  Icd2Programmer();
  virtual Type type() const { return ICD2Programmer; }
  virtual QString port() const;
  const pic_device &dpic() const { return _icd2.dpic(); }

private slots:
  void slotAddProgress(uint steps);

private:
  uint _steps;
  Icd2 _icd2;

  virtual void internalSetDevice(const QString &name);
  virtual bool internalConnectDevice();
  virtual bool internalSetupTarget();
  virtual bool internalReadDeviceId(QString &name, uint &revision);
  virtual bool internalRead(Memory &memory);
  virtual bool internalErase();
  virtual bool internalVerify(const Memory &memory);
  virtual bool internalProgram(const Memory &memory);
  virtual uint internalTotalSteps(Action action, const Memory &memory) const;
};

#endif
