/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *   Copyright (C) 2003-2004 Alain Gibaud <alain.gibaud@free.fr>           *
 *   Copyright (C) 2002-2003 Stephen Landamore <stephen@landamore.com>     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "serial.h"

#include <stdio.h>
#include <errno.h>
#include <termio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/poll.h>
#include <sys/ioctl.h>
#include <linux/ppdev.h>

#include <qfile.h>

#include <klocale.h>

//------------------------------------------------------------------------------
const char *Port::LABELS[Nb_Types] = {
  I18N_NOOP("Serial Port"), I18N_NOOP("Parallel Port")
};

void Port::log(Log::Type logType, const QString &message)
{
  Log::Client::log(logType, i18n(LABELS[type()]) + ": " + message);
}

void Port::logData(const QString &s)
{
/*
  QString vs;
  for (uint i=0; i<s.length(); i++) {
    char c = s[i];
    switch (c) {
      case '\r': vs += "\\r"; break;
      case '\n': vs += "\\n"; break;
      case '<': vs += "&lt;"; break;
      case '>': vs += "&gt;"; break;
      default: {
        if ( c>=32 && c<=126 ) vs += c;
        else {
          QString tmp;
          tmp.sprintf("\\x%02x", c);
          vs += tmp;
        }
        break;
      }
    }
  }
  qDebug("%s", vs.latin1());
*/
  //log(Log::Debug, vs);
}

const QStringList &Port::deviceList(Type type)
{
  switch (type) {
    case Serial: return SerialPort::deviceList();
    case Parallel: return ParallelPort::deviceList();
    case Nb_Types: Q_ASSERT(false); break;
  }
  return SerialPort::deviceList();
}

//------------------------------------------------------------------------------
QStringList *SerialPort::_list = 0;

SerialPort::SerialPort()
  : _handle(-1), _device(0), _rlen(0), _rindex(0)
{}

int SerialPort::probePort(const QString &device)
{
  int id = open (device.latin1(), O_RDONLY|O_NOCTTY|O_NONBLOCK) ;
  if ( id==-1 ) return -2;
  ::close(id);
  id = open (device.latin1(), O_RDWR|O_NOCTTY|O_NONBLOCK) ;
  if ( id==-1 ) return -1; // read-only
  ::close(id);
  return 0;
}

const QStringList &SerialPort::deviceList()
{
  if ( _list==0 ) {
    _list = new QStringList;
    // standard serport in user space
    for(int i = 0; i < 8; ++i) {
      QString name = QString("/dev/ttyS%1").arg(i);
      if( probePort(name)==0 ) (*_list) += name;
    }
    // new devfs serport flavour
    for(int i = 0; i < 8; ++i) {
      QString name = QString("/dev/tts/%1").arg(i);
      if( probePort(name)==0 ) (*_list) += name;
    }
    // standard USB serport in user space
    for(int i = 0; i < 8; ++i) {
      QString name = QString("/dev/ttyUSB%1").arg(i);
      if( probePort(name)==0 ) (*_list) += name;
    }
    // new devfs USB serport flavour
    for(int i = 0; i < 8; ++i) {
      QString name = QString("/dev/usb/tts/%1").arg(i);
      if( probePort(name)==0 ) (*_list) += name;
    }
  }
  return *_list;
}

int SerialPort::baudToSpeed(int baud)
{
    switch( baud ) {
    case   9600: return B9600;
    case  19200: return B19200;
    case  38400: return B38400;
    case  57600: return B57600;
    case 115200: return B115200;
    }
    return 0;
}

bool SerialPort::init(const Device &device, const QString &port, int baud)
{
  close();
  resetError();

  _device = &device;
  int speed = baudToSpeed(baud);
  if ( speed==0 ) {
      setError(i18n("invalid baud rate"));
      return false;
    }
    if ( _handle!=-1 ) close();
    _handle = open(port.latin1(), O_NDELAY|O_RDWR|O_NOCTTY);
    if ( _handle== -1 ) {
      setError(i18n("cannot open device \"%1\"").arg(port));
      return false;
    }
    if ( tcgetattr(_handle, &_initinfo)==-1 ) {
      setError(i18n("\"tcgetattr1\" failed"));
      return false;
    }
    tcflush(_handle, TCIOFLUSH);
    if ( tcgetattr(_handle, &_info)==-1 ) {
      setError(i18n("\"tcgetattr2\" failed"));
      return false;
    }
    device.serialInit(speed, _info);
    if ( tcsetattr(_handle, TCSANOW, &_info)==-1 ) {
      setError(i18n("\"tcsetattr\" failed"));
      return false;
    }
    //fcntl(_handle, F_SETFL, 0); // BLOCKING
    return true;
}

bool SerialPort::setSpeed(int baud)
{
  if ( _handle==-1 || hasError() ) return false;
  tcdrain(_handle);
  tcflush(_handle, TCIOFLUSH);
  int speed = baudToSpeed(baud);
  cfsetispeed(&_info, speed);
  cfsetospeed(&_info, speed);
  if ( tcsetattr(_handle, TCSANOW, &_info)==-1 ) {
    setError(i18n("cannot set speed at %1 baud (tcsetattr failed)").arg(baud));
    return false;
  }
  return true;
}

void SerialPort::close()
{
    flush();
    if ( _handle!=-1 ) {
      tcsetattr(_handle, TCSANOW, &_initinfo);
      ::close(_handle);
    }
    _handle = -1;
    _device = 0;
}

bool SerialPort::setFlowControl(bool on)
{
    if ( _handle==-1 || hasError() ) return false;
    if (on) _info.c_cflag |= CRTSCTS;
    else _info.c_cflag &= ~CRTSCTS;
    _info.c_cc[VMIN]  = 0;
    _info.c_cc[VTIME] = 0;
    if ( tcsetattr(_handle, TCSANOW, &_info)==-1 ) {
      setError(i18n("cannot set flow control (tcsetattr failed)"));
      return false;
    }
    return true;
}

bool SerialPort::getCts()
{
    if ( _handle==-1 || hasError() ) return false;
    int status;
    if ( ioctl(_handle, TIOCMGET, &status)==-1 ) {
      setError(i18n("ioctl TIOCMGET failed"));
      return false;
    }
    return ( status & TIOCM_CTS );
}

bool SerialPort::setDtr(bool on)
{
    if ( _handle==-1 || hasError() ) return false;
    int control = TIOCM_DTR;
    return ( ioctl(_handle, on ? TIOCMBIS : TIOCMBIC, &control)!=-1 );
}

bool SerialPort::setRts(bool on)
{
    if ( _handle==-1 || hasError() ) return false;
    int control = TIOCM_RTS;
    return ( ioctl(_handle, on ? TIOCMBIS : TIOCMBIC, &control)!=-1 );
}

bool SerialPort::select(int timeout)
{
    if ( _handle==-1 || hasError() ) return false;
    struct pollfd pfd[1];
    pfd[0].fd = _handle;
    pfd[0].events = POLLIN;
    for (;;) {
      int ret = poll(pfd, 1, timeout);
      if ( ret==0 ) {
        setError(i18n("timeout waiting for data"));
        return false;
      }
      if ( ret==-1 ) {
        if ( errno==EINTR || errno==EAGAIN ) continue;
        break;
      }
      //if ( pfd[1].revents ) return -2; #### I don't understand that...
      if ( pfd[0].revents==POLLIN ) return true;
      break;
    }
    setError(i18n("an error occured while waiting for data"));
    return false;
}

bool SerialPort::send(char c)
{
  QCString s;
  s += c;
  return send(s);
}

bool SerialPort::send(const QCString &s)
{
    if ( _handle==-1 || hasError() ) return false;
    logData("send: \"" + s + "\"");
    QCString buf = s;
    while ( !buf.isEmpty() ) {
      int ww = write(_handle, buf.data(), buf.length());
      if ( ww==-1 ) {
        if ( errno==EINTR || errno==EAGAIN ) continue;
        setError(i18n("sending data failed"));
        return false;
      }
      buf = buf.right(buf.length()-ww);
    }
    if ( _device->needDrain() ) tcdrain(_handle);
    return true;
}

char SerialPort::getc(int timeout)
{
  if ( _handle==-1 || hasError() ) return -1;
  if ( _rindex==_rlen ) { // read more stuff into buffer
    if ( !select(timeout) ) return -1;
    _rindex = 0;
    _rlen = read(_handle, _rbuffer, BUFFER_MAX_LEN);
    if ( _rlen==-1 ) {
      _rlen = 0;
      if ( errno == EINTR || errno == EAGAIN ) {
        setError(i18n("reading data failed"));
        return -1;
      }
    }
    QCString s(_rbuffer, _rlen+1);
    logData("recv: \"" + s + "\"");
  }
  char c = _rbuffer[_rindex];
  _rindex++;
  return c;
}

QCString SerialPort::receive(uint length, int timeout)
{
  QCString s;
  if ( _handle==-1 || hasError() ) return s;
  for (uint i=0; i<length; i++) {
    s += getc(timeout);
    if ( s[s.length()-1]==-1 ) break;
  }
  return s;
}

void SerialPort::flush()
{
  if ( _handle==-1 ) return;
  tcdrain(_handle);
  tcflush(_handle, TCIFLUSH);
}

//------------------------------------------------------------------------------
QStringList *ParallelPort::_list = 0;

int ParallelPort::probePort(const QString &device)
{
  int fd = open(device.latin1(), O_RDWR);
  if ( fd==-1 ) return -2;
  int k = ioctl(fd, PPCLAIM);
  if ( k!=0 ) {
    ::close(fd) ;
    return -1; // read-only
  }
  ioctl(fd, PPRELEASE);
  ::close(fd);
  return 0;
}

const QStringList &ParallelPort::deviceList()
{
  if ( _list==0 ) {
    _list = new QStringList;
    // standard parport in user space
    for(int i = 0; i<8; ++i) {
      QString name = QString("/dev/parport%1").arg(i);
      if ( probePort(name)==0 ) (*_list) += name;
    }
    // new devfs parport flavour
    for(int i = 0; i<8; ++i) {
      QString name = QString("/dev/parports/%1").arg(i);
      if ( probePort(name)==0 ) (*_list) += name;
    }
  }
  return *_list;
}
