/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "programmer.h"

#include <klocale.h>

#include "Config.h"
#include "memory.h"
#include "direct_prog/direct_programmer.h"
#include "direct_prog/DirectPrgConfigWidget.h"
#include "microchip/icd2.h"
#include "microchip/Icd2PrgConfigWidget.h"

const char *Programmer::ACTION_LABELS[Nb_Actions] = {
  I18N_NOOP("Programming..."), I18N_NOOP("Reading..."), I18N_NOOP("Erasing..."),
  I18N_NOOP("Verifying..."), I18N_NOOP("Blank Checking...")
};

Programmer *Programmer::create(Type type)
{
  Programmer *p = 0;
  switch (type) {
    case DirectProgrammer: p = new ::DirectProgrammer; break;
    case ICD2Programmer: p = new Icd2Programmer; break;
    case PicstartPlus:
    case Promate2:
    case ICD: Q_ASSERT(false); return 0; // #### not implemented
    case GPAsm:
    case GPDAsm:
    case ICD2Debugger:
    case Nb_Types: Q_ASSERT(false); return 0;
  }
  p->init();
  return p;
}

ConfigWidget *Programmer::createConfigWidget(Type type, QWidget *parent)
{
  switch (type) {
    case DirectProgrammer: return new DirectProgConfigWidget(parent);
    case ICD2Programmer: return new Icd2ProgConfigWidget(parent);
    case PicstartPlus:
    case Promate2:
    case ICD: break; // #### not implemented
    case GPAsm:
    case GPDAsm:
    case ICD2Debugger:
    case Nb_Types: break;
  }
  Q_ASSERT(false);
  return 0;
}

Programmer *Programmer::_current = 0;

Programmer &Programmer::current()
{
  Type type = Config::programmerType();
  if ( _current && _current->type()!=type ) {
    delete _current;
    _current = 0;
  }
  if ( _current==0 ) _current = create(type);
  return *_current;
}

//-----------------------------------------------------------------------------
void Programmer::init()
{
  setDevice(DeviceList::defaultDevice(type()));
}

void Programmer::setDevice(const QString &name)
{
  _name = name;
  internalSetDevice(name);
}

bool Programmer::connectDevice()
{
  log(Log::Information, i18n("Connecting on port \"%1\"...").arg(port()));
  if ( !internalConnectDevice() ) return false;
  log(Log::Information, i18n("Connected"));
  return true;
}

bool Programmer::setupTarget()
{
  log(Log::Information, i18n("Using %1 with processor %2")
      .arg(i18n(DATA[type()].label)).arg(_name));
  return internalSetupTarget();
}

bool Programmer::readDeviceId(QString &name, uint &revision)
{
  log(Log::Information, i18n("Reading device id..."));
  if ( !internalReadDeviceId(name, revision) ) return false;
  if ( !name.isEmpty() && name!="?" )
    log(Log::Information, i18n("Device: %1 (%2)").arg(name).arg(revision));
  return true;
}

bool Programmer::doAction(Action action, Memory &memory)
{
  switch (action) {
    case Erase:
      emit totalSteps(internalTotalSteps(Erase, memory));
      log(Log::Information, i18n("Erasing..."));
      if ( !internalErase() ) return false;
      log(Log::Information, i18n("Erasing done"));
      return true;
    case Read:
      emit totalSteps(internalTotalSteps(Read, memory));
      memory.clearData();
      log(Log::Information, i18n("Reading device memory..."));
      if ( !internalRead(memory) ) return false;
      log(Log::Information, i18n("Reading done"));
      return true;
    case Program:
      emit totalSteps(internalTotalSteps(Program, memory) + internalTotalSteps(Verify, memory));
      log(Log::Information, i18n("Programming device memory..."));
      if ( !internalProgram(memory) ) return false;
      log(Log::Information, i18n("Verifying..."));
      if ( !internalVerify(memory) ) return false;
      log(Log::Information, i18n("Programming successful"));
      return true;
    case Verify:
      emit totalSteps(internalTotalSteps(Verify, memory));
      log(Log::Information, i18n("Verifying..."));
      if ( !internalVerify(memory) ) return false;
      log(Log::Information, i18n("Verifying successful"));
      return true;
    case BlankCheck:
      emit totalSteps(internalTotalSteps(BlankCheck, memory));
      log(Log::Information, i18n("Verifying blank..."));
      if ( !internalVerify(memory) ) return false;
      log(Log::Information, i18n("Verifying blank successful"));
      return true;
    case Nb_Actions: break;
  }
  Q_ASSERT(false);
  return false;
}
