#include "xml_device.h"

#include <qregexp.h>

#include "device_list.h"
#include "microchip/misc.h"
#include "FNU.h"
#include "direct_prog/PicFactory.hh"

const char *MEMORY_TECHNOLOGY_LABELS[Nb_MemoryTechnologies] = {
  "EE", "EPROM", "ROM", "ROMLESS", "SS"
};
const char *ICD2_FAMILY_LABELS[Nb_Icd2Families] = {
  "16F", "16FA", "18C", "18F", "18FEGG", "30F"
};
const char *ICD2_DEBUG_EXEC_LABELS[Nb_Icd2DebugExecs] = {
    "12F629", "16F648A",
    "16F676", "16F684",
    "16F688", "16F716",
    "16F7X7", "16F819",
    "16F872", "16F874",
    "16F877", "16F88",
    "18CX01", "18F_BANK00",
    "18F_BANK01", "18F_BANK02",
    "18F_BANK05", "18F_BANK07",
    "18F_BANK12", "18F_BANK14",
    "30F_REVB"
};
const char *ICD2_DEBUG_SET_LABELS[Nb_Icd2DebugSets] = {
    "12F", "16F", "16F648A",
    "16F684", "16F716", "16F7X7",
    "16F872", "18C", "18F",
    "30F"
};

class MaskData
{
public:
  MaskData(uint v = 0) : value(v) {}
  uint value;
  QMap<QString, uint> values;
};

uint nbCharsAddress(MemorySpace, const pic_device &d)
{
  uint nbBitsAddress = Memory::ARCHITECTURE_DATA[d.arch].nbBitsPC;
  return nbBitsAddress/4 + (nbBitsAddress%4 ? 1 : 0);
}

uint nbCharsWord(MemorySpace, const pic_device_mem_spec &spec)
{
  uint nbBitsWord = 0;
  uint field = QMAX(spec.mask, spec.init);
  for (int i=1; i; i<<=1)
    if ( i & field ) nbBitsWord++;
  return nbBitsWord/4 + (nbBitsWord%4 ? 1 : 0);
}

void XmlDevice::appendMemorySpace(MemorySpace space, const pic_device_mem_spec &spec,
                                  const pic_device &device, pic *p, int start, int end)
{
  QString name = Memory::SPACE_DATA[space].name;
  if ( !spec.present ) {
    if ( p && start!=-1 ) addComment(QString("#### %1 present in pikdev ???").arg(name));
    return;
  }
  if (p) {
    if ( start==-1 ) addComment(QString("#### %1 not present in pikdev ???").arg(name));
    else {
      if ( p->byteAdresses() ) {
        start *= 2;
        end *= 2;
      }
      if ( space==MemoryEeprom ) {
        start -= device.eeprom_hexfile_offset;
        end -= device.eeprom_hexfile_offset;
      }
      if ( start!=spec.start )
        addComment(QString("#### %1 start different in pikdev: %2").arg(name).arg(toHexLabel(start, nbCharsAddress(space, device))));
      if ( end!=spec.end )
        addComment(QString("#### %1 end different in pikdev: %2").arg(name).arg(toHexLabel(end, nbCharsAddress(space, device))));
    }
  }

  QDomElement e = _document.createElement("memory");
  e.setAttribute("name", name);
  e.setAttribute("start", toHexLabel(spec.start, nbCharsAddress(space, device)));
  e.setAttribute("end", toHexLabel(spec.end, nbCharsAddress(space, device)));
  e.setAttribute("width", spec.width);
  e.setAttribute("inc", spec.inc);
  e.setAttribute("mask", toHexLabel(spec.mask, nbCharsWord(space, spec)));
  e.setAttribute("init", toHexLabel(spec.init, nbCharsWord(space, spec)));
  _device.appendChild(e);
}

void XmlDevice::appendConfigBits(const pic_device &d, pic *p)
{
  const Memory::Configs configs =
    Memory::processPikdevConfig(p, d.memory_config.start, nbCharsAddress(MemoryConfig, d));

  uint k = 0;
  while ( d.config[k].name ) {
    QDomElement e = _document.createElement("config");
    _device.appendChild(e);
    e.setAttribute("name", d.config[k].name);
    e.setAttribute("address", toHexLabel(d.config[k].address, nbCharsAddress(MemoryConfig, d)));
    const pic_config_bit *bits = d.config[k].bits;
    if ( bits[0].mask_name==0 ) addComment("#### no mask for this config address ???");

    QMap<QString, MaskData> masks;
    uint i = 0;
    while ( bits[i].mask_name ) {
      if ( strlen(bits[i].mask_name)==0 ) addComment("#### empty mask name...");
      if ( masks.contains(bits[i].mask_name) ) {
        if ( masks[bits[i].mask_name].value!=uint(bits[i].mask) )
          addComment("#### mask not consistent !!!!!");
      } else {
        QMapIterator<QString, MaskData> it = masks.begin();
        for (; it!=masks.end(); ++it)
          if ( (it.data().value & bits[i].mask) ) addComment("#### overlapping masks !!!!");
        masks[bits[i].mask_name] = MaskData(bits[i].mask);
      }
      MaskData &mdata = masks[bits[i].mask_name];
      if ( strlen(bits[i].val_name)==0 ) addComment("#### empty bit config name...");
      if ( mdata.values.contains(bits[i].val_name) ) addComment("#### duplicate bit config name...");
      if ( (bits[i].val & bits[i].mask)!=bits[i].val ) addComment("#### bit not in mask !!!!");
      mdata.values[bits[i].val_name] = bits[i].val;
      i++;
    }

    QMapIterator<QString, MaskData> it = masks.begin();
    for (; it!=masks.end(); ++it) {
      QDomElement m = _document.createElement("mask");
      e.appendChild(m);
      m.setAttribute("name", it.key());
      m.setAttribute("value", toHexLabel(it.data().value, nbCharsWord(MemoryConfig, d.memory_config)));
      QMapIterator<QString, uint> itd = it.data().values.begin();
      for (; itd!=it.data().values.end(); ++itd) {
        QDomElement v = _document.createElement("value");
        m.appendChild(v);
        v.setAttribute("name", itd.key());
        v.setAttribute("value", toHexLabel(itd.data(), nbCharsWord(MemoryConfig, d.memory_config)));
      }
    }

    k++;
  }
}

void XmlDevice::appendDevice(const QString &name)
{
  const DeviceList::Data *data = DeviceList::data(name);
  Q_ASSERT(data->device);
  const pic_device &d = *data->device;
  pic *p = PicFactory::factory()->getPic(DeviceList::toGPUtilsName(name));
  int prog_s, prog_e, ee_s, ee_e, id_s, id_e, conf_s, conf_e, cal_s, cal_e;
  if (p) p->getBuffers(prog_s, prog_e, ee_s, ee_e, id_s, id_e, conf_s, conf_e, cal_s, cal_e);

  // attributes
  _device = _document.createElement("device");
  _root.appendChild(_device);
  _device.setAttribute("name", name);
  _device.setAttribute("architecture", Memory::ARCHITECTURE_DATA[d.arch].name);
  if (p) {
    int f = p->picFamily();
    bool ok = false;
    switch (f) {
      case 12: ok = ( d.arch==ARCH_16C5X ); break;
      case 16: ok = ( d.arch==ARCH_16X ); break;
      case 17: ok = ( d.arch==ARCH_17X ); break;
      case 18: ok = ( d.arch==ARCH_18X ); break;
    }
    if ( !ok ) addComment(QString("#### not matching family in pikdev: %1").arg(f));
  }
  _device.setAttribute("memory_technology", MEMORY_TECHNOLOGY_LABELS[d.memtech]);
  _device.setAttribute("vdd_min", d.vdd.min);
  _device.setAttribute("vdd_max", d.vdd.max);
  _device.setAttribute("vdd_nominal", d.vdd.nominal);
  _device.setAttribute("vpp_min", d.vpp.min);
  _device.setAttribute("vpp_max", d.vpp.max);
  _device.setAttribute("vpp_nominal", d.vpp.nominal);

  if ( p==0 ) addComment("#### device ID ???");
  else {
    QStringList names = QStringList::split(' ', p->names());
    int i = names.findIndex(DeviceList::toGPUtilsName(name));
    if ( i==-1 ) addComment("#### cannot find device name in pikdev list !!!");
    else {
      QStringList ids = QStringList::split(' ', p->IDs());
      if ( ids.count() ) {
        if ( ids.count()!=names.count() ) addComment("#### No. of IDs different from no. of names !!!");
        else _device.setAttribute("id", "0x" + ids[i]);
      } else addComment("#### not autodetectable... ??");
    }
  }

  // device memory
  if ( d.memory_code.present && p ) {
    if ( p->wmask()!=d.memory_code.mask )
      addComment(QString("#### program mask different in pikdev: %1").arg(toHexLabel(p->wmask(), nbCharsWord(MemoryCode, d.memory_code))));
    if ( p->byteAdresses() && d.memory_code.width!=2 )
      addComment(QString("#### program width different in pikdev: %i").arg(p->byteAdresses()));
  }
  appendMemorySpace(MemoryCode, d.memory_code, d, p, prog_s, prog_e);
  appendMemorySpace(MemoryCal, d.memory_cal, d, p, cal_s, cal_e);
  appendMemorySpace(MemoryUserId, d.memory_id, d, p, id_s, id_e);
  appendMemorySpace(MemoryConfig, d.memory_config, d, p, conf_s, conf_e);
  appendMemorySpace(MemoryEeprom, d.memory_eeprom, d, p, ee_s, ee_e);
  if ( d.memory_eeprom.present )
    _device.setAttribute("eeprom_hexfile_offset", toHexLabel(d.eeprom_hexfile_offset, nbCharsAddress(MemoryEeprom, d)).latin1());
  appendMemorySpace(MemoryDbgvec, d.memory_dbgvec, d, 0, 0, 0);
  // hwstack is never filled... (?)

  // config bits
  appendConfigBits(d, p);

  // supported programmers
  if ( d.support & SUPPORT_PICSTART_PLUS ) {
    QDomElement e = _document.createElement("prog");
    _device.appendChild(e);
    e.setAttribute("name", "picstart_plus");
    e.setAttribute("first_revision", toHexLabel(d.picstartplus.firstrev, 6));
    e.setAttribute("algorithm", d.picstartplus.algorithm);
    e.setAttribute("overprogram", d.picstartplus.overprogram);
    e.setAttribute("tries", d.picstartplus.tries);
  }
  if ( d.support & SUPPORT_PROMATE2 ) {
    QDomElement e = _document.createElement("prog");
    _device.appendChild(e);
    e.setAttribute("name", "promate2");
    e.setAttribute("first_revision", toHexLabel(d.promate2.firstrev, 6));
    e.setAttribute("os", toHexLabel(d.promate2.os, 2));
    e.setAttribute("socket", toHexLabel(d.promate2.socket, 2));
    e.setAttribute("part", toHexLabel(d.promate2.part, 2));
  }
  if ( d.support & SUPPORT_ICD ) {
    QDomElement e = _document.createElement("prog");
    _device.appendChild(e);
    e.setAttribute("name", "icd");
    e.setAttribute("first_revision", toHexLabel(d.icd.firstrev, 6));
    e.setAttribute("part", toHexLabel(d.icd.part, 2));
  }
  // if supports icd2 debugger: it should support icd2 programmer...
  if ( d.support & SUPPORT_ICD2D ) Q_ASSERT( (d.support & SUPPORT_ICD2P) );
  if ( d.support & SUPPORT_ICD2P ) {
    QDomElement e = _document.createElement("prog");
    _device.appendChild(e);
    e.setAttribute("name", "icd2");
    e.setAttribute("family", ICD2_FAMILY_LABELS[d.icd2.family]);
    e.setAttribute("efid", toHexLabel(d.icd2.efid, 2));
    e.setAttribute("famid", toHexLabel(d.icd2.famid, 2));
    if ( d.support & SUPPORT_ICD2D ) {
      e.setAttribute("debug_exec", ICD2_DEBUG_EXEC_LABELS[d.icd2.debugexec]);
      e.setAttribute("debug_set", ICD2_DEBUG_SET_LABELS[d.icd2.debugset]);
    }
  }
  if (p) {
    QDomElement e = _document.createElement("prog");
    _device.appendChild(e);
    e.setAttribute("name", "direct_programmers");
    if ( p->progOptions() ) addComment(QString("#### program options: %1").arg(p->progOptions()));
  }

  if (p) {
    if ( p->SPLocs() ) addComment(QString("#### special locations: %1").arg(p->SPLocs()));
  } else addComment("#### has special locations ???");
}

void XmlDevice::createFile()
{
  _document = QDomDocument("piclab");
  _root = _document.createElement("xml");
  _document.appendChild(_root);

  // use device from lplab
  QStringList names = DeviceList::names();
  QStringList::const_iterator it = names.begin();
  for (; it!=names.end(); ++it) {
    const DeviceList::Data *data = DeviceList::data(*it);
    Q_ASSERT(data);
    //qDebug("--> device %s", (*it).latin1());
    QDomComment comment = _document.createComment("------ Device: " + *it + " ------");
    _root.appendChild(comment);
    if ( data->supported[DeviceUser::DirectProgrammer] )
      comment.appendData(" in pikdev");

    if ( data->device==0 ) {
      comment.appendData("#### not in lplab\n");
      continue;
    }

    comment.appendData("\n");
    appendDevice(*it);
  }

  // save file
  PURL url = PURL::getSave(PDirectory::home(), QString::null, 0, QString::null);
  if ( url.isEmpty() ) return;
  url.write(_document.toString(), 0);
}

void XmlDevice::addComment(const QString &text)
{
  QDomComment c = _document.createComment(text + "\n");
  _device.appendChild(c);
}
